/*
 * prof.c
 *
 * Copyright (c) ProSyst Software
 * All rights reserved
 *
 * Written by Vladislav Doshev,
 *            Alexi     Jordanov   (alexi91@yahoo.com),
 *            Ivo       Karabashev
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <stddef.h>

#ifdef _MEM_USAGE_
#include <malloc.h>
#include <crtdbg.h>
#endif

#include "prof.h"

/**
 * @author  Vladislav Doshev
 * @author  Alexi     Jordanov (alexi91@yahoo.com)
 * @author  Ivo       Karabashev
 * @version 1.00, 02/11/2000
 */

static void prof_notify_event(JVMPI_Event *event);

JavaVM *jvm;

int debug = FALSE;
int prof_is_on;                        /* whether prof is enabled */
int prof_fd = -1;                      /* non-zero file or socket descriptor */

#define PROF_DEFAULT_TRACE_DEPTH 4
int max_trace_depth = PROF_DEFAULT_TRACE_DEPTH;
int prof_trace_depth = PROF_DEFAULT_TRACE_DEPTH;
int thread_in_traces = TRUE;
int lineno_in_traces = TRUE;
float prof_cutoff_point = 0.0001f;
int dump_on_exit = FALSE;
int alloc_sites = FALSE;
int cpu_sampling = FALSE;
int send_data_on_exit = FALSE;

/* triggers */
int triggers = FALSE;

jlong total_alloced_bytes;
jlong total_alloced_instances;
long total_live_bytes = 0;
long total_live_instances = 0;

/*
 * PRESERVE THIS LOCK ORDER TO AVOID DEADLOCKS!!!
 *
 * prof_dump_lock => DisableGC/EnableGC => data_access_lock
 *
 */

/* Data access Lock */
JVMPI_RawMonitor data_access_lock;
/* Dump lock */
JVMPI_RawMonitor prof_dump_lock;

/* profiler interface */
JVMPI_Interface *prof_jvm_interface;

unsigned int micro_sec_ticks;

#ifdef _MEM_USAGE_
unsigned int mem_usage_total = 0;
unsigned int mem_usage = 0;
unsigned int mem_usage_peak = 0;
unsigned int mem_blocks = 0;
#endif

/**
 * Memory allocation and deallocation.
 */
void *prof_calloc(unsigned int size)
{
#ifdef _MEM_USAGE_
  void *p = _malloc_dbg(size, _CLIENT_BLOCK, "", __LINE__);
	mem_usage_total += size;
	mem_usage += size;
	if (mem_usage > mem_usage_peak) mem_usage_peak = mem_usage;
	mem_blocks++;
#else
  void *p = malloc(size);
#endif
  if (p == NULL) {
    fprintf(stderr, "PROF ERROR: ***Out of Memory***, exiting.\n");
    CALL(ProfilerExit)((jint)1);
  }

  memset(p, 0, size);
  return p;
}

#ifdef _MEM_USAGE_
void prof_free(void *memblock) {
	unsigned int i = _msize_dbg(memblock, _CLIENT_BLOCK);
	mem_usage -= i;
	mem_blocks--;
	_free_dbg(memblock, _CLIENT_BLOCK);
}
#endif

/* prof initialisation */
JNIEXPORT jint JNICALL JVM_OnLoad(JavaVM *vm, char *options, void *reserved)
{
  int res;
  jvm = vm;
  res = (*jvm)->GetEnv(jvm, (void **)&prof_jvm_interface, JVMPI_VERSION_1);
  if (res < 0) {
    return JNI_ERR;
  }
  prof_jvm_interface->NotifyEvent = prof_notify_event;
  prof_dump_lock = CALL(RawMonitorCreate)("_prof_dump_lock");
  prof_init_setup(options);

  return JNI_OK;
}

static void prof_reset_data(void)
{
  if (cpu_sampling) {
    prof_clear_trace_cost();
  }
}

static void prof_dump_data(void)
{
  fprintf(stderr, "Dumping");
  if (alloc_sites) {
    fprintf(stderr, " allocation sites ...");
    prof_output_sites(0, prof_cutoff_point);
  }
  if (cpu_sampling) {
    fprintf(stderr, " CPU usage by sampling running threads ...");
    prof_output_trace_cost(prof_cutoff_point);
  }
  prof_reset_data();
  prof_flush();
  fprintf(stderr, " done.\n");
}


static void prof_jvm_shut_down_event(void)
{
  static int already_dumped = FALSE;
  CALL(RawMonitorEnter)(prof_dump_lock);
  if (!prof_is_on || already_dumped) {
    CALL(RawMonitorExit)(prof_dump_lock);
    return;
  }
#ifdef _MEM_USAGE_
  printf("\nUsed memory: (instant) %u KB  (peak) %u KB  (total) %u KB  (blocks) %u\n", mem_usage / 1024, mem_usage_peak / 1024, mem_usage_total / 1024, mem_blocks);
#endif
  already_dumped = TRUE;
  {
    /* Disable all JVM/PI events since the VM has shutdown */
    jint i;
    for (i = 1; i <= JVMPI_MAX_EVENT_TYPE_VAL; i++) {
      (void) CALL(DisableEvent)(i, NULL);
    }
  }
  CALL(RawMonitorEnter)(data_access_lock);
  prof_output_unmarked_traces();
  CALL(RawMonitorExit)(data_access_lock);
  if (send_data_on_exit && cpu_sampling) {
    prof_output_trace_cost(prof_cutoff_point);
    prof_flush();
  }
  if (dump_on_exit) {
    prof_dump_data();
  }
  if (debug) {
    prof_print_tbl_hash_stats(&prof_name_table,   "NAME");
    prof_print_tbl_hash_stats(&prof_thread_table, "THREAD");
    prof_print_tbl_hash_stats(&prof_trace_table,  "TRACE");
    prof_print_tbl_hash_stats(&prof_frame_table,  "FRAME");
    prof_print_tbl_hash_stats(&prof_method_table, "METHOD");
    prof_print_tbl_hash_stats(&prof_class_table,  "CLASS");
    prof_print_tbl_hash_stats(&prof_site_table,   "SITE");
    prof_print_objmap_tbl_hash_stats();
  }
  prof_is_on = FALSE;
  prof_real_disconnect(); /* close(prof_fd); */
  CALL(RawMonitorExit)(prof_dump_lock);
}

static void prof_jvm_init_done_event(void)
{
  prof_start_listener_thread();
  prof_start_cpu_sampling_thread();
}

static void prof_notify_event(JVMPI_Event *event)
{
  switch(event->event_type) {
    case JVMPI_EVENT_JVM_INIT_DONE:
      prof_jvm_init_done_event();
      return;
    case JVMPI_EVENT_JVM_SHUT_DOWN:
      prof_jvm_shut_down_event();
      return;
    case JVMPI_EVENT_DATA_DUMP_REQUEST:
    case JVMPI_EVENT_DATA_RESET_REQUEST:
      return;
    case JVMPI_EVENT_CLASS_LOAD:
    case JVMPI_EVENT_CLASS_LOAD | JVMPI_REQUESTED_EVENT:
      prof_class_load_event(event->env_id,
        event->u.class_load.class_name,
        event->u.class_load.source_name,
        event->u.class_load.num_interfaces,
        event->u.class_load.num_static_fields,
        event->u.class_load.statics,
        event->u.class_load.num_instance_fields,
        event->u.class_load.instances,
        event->u.class_load.num_methods,
        event->u.class_load.methods,
        event->u.class_load.class_id,
        event->event_type & JVMPI_REQUESTED_EVENT);
      return;
    case JVMPI_EVENT_CLASS_UNLOAD:
      prof_class_unload_event(event->env_id, event->u.class_unload.class_id);
      return;
    case JVMPI_EVENT_OBJECT_DUMP | JVMPI_REQUESTED_EVENT:
      prof_object_dump_event(event->u.object_dump.data);
      return;
    case JVMPI_EVENT_OBJECT_ALLOC:
    case JVMPI_EVENT_OBJECT_ALLOC | JVMPI_REQUESTED_EVENT:
      prof_obj_alloc_event(event->env_id,
        event->u.obj_alloc.class_id,
        event->u.obj_alloc.is_array,
        event->u.obj_alloc.size,
        event->u.obj_alloc.obj_id,
        event->u.obj_alloc.arena_id,
        event->event_type & JVMPI_REQUESTED_EVENT);
      return;
    case JVMPI_EVENT_OBJECT_FREE:
      prof_obj_free_event(event->env_id, event->u.obj_free.obj_id);
      return;
    case JVMPI_EVENT_OBJECT_MOVE:
      prof_obj_move_event(event->env_id,
        event->u.obj_move.obj_id,
        event->u.obj_move.arena_id,
        event->u.obj_move.new_obj_id,
        event->u.obj_move.new_arena_id);
      return;
    case JVMPI_EVENT_THREAD_START:
    case JVMPI_EVENT_THREAD_START | JVMPI_REQUESTED_EVENT:
      prof_thread_start_event(event->u.thread_start.thread_env_id,
        event->u.thread_start.thread_name,
        event->u.thread_start.group_name,
        event->u.thread_start.parent_name,
        event->u.thread_start.thread_id,
        event->event_type & JVMPI_REQUESTED_EVENT);
      return;
    case JVMPI_EVENT_THREAD_END:
      prof_thread_end_event(event->env_id);
      return;
    case JVMPI_EVENT_METHOD_ENTRY:
      prof_method_entry_event(event->env_id, event->u.method.method_id);
      return;
    case JVMPI_EVENT_METHOD_EXIT:
      prof_method_exit_event(event->env_id, event->u.method.method_id);
      return;
    case JVMPI_EVENT_HEAP_DUMP | JVMPI_REQUESTED_EVENT:
      prof_heap_dump_event(event->u.heap_dump.begin,
        event->u.heap_dump.end,
        event->u.heap_dump.num_traces,
        event->u.heap_dump.traces);
      return;
    case JVMPI_EVENT_JNI_GLOBALREF_ALLOC:
      prof_jni_globalref_alloc_event(event->env_id,
        event->u.jni_globalref_alloc.obj_id,
        event->u.jni_globalref_alloc.ref_id);
      return;
    case JVMPI_EVENT_JNI_GLOBALREF_FREE:
      prof_jni_globalref_free_event(event->env_id, event->u.jni_globalref_free.ref_id);
      return;
    case JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTER:
    case JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTERED:
    case JVMPI_EVENT_RAW_MONITOR_CONTENDED_EXIT:
      prof_raw_monitor_event(event, event->u.raw_monitor.name, event->u.raw_monitor.id);
      return;
    case JVMPI_EVENT_MONITOR_CONTENDED_ENTER:
    case JVMPI_EVENT_MONITOR_CONTENDED_ENTERED:
    case JVMPI_EVENT_MONITOR_CONTENDED_EXIT:
      prof_monitor_event(event, event->u.monitor.object);
      return;
    case JVMPI_EVENT_MONITOR_WAITED:
    case JVMPI_EVENT_MONITOR_WAIT:
      prof_monitor_wait_event(event, event->u.monitor_wait.object, event->u.monitor_wait.timeout);
      return;
    case JVMPI_EVENT_MONITOR_DUMP | JVMPI_REQUESTED_EVENT:
      prof_monitor_dump_event(event);
      return;
    case JVMPI_EVENT_GC_START:
      prof_gc_start_event(event->env_id);
      return;
    case JVMPI_EVENT_GC_FINISH:
      prof_gc_finish_event(event->env_id,
        event->u.gc_info.used_objects,
        event->u.gc_info.used_object_space,
        event->u.gc_info.total_object_space);
      return;
  }
}
