/*
 * prof.h
 *
 * Copyright (c) ProSyst Software
 * All rights reserved
 *
 * Written by Vladislav Doshev,
 *            Alexi     Jordanov   (alexi91@yahoo.com),
 *            Ivo       Karabashev
 */

#ifndef _PROF_H
#define _PROF_H

#include "prof_md.h"
#include "jni.h"
#include "jvmpi.h"
#include "jlong_md.h"

#include "prof_global.h"
#include "prof_setup.h"
#include "prof_trace.h"
#include "prof_thread.h"
#include "prof_heapdump.h"
#include "prof_site.h"
#include "prof_io.h"
#include "prof_listener.h"
#include "prof_cpu.h"
#include "prof_monitor.h"
#include "prof_class.h"
#include "prof_method.h"
#include "prof_jni.h"
#include "prof_object.h"
#include "prof_name.h"
#include "prof_sys.h"
#include "prof_gc.h"
#include "prof_trigger.h"

/**
 * @author  Vladislav Doshev
 * @author  Alexi     Jordanov (alexi91@yahoo.com)
 * @author  Ivo       Karabashev
 * @version 1.00, 02/11/2000
 */

/*
 * -Xprof binary format:
 * (result either written to a file or sent over the network).
 * 
 *  header    "JAVA PROFILER AGENT 1.2" (0-terminated)
 *  u4        size of identifiers. Identifiers are used to represent
 *            UTF8 strings, objects, stack traces, etc. They usually
 *            have the same size as host pointers. For example, on
 *            Solaris and Win32, the size is 4.
 *  u4        high word 
 *  u4        low word    number of milliseconds since 0:00 GMT, 1/1/70
 * [record]*  a sequence of records.
 */

/*
 * Record format:
 *
 *  u1        a TAG denoting the type of the record
 *  u4        number of *microseconds* since the time stamp in the
 *            header. (wraps around in a little more than an hour)
 *  u4        number of bytes *remaining* in the record. Note that
 *            this number excludes the tag and the length field itself.
 * [u1]*      BODY of the record (a sequence of bytes)
 */

/*
 * The following TAGs are supported:
 *
 * TAG           BODY       notes
 * ----------------------------------------------------------
 * PROF_UTF8               a UTF8-encoded name  
 *
 *                id        name ID
 *               [u1]*      UTF8 characters (no trailing zero)
 *
 * PROF_LOAD_CLASS         a newly loaded class
 *
 *                u4        class serial number (> 0)
 *                id        class object ID
 *                u4        stack trace serial number
 *                id        class name ID
 *
 * PROF_UNLOAD_CLASS       an unloading class
 *
 *                u4        class serial_number
 *
 * PROF_FRAME              a Java stack frame
 *
 *                id        stack frame ID
 *                id        method name ID
 *                id        method signature ID
 *                id        source file name ID
 *                u4        class serial number
 *                i4        line number. >0: normal
 *                                       -1: unknown
 *                                       -2: compiled method
 *                                       -3: native method
 *
 * PROF_TRACE              a Java stack trace
 *
 *                u4        stack trace serial number
 *                u4        thread serial number
 *                u4        number of frames
 *               [id]*      stack frame IDs
 *
 *
 * PROF_ALLOC_SITES        a set of heap allocation sites, obtained after GC
 *
 *                u2        flags 0x0001: incremental vs. complete
 *                                0x0002: sorted by allocation vs. live
 *                                0x0004: whether to force a GC
 *                u4        cutoff ratio
 *                u4        total live bytes
 *                u4        total live instances
 *                u8        total bytes allocated
 *                u8        total instances allocated
 *                u4        number of sites that follow
 *               [u1        is_array: 0: normal object
 *                                    2: object array
 *                                    4: boolean array
 *                                    5: char array
 *                                    6: float array
 *                                    7: double array
 *                                    8: byte array
 *                                    9: short array
 *                                   10: int array
 *                                   11: long array
 *                u4        class serial number (may be zero during startup)
 *                u4        stack trace serial number
 *                u4        number of bytes alive
 *                u4        number of instances alive
 *                u4        number of bytes allocated
 *                u4]*      number of instance allocated
 *
 * PROF_START_THREAD       a newly started thread.
 *
 *                u4        thread serial number (> 0)
 *                id        thread object ID
 *                u4        stack trace serial number
 *                id        thread name ID
 *                id        thread group name ID
 *                id        thread group parent name ID
 *
 * PROF_END_THREAD         a terminating thread. 
 *
 *                u4        thread serial number
 *
 * PROF_HEAP_SUMMARY       heap summary
 *
 *                u4        total live bytes
 *                u4        total live instances
 *                u8        total bytes allocated
 *                u8        total instances allocated
 *
 * PROF_HEAP_DUMP          denote a heap dump
 *
 *               [heap dump sub-records]*
 *
 *                          There are four kinds of heap dump sub-records:
 *
 *                u1        sub-record type
 *
 *               PROF_GC_ROOT_UNKNOWN         unknown root
 *
 *                          id         object ID
 *
 *               PROF_GC_ROOT_THREAD_OBJ      thread object
 *
 *                          id         thread object ID  (may be 0 for a
 *                                     thread newly attached through JNI)
 *                          u4         thread sequence number
 *                          u4         stack trace sequence number
 *
 *               PROF_GC_ROOT_JNI_GLOBAL      JNI global ref root
 *
 *                          id         object ID
 *                          id         JNI global ref ID
 *
 *               PROF_GC_ROOT_JNI_LOCAL       JNI local ref
 *
 *                          id         object ID
 *                          u4         thread serial number
 *                          u4         frame # in stack trace (-1 for empty)
 *
 *               PROF_GC_ROOT_JAVA_FRAME      Java stack frame
 *
 *                          id         object ID
 *                          u4         thread serial number
 *                          u4         frame # in stack trace (-1 for empty)
 *
 *               PROF_GC_ROOT_NATIVE_STACK    Native stack
 *
 *                          id         object ID
 *                          u4         thread serial number
 *
 *               PROF_GC_ROOT_STICKY_CLASS    System class
 *
 *                          id         object ID
 *
 *               PROF_GC_ROOT_THREAD_BLOCK    Reference from thread block
 *
 *                          id         object ID
 *                          u4         thread serial number
 *
 *               PROF_GC_ROOT_MONITOR_USED    Busy monitor
 *
 *                          id         object ID
 *
 *               PROF_GC_CLASS_DUMP           dump of a class object
 *
 *                          id         class object ID
 *                          u4         stack trace serial number
 *                          id         super class object ID
 *                          id         class loader object ID
 *                          id         signers object ID
 *                          id         protection domain object ID
 *                          id         reserved
 *                          id         reserved
 *
 *                          u4         instance size (in bytes)
 *
 *                          u2         size of constant pool
 *                         [u2,        constant pool index,
 *                          ty,        type 
 *                                      2: object
 *                                      4: boolean
 *                                      5: char
 *                                      6: float
 *                                      7: double
 *                                      8: byte
 *                                      9: short
 *                                     10: int
 *                                     11: long
 *                          vl]*       and value
 *
 *                          u2         number of static fields
 *                         [id,        static field name,
 *                          ty,        type,
 *                          vl]*       and value
 *
 *                          u2         number of inst. fields (not inc. super)
 *                         [id,        instance field name,
 *                          ty]*       type
 *
 *               PROF_GC_INSTANCE_DUMP        dump of a normal object
 *
 *                          id         object ID
 *                          u4         stack trace serial number
 *                          id         class object ID
 *                          u4         number of bytes that follow
 *                          [vl]*      instance field values (class, followed
 *                                     by super, super's super ...)
 *
 *               PROF_GC_OBJ_ARRAY_DUMP       dump of an object array
 *
 *                          id         array object ID
 *                          u4         stack trace serial number
 *                          u4         number of elements
 *                          id         element class ID
 *                          [id]*      elements
 *
 *               PROF_GC_PRIM_ARRAY_DUMP      dump of a primitive array
 *
 *                          id         array object ID
 *                          u4         stack trace serial number
 *                          u4         number of elements
 *                          u1         element type
 *                                     4:  boolean array
 *                                     5:  char array
 *                                     6:  float array
 *                                     7:  double array
 *                                     8:  byte array
 *                                     9:  short array
 *                                     10: int array
 *                                     11: long array
 *                          [u1]*      elements
 *
 * PROF_CPU_SAMPLES        a set of sample traces of running threads
 *
 *                u4        total number of samples
 *                u4        # of traces
 *               [u4        # of samples
 *                u4]*      stack trace serial number
 *
 * PROF_CONTROL_SETTINGS   the settings of on/off switches
 *
 *                u4        0x00000001: alloc traces on/off
 *                          0x00000002: cpu sampling on/off
 *                u2        stack trace depth
 *
 */

#define PROF_UTF8                  0x01
#define PROF_LOAD_CLASS            0x02
#define PROF_UNLOAD_CLASS          0x03
#define PROF_FRAME                 0x04
#define PROF_TRACE                 0x05
#define PROF_ALLOC_SITES           0x06
#define PROF_HEAP_SUMMARY          0x07
#define PROF_START_THREAD          0x0A
#define PROF_END_THREAD            0x0B
#define PROF_HEAP_DUMP             0x0C
#define PROF_CPU_SAMPLES           0x0D
#define PROF_CONTROL_SETTINGS      0x0E

#define PROF_GRAPH_INFO            0x40
#define PROF_MONITOR_DUMP          0x41
#define PROF_TRIGGER_EVENT         0x42

#define PROF_MONITOR_ENTER         0x50
#define PROF_MONITOR_ENTERED       0x51
#define PROF_MONITOR_EXIT          0x52
#define PROF_MONITOR_WAIT          0x53
#define PROF_MONITOR_WAITED        0x54

/* heap dump constants */
#define PROF_GC_ROOT_UNKNOWN       0xff
#define PROF_GC_ROOT_JNI_GLOBAL    0x01
#define PROF_GC_ROOT_JNI_LOCAL     0x02
#define PROF_GC_ROOT_JAVA_FRAME    0x03
#define PROF_GC_ROOT_NATIVE_STACK  0x04
#define PROF_GC_ROOT_STICKY_CLASS  0x05
#define PROF_GC_ROOT_THREAD_BLOCK  0x06
#define PROF_GC_ROOT_MONITOR_USED  0x07
#define PROF_GC_ROOT_THREAD_OBJ    0x08

#define PROF_GC_CLASS_DUMP         0x20
#define PROF_GC_INSTANCE_DUMP      0x21 
#define PROF_GC_OBJ_ARRAY_DUMP     0x22
#define PROF_GC_PRIM_ARRAY_DUMP    0x23

/* When the VM is collected via a socket to the profiling client, the
 * client may send the VM a set of commands. The commands have the 
 * following format:
 *
 *  u1        a TAG denoting the type of the record
 *  u4        a serial number
 *  u4        number of bytes *remaining* in the record. Note that
 *            this number excludes the tag and the length field itself.
 * [u1]*      BODY of the record (a sequence of bytes)
 */

/* The following commands are presently supported:
 *
 * TAG           BODY       notes
 * ----------------------------------------------------------
 * PROF_CMD_GC             force a GC.
 *
 * PROF_CMD_DUMP_HEAP      obtain a heap dump
 *
 * PROF_CMD_ALLOC_SITES    obtain allocation sites
 *
 *                u2        flags 0x0001: incremental vs. complete
 *                                0x0002: sorted by allocation vs. live
 *                                0x0004: whether to force a GC
 *                u4        cutoff ratio (0.0 ~ 1.0)
 *
 * PROF_CMD_HEAP_SUMMARY   obtain heap summary
 *
 * PROF_CMD_DUMP_TRACES    obtain all newly created traces
 *
 * PROF_CMD_CPU_SAMPLES    obtain a PROF_CPU_SAMPLES record
 *
 *                u2        ignored for now
 *                u4        cutoff ratio (0.0 ~ 1.0)
 *
 * PROF_CMD_CONTROL        changing settings
 *
 *                u2        0x0001: alloc traces on
 *                          0x0002: alloc traces off
 *
 *                          0x0003: CPU sampling on
 *                                  id:   thread object id (NULL for all)
 *
 *                          0x0004: CPU sampling off
 *                                  id:   thread object id (NULL for all)
 *
 *                          0x0005: CPU sampling clear
 *
 *                          0x0006: clear alloc sites info
 *
 *                          0x0007: set max stack depth in CPU samples and alloc traces
 *                                  u2:   new depth
 */

#define PROF_CMD_GC                0x01
#define PROF_CMD_DUMP_HEAP         0x02
#define PROF_CMD_ALLOC_SITES       0x03
#define PROF_CMD_HEAP_SUMMARY      0x04
#define PROF_CMD_EXIT              0x05
#define PROF_CMD_DUMP_TRACES       0x06
#define PROF_CMD_CPU_SAMPLES       0x07
#define PROF_CMD_CONTROL           0x08

#define PROF_CMD_SUSPEND_ALL       0x40
#define PROF_CMD_RESUME_ALL        0x41
#define PROF_CMD_GRAPH_INFO        0x42
#define PROF_CMD_DUMP_MONITORS     0x43
#define PROF_CMD_DUMP_ON_EXIT      0x44
#define PROF_CMD_MONITOR_EVENTS    0x45

#define PROF_CMD_SET_TRIGGER       0x50

/* flags used in output of allocation sites */
#define PROF_SITE_DUMP_INCREMENTAL 0x01
#define PROF_SITE_SORT_BY_ALLOC    0x02
#define PROF_SITE_FORCE_GC         0x04

#endif /* _PROF_H */
