/*
 * prof_class.c
 *
 * Copyright (c) ProSyst Software
 * All rights reserved
 *
 * Written by Vladislav Doshev,
 *            Alexi     Jordanov   (alexi91@yahoo.com),
 *            Ivo       Karabashev
 */

#include "jvmpi.h"
#include "prof.h"
#include <string.h>

/**
 * @author  Vladislav Doshev
 * @author  Alexi     Jordanov (alexi91@yahoo.com)
 * @author  Ivo       Karabashev
 * @version 1.00, 02/11/2000
 */

#define PROF_CLASS_TABLE_SIZE 3001

static unsigned int class_serial_number = 1;
prof_hash_t prof_class_table;
prof_class_t *java_lang_object_class = NULL;

static unsigned int hash_class(void *_class)
{
  prof_class_t *class = _class;
  return (unsigned int)class->class_id;
}

static unsigned int size_class(void *_class)
{
  return sizeof(prof_class_t);
}

static int compare_class(void *_class1, void *_class2)
{
  prof_class_t *class1 = _class1;
  prof_class_t *class2 = _class2;

  return (int)((unsigned int)class1->class_id - (unsigned int)class2->class_id);
}

void prof_class_table_init(void)
{
  prof_hash_init(&prof_class_table, PROF_CLASS_TABLE_SIZE, hash_class, size_class, compare_class);
}

prof_class_t * prof_lookup_class_objmap(prof_objmap_t *objmap)
{
  prof_class_t class_tmp;

  class_tmp.class_id = objmap;
  return prof_hash_lookup(&prof_class_table, &class_tmp);
}

prof_class_t * prof_lookup_class(jobjectID class_id)
{
  prof_objmap_t *objmap = prof_fetch_object_info(class_id);

  if (objmap == NULL) return NULL;
  return prof_lookup_class_objmap(objmap);
}

static unsigned char sigToTy(char sig)
{
  switch(sig) {
    case 'L':
    case '[':
      return JVMPI_CLASS;
    case 'Z':
      return JVMPI_BOOLEAN;
    case 'B':
      return JVMPI_BYTE;
    case 'C':
      return JVMPI_CHAR;
    case 'S':
      return JVMPI_SHORT;
    case 'I':
      return JVMPI_INT;
    case 'J':
      return JVMPI_LONG;
    case 'F':
      return JVMPI_FLOAT;
    case 'D':
      return JVMPI_DOUBLE;
    default:
      return 0;
  }
}

void prof_class_load_event(JNIEnv *env_id,
         char *name, char *src_name, unsigned int num_interfaces,
         unsigned int num_statics,   JVMPI_Field *statics,
         unsigned int num_instances, JVMPI_Field *instances,
         unsigned int num_methods,   JVMPI_Method *methods,
         jobjectID class_id, int requested)
{
  prof_class_t *result;
  prof_class_t class_tmp;
  prof_field_t *hstatics;
  prof_field_t *hinstances;
  prof_objmap_t *objmap;
  unsigned int i;
  unsigned int trace_num;

  CALL(RawMonitorEnter)(data_access_lock);

  objmap = prof_fetch_object_info(class_id);

  if (objmap == NULL) {
    if (debug) fprintf(stderr, "PROF ERROR: unable to map JVMPI class ID to prof class ID in class_load \n");
    goto classload_done;
  }

  class_tmp.class_id = objmap;
  if (prof_hash_lookup(&prof_class_table, &class_tmp) != NULL) {
    if (debug) fprintf(stderr, "PROF ERROR: class ID already in use\n");
    goto classload_done;
  }
  result = prof_hash_put(&prof_class_table, &class_tmp);

  result->super = NULL;
  result->loader = NULL;
  result->num_methods = num_methods;
  result->num_interfaces = num_interfaces;
  result->num_statics = num_statics;
  result->num_instances = num_instances;
  result->name = prof_intern_name(name);
  result->src_name = prof_intern_name(src_name);
  result->serial_num = class_serial_number++;

  if (num_methods)
    result->methods = (prof_method_t **)prof_calloc(sizeof(prof_method_t *) * num_methods);
  else
    result->methods = NULL;

  /* intern methods in the class */
  for (i = 0; i < num_methods; i++) {
    result->methods[i] = prof_intern_method(&methods[i], result);
  }

  if (num_statics) {
    hstatics = (prof_field_t *)prof_calloc(num_statics * sizeof(prof_field_t));
    for (i = 0; i < num_statics; i++) {
      hstatics[i].name = prof_intern_name(statics[i].field_name);
      hstatics[i].type = sigToTy(statics[i].field_signature[0]);
    }
  } else {
    hstatics = NULL;
  }
  result->statics = hstatics;

  if (num_instances) {
    hinstances = (prof_field_t *)prof_calloc(num_instances * sizeof(prof_field_t));
    for (i = 0; i < num_instances; i++) {
      hinstances[i].name = prof_intern_name(instances[i].field_name);
      hinstances[i].type = sigToTy(instances[i].field_signature[0]);
    }
  } else {
    hinstances = NULL;
  }
  result->instances = hinstances;

  if (java_lang_object_class == NULL && strcmp(name, "java/lang/Object") == 0) {
    java_lang_object_class = result;
  }

  if (requested) {
    trace_num = 0;
  } else {
    prof_trace_t *htrace = prof_get_trace(env_id, max_trace_depth);
    if (htrace == NULL) {
      if (debug) fprintf(stderr, "PROF ERROR : got a NULL trace in class_load\n");
      goto classload_done;
    }
    trace_num = htrace->serial_num;
  }
  prof_write_header(PROF_LOAD_CLASS, 2 * sizeof(void *) + 8);
  prof_write_u4(result->serial_num);
  prof_write_id(objmap);
  prof_write_u4(trace_num);
  prof_write_id(result->name);
classload_done:
  CALL(RawMonitorExit)(data_access_lock);
}

void prof_class_unload_event(JNIEnv *env_id, jobjectID class_id)
{
  prof_class_t *class;
  unsigned int i;

  CALL(RawMonitorEnter)(data_access_lock);
  class = prof_lookup_class(class_id);
  if (class == NULL) {
    if (debug) fprintf(stderr, "PROF ERROR : unknown class ID in class_unload\n");
  } else {
    prof_write_header(PROF_UNLOAD_CLASS, 4);
    prof_write_u4(class->serial_num);
    /* resolve the methodIDs in the thread local tables, as they may have become invalid now. */
    /* invalidate this entry */
    class->class_id = (prof_objmap_t *) -1;
    /* invalidate all the method entries defined in this class */
    for (i = 0; i < class->num_methods; i++) {
      class->methods[i]->method_id = (jmethodID) -1;
    }
  }
  CALL(RawMonitorExit)(data_access_lock);
}

void prof_superclass_link(jobjectID class_id, jobjectID super_id, jobjectID loader_id)
{
  prof_class_t  *class;
  prof_class_t  *super;
  prof_objmap_t *loader;

  CALL(RawMonitorEnter)(data_access_lock);
  class = prof_lookup_class(class_id);
  if (class == NULL) {
    if (debug) fprintf(stderr, "PROF ERROR: unknown class ID in superclass_link\n");
  } else {
    if (super_id == NULL) {
      super = NULL;
    } else {
      super = prof_lookup_class(super_id);
      if (super == NULL) {
        if (debug) fprintf(stderr, "PROF ERROR: unknown superclass ID in superclass_link\n");
      }
    }
    if (loader_id == NULL) {
      loader = NULL;
    } else {
      loader = prof_fetch_object_info(loader_id);
      if (loader == NULL) {
        if (debug) fprintf(stderr, "PROF ERROR: unknown class loader ID in superclass_link\n");
      }
    }
    class->super = super;
    class->loader = loader;
  }
  CALL(RawMonitorExit)(data_access_lock);
}
