/*
 * prof_cpu.c
 *
 * Copyright (c) ProSyst Software
 * All rights reserved
 *
 * Written by Vladislav Doshev,
 *            Alexi     Jordanov   (alexi91@yahoo.com),
 *            Ivo       Karabashev
 */

#include "jvmpi.h"
#include "prof.h"

/**
 * @author  Vladislav Doshev
 * @author  Alexi     Jordanov (alexi91@yahoo.com)
 * @author  Ivo       Karabashev
 * @version 1.00, 02/11/2000
 */

static JVMPI_RawMonitor prof_cpu_lock;

void prof_cpu_sample_off(prof_objmap_t *thread_id)
{
  cpu_sampling = FALSE;
  if (thread_id != NULL) {
    if (live_thread_list != NULL) {
      live_thread_t *list;
      CALL(RawMonitorEnter)(data_access_lock);
      for(list = live_thread_list; list; list = list->next) {
        if (list->tid == thread_id) {
          list->cpu_sampled = 0;
        }
        if (list->cpu_sampled) {
          cpu_sampling = TRUE;
        }
      }
      CALL(RawMonitorExit)(data_access_lock);
    }
  }
}

/* handler for enabling cpu sampling */
void prof_cpu_sample_on(prof_objmap_t *thread_id)
{
  prof_start_cpu_sampling_thread();
  if (thread_id == NULL) {
    cpu_sampling = TRUE;
  } else {
    /* turn on cpu sampling for the specified thread */
    if (live_thread_list) {
      live_thread_t *list;
      CALL(RawMonitorEnter)(data_access_lock);
      for(list = live_thread_list; list; list = list->next) {
        if (list->tid == thread_id) {
          list->cpu_sampled = 1;
          cpu_sampling = TRUE;
        }
      }
      CALL(RawMonitorExit)(data_access_lock);
    }
  }
  if (prof_cpu_lock) {
    /* notify the CPU sampling thread */
    CALL(RawMonitorEnter)(prof_cpu_lock);
    CALL(RawMonitorNotifyAll)(prof_cpu_lock);
    CALL(RawMonitorExit)(prof_cpu_lock);
  }
}

static void prof_cpu_loop(void *p)
{
  float avg_sample_time = 1;
  jint last_sample_time = 1;
  jint pause_time = 1;
  JNIEnv *my_env;

  (*jvm)->GetEnv(jvm, (void **)&my_env, JNI_VERSION_1_2);

  prof_cpu_lock = CALL(RawMonitorCreate)("_Prof CPU sampling lock");

  CALL(RawMonitorEnter)(prof_cpu_lock);

  while (1) {
    avg_sample_time = last_sample_time * 0.1f + avg_sample_time * 0.9f;

    if (cpu_sampling) {
      /* Adjust the sampling interval according to the cost of each sample. */
      jint avg_time = (jint)avg_sample_time;
      if (avg_time == 0) {
        avg_time = 1;
      }
      if (avg_time > pause_time * 2) {
        pause_time = avg_time;
      }
      if (pause_time > 1 && avg_time < pause_time) {
        pause_time = avg_time;
      }
      CALL(RawMonitorWait)(prof_cpu_lock, (jlong)pause_time);
    } else {
      /* If cpu profiling has been turned off, wait forever on the lock. */
      CALL(RawMonitorWait)(prof_cpu_lock, (jlong)10000000000);
      continue;
    }

    CALL(RawMonitorEnter)(prof_dump_lock);
    CALL(DisableGC)();
    CALL(RawMonitorEnter)(data_access_lock);

    last_sample_time = prof_get_milliticks();

    {
      live_thread_t *list;
      live_thread_t *suspended_list = NULL;
      int n_traces = 0;
      int i;
      JVMPI_CallTrace *traces;

      /* Allocate space for all the traces we might collect, maximum
         value is number of live threads */
      traces = prof_calloc(num_live_threads * sizeof(JVMPI_CallTrace));
      for (i = 0; i < num_live_threads; i++) {
        traces[i].frames = prof_calloc(max_trace_depth * sizeof(JVMPI_CallFrame));
      }

      /* suspend all the runnable threads */
      for (list = live_thread_list; list; list = list->next) {
        if (list->cpu_sampled && list->env != my_env) {
          jint thread_status = CALL(GetThreadStatus)(list->env);
          if ((thread_status & (~JVMPI_THREAD_INTERRUPTED)) == JVMPI_THREAD_RUNNABLE) {
            CALL(SuspendThread)(list->env);
            list->nextSuspended = suspended_list;
            suspended_list = list;
          }
        }
      }

      /* get traces for all the threads that were actually running */
      for (list = suspended_list; list; list = list->nextSuspended) {
        if (CALL(ThreadHasRun)(list->env)) {
          traces[n_traces].env_id = list->env;
          CALL(GetCallTrace)(&(traces[n_traces]), max_trace_depth);
          if (traces[n_traces].num_frames > 0) {
            n_traces++;
          }
        }
      }

      /* resume all the suspended threads */
      for (list = suspended_list; list; list = list->nextSuspended) {
        CALL(ResumeThread)(list->env);
      }

      /* record all the traces */
      if (n_traces > 0) {
        for (i = 0; i < n_traces; i++) {
          JVMPI_CallTrace *jtrace = &(traces[i]);
          prof_trace_t *htrace = prof_intern_jvmpi_trace(jtrace->frames,
              jtrace->num_frames, jtrace->env_id);
          if (htrace == NULL) {
            if (debug) fprintf(stderr, "PROF ERROR: NULL trace in cpu_sample\n");
          } else {
            htrace->num_hits++;
          }
        }
      }

      /* free up the space allocated to collect traces */
      for(i = 0; i < num_live_threads; i++) {
        prof_free(traces[i].frames);
      }
      prof_free(traces);
    }

    last_sample_time = prof_get_milliticks() - last_sample_time;

    CALL(RawMonitorExit)(data_access_lock);
    CALL(EnableGC)();
    CALL(RawMonitorExit)(prof_dump_lock);
  }
  /* never reached */
}

void prof_start_cpu_sampling_thread(void)
{
  static int started = FALSE;
  if (started || cpu_sampling == FALSE) return;
  started = TRUE;
  /* start the cpu sampling thread */
  if (CALL(CreateSystemThread)("PROF CPU profiler", JVMPI_NORMAL_PRIORITY, prof_cpu_loop) < 0) {
    fprintf(stderr, "PROF ERROR: unable to create CPU sampling thread\n");
  }
}
