/*
 * prof_hash.c
 *
 * Copyright (c) ProSyst Software
 * All rights reserved
 *
 * Written by Vladislav Doshev,
 *            Alexi     Jordanov   (alexi91@yahoo.com),
 *            Ivo       Karabashev
 */

#include <string.h>

#include "prof.h"

/**
 * @author  Vladislav Doshev
 * @author  Alexi     Jordanov (alexi91@yahoo.com)
 * @author  Ivo       Karabashev
 * @version 1.00, 02/11/2000
 */

/* initialise a hashtable */
void prof_hash_init(prof_hash_t *table, unsigned int size,
         unsigned int (*hash_f)(void *),
         unsigned int (*size_f)(void *),
         int (*compare_f)(void *, void *))
{
  table->n_entries = 0;
  table->size = size;
  table->entries = prof_calloc(size * sizeof(prof_bucket_t *));
  table->hash_f = hash_f;
  table->size_f = size_f;
  table->compare_f = compare_f;
}

/* iterate through the hash table */
void * prof_hash_iterate(prof_hash_t *table, void * (*f)(void *, void *), void *arg)
{
  unsigned int i;
  for (i = 0; i < table->size; i++) {
    prof_bucket_t *bucket = table->entries[i];
    while (bucket) {
      arg = f(bucket, arg);
      bucket = bucket->next;
    }
  }
  return arg;
}

/* remove an entry from the hash table */
void prof_hash_remove(prof_hash_t *table, int (*f)(void *, void *), void *arg)
{
  unsigned int i;
  for (i = 0; i < table->size; i++) {
    prof_bucket_t **p = &(table->entries[i]);
    prof_bucket_t *bucket;
    while ((bucket = *p)) {
      if (f(bucket, arg)) {
        *p = bucket->next;
        prof_free(bucket);
        table->n_entries--;
        continue;
      }
      p = &(bucket->next);
    }
  }
}

/* remove all entries from the hash table */
void prof_hash_removeall(prof_hash_t *table)
{
  unsigned int i;
  for (i = 0; i < table->size; i++) {
    prof_bucket_t **p = &(table->entries[i]);
    prof_bucket_t *bucket;
    while ((bucket = *p)) {
      *p = bucket->next;
      prof_free(bucket);
      table->n_entries--;
    }
  }
}

/* lookup a hashtable, if present, a ptr to the contents is returned, else a NULL. */
void * prof_hash_lookup(prof_hash_t *table, void *new)
{
  int index = (int)(table->hash_f(new) % table->size);
  prof_bucket_t **p = &(table->entries[index]);

#ifdef _HASH_STAT_
  table->num_calls_hash_f++;
#endif
  while (*p) {
#ifdef _HASH_STAT_
    table->num_calls_compare_f++;
#endif
    if (table->compare_f(new, *p) == 0) {
      return (*p);
    }
    p = &((*p)->next);
  }
  return NULL;
}

/* put an entry into a hashtable
   CAUTION - this should be used only if prof_hash_lookup returns NULL */
void * prof_hash_put(prof_hash_t *table, void *new)
{
	int index;
	int bytes;
	prof_bucket_t *new_bucket;

  if (table->n_entries >= table->size * 0.5) {
		//prof_hash_rehash(table);
	}
  index = (int)(table->hash_f(new) % table->size);
  bytes = table->size_f(new);
  new_bucket = prof_calloc(bytes);

#ifdef _HASH_STAT_
  table->num_calls_hash_f++;
#endif
  memcpy(new_bucket, new, bytes);
  new_bucket->next = table->entries[index];
  table->entries[index] = new_bucket;
  table->n_entries++;
  return new_bucket;
}

/* intern an entry into the hash table */
void * prof_hash_intern(prof_hash_t *table, void *new)
{
  void *ptr = prof_hash_lookup(table, new);
  if (ptr == NULL) {
    ptr = prof_hash_put(table, new);
  }
  return ptr;
}

void prof_hash_rehash(prof_hash_t *table)
{
  unsigned int i;
  unsigned int index;

	unsigned int old_n_entries = table->n_entries;
	unsigned int old_size = table->size;
	prof_bucket_t **old_entries = table->entries;

  table->n_entries = 0;
  table->size = table->size * 2 + 1;
  table->entries = prof_calloc(table->size * sizeof(prof_bucket_t *));

  for (i = 0; i < old_size; i++)	{
    prof_bucket_t **p = &(old_entries[i]);
	  prof_bucket_t *bucket;
	  while ((bucket = *p)) {
      *p = bucket->next;
			index = (int)(table->hash_f(bucket) % table->size);
			bucket->next = table->entries[index];
			table->entries[index] = bucket;
			table->n_entries++;
		}
  }
	prof_free(old_entries);
}

void prof_print_tbl_hash_stats(prof_hash_t *table, char *tbl_name)
{
  unsigned int i;
  unsigned int empties = 0;
  unsigned int max_elems = 0;

  fprintf(stderr, "\nHash Table Statistics for '%s'\n", tbl_name);

  if (table->n_entries == 0) {
    fprintf(stderr, "TABLE IS EMPTY!!\n");
  } else {
#ifdef _HASH_STAT_
		unsigned int mem_size = 0;
#endif
    for (i = 0; i < table->size; i++) {
      prof_bucket_t *bucket;
      unsigned int elems = 0;

      for (bucket = table->entries[i]; bucket; bucket = bucket->next) {
				elems++;
#ifdef _HASH_STAT_
				mem_size += table->size_f(bucket);
#endif
			}
      if (elems == 0) empties++;
      else if (elems > max_elems) max_elems = elems;
    }

    fprintf(stderr, "tbl-size=%u   #-elems=%u   empty-slots=%.2f%%   avg-#-elems=%.2f   max-#-elems=%u\n",
      table->size,
      table->n_entries,
      ((float)empties / table->size) * 100.0f,
      (float)table->n_entries / (table->size - empties),
      max_elems);
#ifdef _HASH_STAT_
    fprintf(stderr, "hash-calls=%u   compare-calls=%u   mem-size=%u\n",
      table->num_calls_hash_f, table->num_calls_compare_f, mem_size);
#endif
  }
  fflush(stderr);
}
