/*
 * prof_heapdump.c
 *
 * Copyright (c) ProSyst Software
 * All rights reserved
 *
 * Written by Vladislav Doshev,
 *            Alexi     Jordanov   (alexi91@yahoo.com),
 *            Ivo       Karabashev
 */

#include <string.h>

#include "jvmpi.h"
#include "prof.h"

/**
 * @author  Vladislav Doshev
 * @author  Alexi     Jordanov (alexi91@yahoo.com)
 * @author  Ivo       Karabashev
 * @version 1.00, 02/11/2000
 */

void prof_get_heap_dump(void)
{
  if (CALL(RequestEvent)(JVMPI_EVENT_HEAP_DUMP, NULL) != JVMPI_SUCCESS) {
    if (debug) fprintf(stderr, "PROF ERROR: heap dump from JVM did not succeed\n");
  }
}

/* this function is used to get the size of the heap dump
   record output by prof if the output format is binary */
static int prof_get_dump_size(char *begin_roots, char *end_objects)
{
  int size = 0;
  prof_dump_seek(begin_roots);
  while (prof_dump_cur() < end_objects) {
    unsigned char tag;
    int in;
    int out;
    int sizeof_u1 = sizeof(unsigned char);
    int sizeof_u2 = sizeof(unsigned short);
    int sizeof_u4 = sizeof(unsigned int);
    int sizeof_id = sizeof(void *);
    tag = prof_dump_read_u1();
    size += sizeof_u1;
    switch (tag) {
      case PROF_GC_ROOT_MONITOR_USED:
      case PROF_GC_ROOT_STICKY_CLASS:
      case PROF_GC_ROOT_UNKNOWN: {
        in = sizeof_id;
        out = sizeof_id;
        size += out;
        prof_dump_seek(prof_dump_cur() + in);
        break;
      }
      case PROF_GC_ROOT_JNI_GLOBAL: {
        in = 2 * sizeof_id;
        out = 2 * sizeof_id;
        size += out;
        prof_dump_seek(prof_dump_cur() + in);
        break;
      }
      case PROF_GC_ROOT_THREAD_BLOCK:
      case PROF_GC_ROOT_NATIVE_STACK: {
        in = 2 * sizeof_id;
        out = sizeof_id + sizeof_u4;
        size += out;
        prof_dump_seek(prof_dump_cur() + in);
        break;
      }
      case PROF_GC_ROOT_JAVA_FRAME:
      case PROF_GC_ROOT_JNI_LOCAL: {
        in = 2 * sizeof_id + sizeof_u4;
        out = sizeof_id + 2 * sizeof_u4;
        size += out;
        prof_dump_seek(prof_dump_cur() + in);
        break;
      }
      case PROF_GC_CLASS_DUMP: {
        int n;
        int i;

        jobjectID class_id = prof_dump_read_id();
        prof_class_t *class = prof_lookup_class(class_id);
        if (class == NULL) {
          return -1;
        }

        in = 6 * sizeof_id + sizeof_u4;
        out = 7 * sizeof_id + 2 * sizeof_u4;
        size += out;
        prof_dump_seek(prof_dump_cur() + in);

        in = class->num_interfaces * sizeof_id;
        prof_dump_seek(prof_dump_cur() + in);

        n = prof_dump_read_u2();
        size += sizeof_u2;
        in = n * (sizeof_u2 + sizeof_u1 + sizeof_id);
        out = in;
        size += out;
        prof_dump_seek(prof_dump_cur() + in);

        n = class->num_statics;
        size += sizeof_u2;
        for (i = 0; i < n; i++) {
          jint ty = class->statics[i].type;
          size += sizeof_id + sizeof_u1;
          if (ty == JVMPI_LONG || ty == JVMPI_DOUBLE) {
            in = sizeof(jvalue);
            out = in;
            size += out;
            prof_dump_seek(prof_dump_cur() + in);
          } else if (ty == JVMPI_CLASS) {
            in = sizeof_id;
            out = sizeof_id;
            size += out;
            prof_dump_seek(prof_dump_cur() + in);
          } else if (ty == JVMPI_INT || ty == JVMPI_FLOAT) {
            in = sizeof_u4;
            out = sizeof_u4;
            size += out;
            prof_dump_seek(prof_dump_cur() + in);
          } else if (ty == JVMPI_CHAR || ty == JVMPI_SHORT) {
            in = sizeof_u2;
            out = sizeof_u2;
            size += out;
            prof_dump_seek(prof_dump_cur() + in);
          } else if (ty == JVMPI_BYTE || ty == JVMPI_BOOLEAN) {
            in = sizeof_u1;
            out = sizeof_u1;
            size += out;
            prof_dump_seek(prof_dump_cur() + in);
          }
        }

        size += sizeof_u2 + class->num_instances * (sizeof_id + sizeof_u1);
        break;
      }
      case PROF_GC_INSTANCE_DUMP: {
        unsigned int valbytes;

        in = 2 * sizeof_id;
        out = 2 * sizeof_id + sizeof_u4;
        size += out;
        prof_dump_seek(prof_dump_cur() + in);

        valbytes = prof_dump_read_u4();
        size += sizeof_u4;

        size += valbytes;
        prof_dump_seek(prof_dump_cur() + valbytes);
        break;
      }
      case PROF_GC_OBJ_ARRAY_DUMP: {
        unsigned int num_elements;

        in = sizeof_id;
        out = sizeof_id + sizeof_u4;
        size += out;
        prof_dump_seek(prof_dump_cur() + in);

        num_elements = prof_dump_read_u4();
        size += sizeof_u4;

        in = sizeof_id + num_elements * sizeof_id;
        out = in;
        size += out;
        prof_dump_seek(prof_dump_cur() + in);
        break;
      }
      case PROF_GC_PRIM_ARRAY_DUMP: {
        unsigned int num_elements;
        unsigned char ty;

        in = sizeof_id;
        out = sizeof_id + sizeof_u4;
        size += out;
        prof_dump_seek(prof_dump_cur() + in);

        num_elements = prof_dump_read_u4();
        size += sizeof_u4;
        ty = prof_dump_read_u1();
        size += sizeof_u1;

        switch (ty) {
          case JVMPI_BYTE:
          case JVMPI_BOOLEAN:
            in = num_elements * sizeof_u1;
            /* sending values of primitive array elements is removed
            out = in;
            size += out; */
            prof_dump_seek(prof_dump_cur() + in);
            break;
          case JVMPI_SHORT:
          case JVMPI_CHAR:
            in = num_elements * sizeof_u2;
            /* sending values of primitive array elements is removed
            out = in;
            size += out; */
            prof_dump_seek(prof_dump_cur() + in);
            break;
          case JVMPI_FLOAT:
          case JVMPI_INT:
            in = num_elements * sizeof_u4;
            /* sending values of primitive array elements is removed
            out = in;
            size += out; */
            prof_dump_seek(prof_dump_cur() + in);
            break;
          case JVMPI_DOUBLE:
          case JVMPI_LONG:
            in = num_elements * 2 * sizeof_u4;
            /* sending values of primitive array elements is removed
            out = in;
            size += out; */
            prof_dump_seek(prof_dump_cur() + in);
            break;
        }
        break;
      }
    }
  }
  return size;
}

/* this function processes the heap dump buffer
   and outputs the second part of the heap dump record */
static void prof_process_dump_buffer(char *begin_roots, char *end_objects)
{
  unsigned int i;

  prof_dump_seek(begin_roots);
  while (prof_dump_cur() < end_objects) {
    unsigned char tag;
    tag = prof_dump_read_u1();
    prof_write_u1(tag);
    switch (tag) {
      case PROF_GC_ROOT_UNKNOWN: {
        void *obj_id = prof_dump_read_id();
        prof_objmap_t *objmap = prof_fetch_object_info(obj_id);
        prof_write_id(objmap);
        break;
      }
      case PROF_GC_ROOT_JNI_GLOBAL: {
        void *obj_id;
        void *ref_id;
        prof_globalref_t *gref;

        prof_dump_read_id(); /* skip object id, we can get it from gref */
        ref_id = prof_dump_read_id();
        gref = prof_globalref_find(ref_id);

        if (gref == NULL) {
          obj_id = NULL;
        } else {
          obj_id = gref->obj_id; /* already objmapped */
        }

        prof_write_id(obj_id);
        prof_write_id(ref_id);
        break;
      }
      case PROF_GC_ROOT_NATIVE_STACK: {
        void *h = prof_dump_read_id();
        prof_objmap_t *objmap = prof_fetch_object_info(h);
        void *env = prof_dump_read_id();
        prof_thread_t *thread = prof_lookup_thread(env);
        unsigned int thread_serial_num;

        if (thread == NULL) {
          thread_serial_num = 0;
        } else {
          thread_serial_num = thread->serial_num;
        }

        prof_write_id(objmap);
        prof_write_u4(thread_serial_num);
        break;
      }
      case PROF_GC_ROOT_STICKY_CLASS: {
        void *class_id = prof_dump_read_id();
        prof_objmap_t *objmap = prof_fetch_object_info(class_id);
        prof_write_id(objmap);
        break;
      }
      case PROF_GC_ROOT_THREAD_BLOCK: {
        void *h = prof_dump_read_id();
        prof_objmap_t *objmap = prof_fetch_object_info(h);
        void *env = prof_dump_read_id();
        prof_thread_t *thread = prof_lookup_thread(env);
        unsigned int thread_serial_num = (thread == NULL) ? 0 : thread->serial_num;
        prof_write_id(objmap);
        prof_write_u4(thread_serial_num);
        break;
      }
      case PROF_GC_ROOT_MONITOR_USED: {
        void *h = prof_dump_read_id();
        prof_objmap_t *objmap = prof_fetch_object_info(h);
        prof_write_id(objmap);
        break;
      }
      case PROF_GC_ROOT_JNI_LOCAL: {
        void *h = prof_dump_read_id();
        prof_objmap_t *objmap = prof_fetch_object_info(h);
        void *env = prof_dump_read_id();
        int depth = prof_dump_read_u4();
        prof_thread_t *thread = prof_lookup_thread(env);
        unsigned int thread_serial_num = (thread == NULL) ? 0 : thread->serial_num;
        prof_write_id(objmap);
        prof_write_u4(thread_serial_num);
        prof_write_u4(depth);
        break;
      }
      case PROF_GC_ROOT_JAVA_FRAME: {
        void *h = prof_dump_read_id();
        prof_objmap_t *objmap = prof_fetch_object_info(h);
        void *env = prof_dump_read_id();
        int depth = prof_dump_read_u4();
        prof_thread_t *thread = prof_lookup_thread(env);
        unsigned int thread_serial_num = (thread == NULL) ? 0 : thread->serial_num;
        prof_write_id(objmap);
        prof_write_u4(thread_serial_num);
        prof_write_u4(depth);
        break;
      }
      case PROF_GC_CLASS_DUMP: {
        void *class_id;
        prof_objmap_t *class_map;
        prof_objmap_t *super_map;
        prof_objmap_t *loader_map;
        prof_objmap_t *signers_map;
        prof_objmap_t *domain_map;
        unsigned int trace_serial_num;
        prof_class_t *class;
        char *class_name;
        void *resv1;
        void *resv2;
        unsigned int size;
        unsigned short n;
        unsigned int i;

        class_id = prof_dump_read_id();
        class_map = prof_fetch_object_info(class_id);
        class = prof_lookup_class_objmap(class_map);
        if (class == NULL) {
          if (debug) fprintf(stderr, "PROF ERROR: unknown class_id in heap dump %p %p\n",
              class_id, class_map);
          return;
        } else if (class->serial_num == 0) {
          if (debug) fprintf(stderr, "PROF ERROR: no class_load event for class_id in heap dump %p %p\n",
              class_id, class_map);
          return;
        }

        trace_serial_num = class_map->site->trace_serial_num;
        class_name = class->name->name;
        super_map = prof_fetch_object_info(prof_dump_read_id());
        loader_map = prof_fetch_object_info(prof_dump_read_id());
        signers_map = prof_fetch_object_info(prof_dump_read_id());
        domain_map = prof_fetch_object_info(prof_dump_read_id());
        resv1 = prof_dump_read_id();  /* reserved */
        resv2 = prof_dump_read_id();  /* reserved */
        size = prof_dump_read_u4();

        prof_write_id(class_map);
        prof_write_u4(trace_serial_num);
        prof_write_id(super_map);
        prof_write_id(loader_map);
        prof_write_id(signers_map);
        prof_write_id(domain_map);
        prof_write_id(resv1);
        prof_write_id(resv2);
        prof_write_u4(size);

        /* skip interfaces */
        for (i = 0; i < class->num_interfaces; i++) {
          prof_dump_read_id();
        }

        n = prof_dump_read_u2();
        prof_write_u2(n);
        for (i = 0; i < n; i++) {
          unsigned short index;
          unsigned char ty;
          void *p;

          index = prof_dump_read_u2();
          ty = prof_dump_read_u1();
          p = prof_fetch_object_info(prof_dump_read_id());
          prof_write_u2(index);
          prof_write_u1(ty);
          prof_write_id(p);
        }
        n = (unsigned short)class->num_statics;
        prof_write_u2(n);
        for (i = 0; i < n; i++) {
          jint ty = class->statics[i].type;
          prof_name_t *name =  class->statics[i].name;
          prof_write_id(name);
          prof_write_u1((unsigned char)ty);
          if (ty == JVMPI_LONG || ty == JVMPI_DOUBLE) {
            jvalue t;
            prof_dump_read(&t, sizeof(jvalue));
            prof_write_raw(&t, sizeof(jvalue));
          } else if (ty == JVMPI_CLASS) {
            void *p = prof_fetch_object_info(prof_dump_read_id());
            prof_write_id(p);
          } else if (ty == JVMPI_INT || ty == JVMPI_FLOAT) {
            unsigned int u4 = prof_dump_read_u4();
            prof_write_u4(u4);
          } else if (ty == JVMPI_CHAR || ty == JVMPI_SHORT) {
            unsigned short u2 = prof_dump_read_u2();
            prof_write_u2(u2);
          } else if (ty == JVMPI_BYTE || ty == JVMPI_BOOLEAN) {
            unsigned char u1 = prof_dump_read_u1();
            prof_write_u1(u1);
          }
        }
        n = (unsigned short)class->num_instances;
        prof_write_u2(n);
        for (i = 0; i < n; i++) {
          jint ty = class->instances[i].type;
          prof_name_t *name = class->instances[i].name;
          prof_write_id(name);
          prof_write_u1((unsigned char)ty);
        }
        break;
      }
      case PROF_GC_INSTANCE_DUMP: {
        void *obj_id = prof_dump_read_id();
        prof_objmap_t *objmap = prof_fetch_object_info(obj_id);
        void *class_id = prof_dump_read_id();
        prof_objmap_t *classmap = prof_fetch_object_info(class_id);
        unsigned int valbytes = prof_dump_read_u4();
        prof_class_t *class;
        char *class_name;
        unsigned int trace_serial_num = 0;

        if (objmap != NULL) {
          trace_serial_num = objmap->site->trace_serial_num;
        }

        class = prof_lookup_class_objmap(classmap);
        class_name = (class == NULL || class->name == NULL) ? "<Unknown>" : class->name->name;

        prof_write_id(objmap);
        prof_write_u4(trace_serial_num);
        prof_write_id(classmap);
        prof_write_u4(valbytes);

        while (class != NULL) {
          int i;
          int n_inst = class->num_instances;
          unsigned int val_u4;
          unsigned short val_u2;
          unsigned char val_u1;
          jvalue val_j;
          void *val_id;
          jint ty;

          for (i = 0; i < n_inst; i++) {
            ty = class->instances[i].type;
            if (ty == JVMPI_LONG || ty == JVMPI_DOUBLE) {
              prof_dump_read(&val_j, sizeof(jvalue));
              prof_write_raw(&val_j, sizeof(jvalue));
            } else if (ty == JVMPI_CLASS) {
              val_id = prof_fetch_object_info(prof_dump_read_id());
              prof_write_id(val_id);
            } else if (ty == JVMPI_INT || ty == JVMPI_FLOAT) {
              val_u4 = prof_dump_read_u4();
              prof_write_u4(val_u4);
            } else if (ty == JVMPI_CHAR || ty == JVMPI_SHORT) {
              val_u2 = prof_dump_read_u2();
              prof_write_u2(val_u2);
            } else if (ty == JVMPI_BYTE || ty == JVMPI_BOOLEAN) {
              val_u1 = prof_dump_read_u1();
              prof_write_u1(val_u1);
            }
          }
          if (class->super == NULL && class != java_lang_object_class) {
            CALL(RequestEvent)(JVMPI_EVENT_OBJECT_DUMP, class->class_id->obj_id);
          }
          class = class->super;
        }
        break;
      }
      case PROF_GC_OBJ_ARRAY_DUMP: {
        void *obj_id = prof_dump_read_id();
        prof_objmap_t *objmap = prof_fetch_object_info(obj_id);
        unsigned int num_elements = prof_dump_read_u4();
        void *class_id = prof_dump_read_id();
        prof_objmap_t *classmap = prof_fetch_object_info(class_id);
        prof_class_t *class;
        unsigned int trace_serial_num = 0;
        char *class_name;

        if (objmap != NULL) {
          trace_serial_num = objmap->site->trace_serial_num;
        }
        class = prof_lookup_class_objmap(classmap);
        class_name = (class == NULL || class->name == NULL) ? "<Unknown>" : class->name->name;

        prof_write_id(objmap);
        prof_write_u4(trace_serial_num);
        prof_write_u4(num_elements);
        prof_write_id(classmap);
        for (i = 0; i < num_elements; i++) {
          void *p = prof_fetch_object_info(prof_dump_read_id());
          prof_write_id(p);
        }
        break;
      }
      case PROF_GC_PRIM_ARRAY_DUMP: {
        void *obj_id = prof_dump_read_id();
        prof_objmap_t *objmap = prof_fetch_object_info(obj_id);
        unsigned int num_elements = prof_dump_read_u4();
        unsigned char ty = prof_dump_read_u1();
        unsigned int trace_serial_num = 0;

        if (objmap != NULL) {
          trace_serial_num = objmap->site->trace_serial_num;
        }
        prof_write_id(objmap);
        prof_write_u4(trace_serial_num);
        prof_write_u4(num_elements);
        prof_write_u1(ty);

        switch (ty) {
          case JVMPI_BOOLEAN:
            for (i = 0; i < num_elements; i++) {
              /* unsigned char u1 = */ prof_dump_read_u1();
              /* sending values of primitive array elements is removed */
            }
            break;
          case JVMPI_BYTE:
            for (i = 0; i < num_elements; i++) {
              /* unsigned char u1 = */ prof_dump_read_u1();
              /* sending values of primitive array elements is removed */
            }
            break;
          case JVMPI_CHAR:
            for (i = 0; i < num_elements; i++) {
              /* unsigned short u2 = */ prof_dump_read_u2();
              /* sending values of primitive array elements is removed */
            }
            break;
          case JVMPI_SHORT:
            for (i = 0; i < num_elements; i++) {
              /* unsigned short u2 = */ prof_dump_read_u2();
              /* sending values of primitive array elements is removed */
            }
            break;
          case JVMPI_INT:
            for (i = 0; i < num_elements; i++) {
              /* unsigned int u4 = */ prof_dump_read_u4();
              /* sending values of primitive array elements is removed */
            }
            break;
          case JVMPI_LONG:
            for (i = 0; i < num_elements; i++) {
              /* unsigned int u41 = */ prof_dump_read_u4();
              /* unsigned int u42 = */ prof_dump_read_u4();
              /* sending values of primitive array elements is removed */
            }
            break;
          case JVMPI_FLOAT:
            for (i = 0; i < num_elements; i++) {
              /* unsigned int u4 = */ prof_dump_read_u4();
              /* sending values of primitive array elements is removed */
            }
            break;
          case JVMPI_DOUBLE:
            for (i = 0; i < num_elements; i++) {
              /* unsigned int u41 = */ prof_dump_read_u4();
              /* unsigned int u42 = */ prof_dump_read_u4();
              /* sending values of primitive array elements is removed */
            }
            break;
        }
        break;
      }
    }
  }
}

void prof_heap_dump_event(char *begin_roots, char *end_objects,
         int num_traces, JVMPI_CallTrace *traces)
{
  int i;
  int dump_size;
  prof_trace_t **htraces = (prof_trace_t **) prof_calloc(sizeof(prof_trace_t *) * num_traces);

  CALL(RawMonitorEnter)(data_access_lock);

  /* intern the jvmpi traces */
  for (i = 0; i < num_traces; i++) {
    JVMPI_CallTrace *jtrace = &(traces[i]);
    htraces[i] = prof_intern_jvmpi_trace(jtrace->frames, jtrace->num_frames, jtrace->env_id);
    if (htraces[i] == NULL) {
      if (debug) fprintf(stderr, "PROF ERROR: got a NULL trace in heap_dump\n");
      goto heap_dump_done;
    }
  }

  /* First write all trace we might refer to. */
  prof_output_unmarked_traces();

  /* get dump size, we need the dump size only for the binary
   * format, but we do it always to check for any errors */
  dump_size = prof_get_dump_size(begin_roots, end_objects);
  if (dump_size < 0) {
    if (debug) fprintf(stderr, "PROF ERROR: heap dump size < 0\n");
    goto heap_dump_done;
  }

  /* print header */
  {
    int threads_size = num_traces * (1 + sizeof(void *) + 8);
    /* Take into account the size for PROF_GC_ROOT_THREAD_OBJ, in addition to other roots and objects. */
    prof_write_header(PROF_HEAP_DUMP, dump_size + threads_size);
  }

  /* Process the root threads first */
  for (i = 0; i < num_traces; i++) {
    prof_thread_t *thread = prof_lookup_thread(traces[i].env_id);
    prof_objmap_t *thread_id;
    unsigned int thread_serial_num;

    if (thread == NULL) {
      thread_id = NULL;
      thread_serial_num = 0;
    } else {
      thread_id = thread->thread_id;
      thread_serial_num = thread->serial_num;
    }

    prof_write_u1(PROF_GC_ROOT_THREAD_OBJ);
    prof_write_id(thread_id);
    prof_write_u4(thread_serial_num);
    prof_write_u4(htraces[i]->serial_num);
  }
  prof_free(htraces);

  /* process the heap dump buffer */
  prof_process_dump_buffer(begin_roots, end_objects);

heap_dump_done:
  CALL(RawMonitorExit)(data_access_lock);
}

void prof_object_dump_event(char *data)
{
  char *ptr_bak = prof_dump_cur();
  int kind;
  prof_dump_seek(data);
  kind = prof_dump_read_u1();
  if (kind == JVMPI_GC_CLASS_DUMP) {
    jobjectID self = prof_dump_read_id();
    jobjectID super = prof_dump_read_id();
    jobjectID loader = prof_dump_read_id();
    prof_superclass_link(self, super, loader);
  }
  prof_dump_seek(ptr_bak);
}
