/*
 * prof_listener.c
 *
 * Copyright (c) ProSyst Software
 * All rights reserved
 *
 * Written by Vladislav Doshev,
 *            Alexi     Jordanov   (alexi91@yahoo.com),
 *            Ivo       Karabashev
 */

#include "prof.h"

/**
 * @author  Vladislav Doshev
 * @author  Alexi     Jordanov (alexi91@yahoo.com)
 * @author  Ivo       Karabashev
 * @version 1.00, 02/11/2000
 */

static jint last_event_time;
static jclass clsSystem;
static jmethodID midSystemGC;
static JNIEnv *env;

static jint recv_fully(int f, char *buf, int len)
{
  int nbytes = 0;
  while (nbytes < len) {
    int res = recv(f, buf + nbytes, len - nbytes, 0);
    if (res < 0) {
      if (prof_is_on) {
        prof_is_on = FALSE;
        fprintf(stderr, "PROF ERROR: failed to read cmd from socket\n");
      }
      CALL(ProfilerExit)((jint)1);
    }
    nbytes += res;
  }
  return nbytes;
}

static unsigned char recv_u1(void)
{
  unsigned char c;
  recv_fully(prof_fd, (char *)&c, sizeof(unsigned char));
  return c;
}

static unsigned short recv_u2(void)
{
  unsigned short s;
  recv_fully(prof_fd, (char *)&s, sizeof(unsigned short));
  return ntohs(s);
}

static unsigned int recv_u4(void)
{
  unsigned int i;
  recv_fully(prof_fd, (char *)&i, sizeof(unsigned int));
  return ntohl(i);
}

static void * recv_id(void)
{
  void *result;
  recv_fully(prof_fd, (char *)&result, sizeof(void *));
  return result;
}

/* the callback thread */
static void prof_callback(void *p)
{
  while (prof_is_on) {
    unsigned char tag = recv_u1();
    /* jint seq_num = */ recv_u4();
    /* jint length = */ recv_u4();

    switch (tag) {
      case PROF_CMD_GC: {

        /* CALL(RunGC);    /* through JVMPI */
        if (env == NULL) { /* through JNI */
          int res = (*jvm)->GetEnv(jvm, (void **)&env, JNI_VERSION_1_1);
          if (res < 0) {
            fprintf(stderr, "JNI Error\n");
            break;
          }
        }
        if (clsSystem == NULL) {
          clsSystem = (*env)->FindClass(env, "java/lang/System");
          if (clsSystem == NULL) {
            fprintf(stderr, "Can't find java.lang.System class\n");
            break;
          }
        }
        if (midSystemGC == NULL) {
          midSystemGC = (*env)->GetStaticMethodID(env, clsSystem, "gc", "()V");
          if (midSystemGC == NULL) {
            fprintf(stderr, "Can't find System.gc method\n");
            break;
          }
        }
        (*env)->CallStaticVoidMethod(env, clsSystem, midSystemGC);
        break;
      }
      case PROF_CMD_DUMP_HEAP: {
        prof_get_heap_dump();
        break;
      }
      case PROF_CMD_ALLOC_SITES: {
        unsigned short flags = recv_u2();
        unsigned int i_tmp = recv_u4();
        float ratio = *(float *)(&i_tmp);
        CALL(RawMonitorEnter)(data_access_lock);
        prof_output_sites(flags, ratio);
        CALL(RawMonitorExit)(data_access_lock);
        break;
      }
      case PROF_CMD_HEAP_SUMMARY: {
        CALL(RawMonitorEnter)(data_access_lock);
        prof_write_header(PROF_HEAP_SUMMARY, 24);
        prof_write_u4(total_live_bytes);
        prof_write_u4(total_live_instances);
        prof_write_u4(jlong_high(total_alloced_bytes));
        prof_write_u4(jlong_low(total_alloced_bytes));
        prof_write_u4(jlong_high(total_alloced_instances));
        prof_write_u4(jlong_low(total_alloced_instances));
        CALL(RawMonitorExit)(data_access_lock);
        break;
      }
      case PROF_CMD_EXIT: {
        prof_is_on = FALSE;
        if (debug) fprintf(stderr, "PROF: received exit event, exiting ...\n");
        CALL(ProfilerExit)((jint)0);
      }
      case PROF_CMD_DUMP_TRACES: {
        CALL(RawMonitorEnter)(data_access_lock);
        prof_output_unmarked_traces();
        CALL(RawMonitorExit)(data_access_lock);
        break;
      }
      case PROF_CMD_CPU_SAMPLES: {
        unsigned short flags = recv_u2();
        unsigned int i_tmp = recv_u4();
        float ratio = *(float *)(&i_tmp);
        send_data_on_exit = FALSE;
        CALL(RawMonitorEnter)(data_access_lock);
        prof_output_trace_cost(ratio);
        CALL(RawMonitorExit)(data_access_lock);
        break;
      }
      case PROF_CMD_CONTROL: {
        unsigned short cmd = recv_u2();
        if (cmd == 0x0001) {
          CALL(EnableEvent)(JVMPI_EVENT_OBJECT_ALLOC, NULL);
          CALL(EnableEvent)(JVMPI_EVENT_OBJECT_MOVE, NULL);
          CALL(EnableEvent)(JVMPI_EVENT_OBJECT_FREE, NULL);
        } else if (cmd == 0x0002) {
          CALL(DisableEvent)(JVMPI_EVENT_OBJECT_ALLOC, NULL);
          CALL(DisableEvent)(JVMPI_EVENT_OBJECT_MOVE, NULL);
          CALL(DisableEvent)(JVMPI_EVENT_OBJECT_FREE, NULL);
        } else if (cmd == 0x0003) {
          prof_objmap_t *thread_id = recv_id();
          prof_cpu_sample_on(thread_id);
        } else if (cmd == 0x0004) {
          prof_objmap_t *thread_id = recv_id();
          prof_cpu_sample_off(thread_id);
        } else if (cmd == 0x0005) {
          send_data_on_exit = TRUE;
          CALL(RawMonitorEnter)(data_access_lock);
          prof_clear_trace_cost();
          CALL(RawMonitorExit)(data_access_lock);
        } else if (cmd == 0x0006) {
          CALL(RawMonitorEnter)(data_access_lock);
          prof_clear_site_table();
          CALL(RawMonitorExit)(data_access_lock);
        } else if (cmd == 0x0007) {
          max_trace_depth = recv_u2();
        }
        break;
      }
      case PROF_CMD_SUSPEND_ALL: {
        if (nonsys_thread_list != NULL) {
          nonsys_thread_t *list;
          CALL(RawMonitorEnter)(data_access_lock);
          for(list = nonsys_thread_list; list; list = list->next) {
            CALL(SuspendThread(list->env));
          }
          CALL(RawMonitorExit)(data_access_lock);
        }
        break;
      }
      case PROF_CMD_RESUME_ALL: {
        if (nonsys_thread_list != NULL) {
          nonsys_thread_t *list;
          CALL(RawMonitorEnter)(data_access_lock);
          for(list = nonsys_thread_list; list; list = list->next) {
            CALL(ResumeThread(list->env));
          }
          CALL(RawMonitorExit)(data_access_lock);
        }
        break;
      }
      case PROF_CMD_GRAPH_INFO: {
        nonsys_thread_t *list;
        jint num_active_threads = 0;
        jint event_duration;

        event_duration = last_event_time;
        last_event_time = prof_get_milliticks();
        event_duration = last_event_time - event_duration;
        for (list = nonsys_thread_list; list; list = list->next) {
          if (CALL(ThreadHasRun)(list->env)) num_active_threads++;
        }

        CALL(RawMonitorEnter)(data_access_lock);
        prof_write_header(PROF_GRAPH_INFO, 44);
        prof_write_u4(jlong_low(event_duration));
        prof_write_u4(total_live_bytes);
        prof_write_u4(total_live_instances);
        prof_write_u4(jlong_high(total_alloced_bytes));
        prof_write_u4(jlong_low(total_alloced_bytes));
        prof_write_u4(jlong_high(total_alloced_instances));
        prof_write_u4(jlong_low(total_alloced_instances));
        prof_write_u4(num_live_threads);
        prof_write_u4(num_active_threads);
        prof_write_u4(prof_class_table.n_entries);
        prof_write_u4(jlong_low(gc_active_time));
        CALL(RawMonitorExit)(data_access_lock);

        gc_active_time = 0;
        break;
      }
      case PROF_CMD_DUMP_MONITORS: {
        prof_dump_monitors();
        break;
      }
      case PROF_CMD_SET_TRIGGER: {
        unsigned char type, action;
        unsigned short len;
        char *cname_buf, *mname_buf;
        if (!triggers) {
          triggers = TRUE;
          CALL(EnableEvent)(JVMPI_EVENT_METHOD_ENTRY, NULL);
          CALL(EnableEvent)(JVMPI_EVENT_METHOD_EXIT,  NULL);
        }
        type = recv_u1();
        action = recv_u1();
        len = recv_u2();
        cname_buf = prof_calloc(len + 1);
        recv_fully(prof_fd, cname_buf, len);
        len = recv_u2();
        if (len == 0) {
          mname_buf = NULL;
        } else {
          mname_buf = prof_calloc(len + 1);
          recv_fully(prof_fd, mname_buf, len);
        }
        CALL(RawMonitorEnter)(data_access_lock);
        if (type == TRIGGER_METHOD_ENTRY) prof_put_trigger_entry(action, cname_buf, mname_buf);
        else                              prof_put_trigger_exit (action, cname_buf, mname_buf);
        CALL(RawMonitorExit)(data_access_lock);
        break;
      }
      case PROF_CMD_DUMP_ON_EXIT: {
        send_data_on_exit = TRUE;
        break;
      }
      case PROF_CMD_MONITOR_EVENTS: {
        recv_u2();
        prof_enable_monitors();
        break;
      }
      default: {
        if (prof_is_on) {
          prof_is_on = FALSE;
          fprintf(stderr, "PROF ERROR : failed to recognize cmd %d, exiting..\n", (int)tag);
        }
        CALL(ProfilerExit)((jint)1);
      }
    }
    CALL(RawMonitorEnter)(data_access_lock);
    prof_flush();
    CALL(RawMonitorExit)(data_access_lock);
  }
}

void prof_start_listener_thread(void)
{
  if (CALL(CreateSystemThread)("Prof listener", JVMPI_MAXIMUM_PRIORITY, prof_callback) == JNI_ERR) {
    fprintf(stderr, "PROF ERROR: unable to create listener thread\n");
  }
  last_event_time = prof_get_milliticks();
  gc_active_time = 0;
}
