/*
 * prof_monitor.c
 *
 * Copyright (c) ProSyst Software
 * All rights reserved
 *
 * Written by Vladislav Doshev,
 *            Alexi     Jordanov   (alexi91@yahoo.com),
 *            Ivo       Karabashev
 */

#include "prof.h"

/**
 * @author  Vladislav Doshev
 * @author  Alexi     Jordanov (alexi91@yahoo.com)
 * @author  Ivo       Karabashev
 * @version 1.00, 02/11/2000
 */

static JVMPI_RawMonitor prof_monitor_lock;

void write_monitor_event(unsigned char event_type, void *mon_info, JNIEnv *env_id, jlong cur_time) {
  prof_write_header(event_type, 25);
  prof_write_object_name(mon_info);
  prof_write_u4(jlong_high(cur_time));
  prof_write_u4(jlong_low(cur_time));
  prof_write_thread_info(env_id);
}

void prof_raw_monitor_event(JVMPI_Event *event, char *name, JVMPI_RawMonitor mid)
{
  jlong cur_time;
  prof_name_t *hname;
  JNIEnv *env_id;

  cur_time = CALL(GetCurrentThreadCpuTime)();
  hname = prof_intern_name(name);
  env_id = event->env_id;

  /* CALL(RawMonitorEnter)(data_access_lock); /* 1.3 problem */

  /* CALL(RawMonitorExit)(data_access_lock); */
}

void prof_monitor_event(JVMPI_Event *event, jobjectID obj)
{
  jlong cur_time;
  JNIEnv *env_id;

  CALL(RawMonitorEnter)(data_access_lock);

  cur_time = CALL(GetCurrentThreadCpuTime)();
  env_id = event->env_id;

  switch (event->event_type) {
    case JVMPI_EVENT_MONITOR_CONTENDED_ENTER:
      write_monitor_event(PROF_MONITOR_ENTER,   obj, env_id, cur_time);
      break;
    case JVMPI_EVENT_MONITOR_CONTENDED_ENTERED:
      write_monitor_event(PROF_MONITOR_ENTERED, obj, env_id, cur_time);
      break;
  }

  CALL(RawMonitorExit)(data_access_lock);
}

void prof_monitor_wait_event(JVMPI_Event *event, jobjectID obj, jlong timeout)
{
  CALL(RawMonitorEnter)(data_access_lock);

  prof_fetch_thread_info(event->env_id);

  switch (event->event_type) {
    case JVMPI_EVENT_MONITOR_WAIT:
      write_monitor_event(PROF_MONITOR_WAIT,   obj, event->env_id, timeout);
      break;
    case JVMPI_EVENT_MONITOR_WAITED:
      write_monitor_event(PROF_MONITOR_WAITED, obj, event->env_id, timeout);
      break;
  }

  CALL(RawMonitorExit)(data_access_lock);
}

void prof_dump_monitors(void)
{
  CALL(RequestEvent)(JVMPI_EVENT_MONITOR_DUMP, NULL);
}

void prof_monitor_dump_event(JVMPI_Event *event)
{
  int i;
  int len = 0;
  int mon_num = 0;
  prof_trace_t **traces = NULL;

  CALL(RawMonitorEnter)(data_access_lock);
  if (event->u.monitor_dump.num_traces) {
    traces = prof_calloc(event->u.monitor_dump.num_traces * sizeof(void *));
  }

  for (i = 0; i < event->u.monitor_dump.num_traces; i++) {
    JVMPI_CallTrace *jtrace = &(event->u.monitor_dump.traces[i]);
    traces[i] = prof_intern_jvmpi_trace(jtrace->frames, jtrace->num_frames, jtrace->env_id);
  }
  prof_output_unmarked_traces();

  prof_dump_seek(event->u.monitor_dump.begin);
  while (prof_dump_cur() < event->u.monitor_dump.end) {
    int kind = prof_dump_read_u1();
    int n;
    JNIEnv *thread_env;
    if (kind == JVMPI_MONITOR_JAVA) {
      /* jobjectID obj = */ prof_dump_read_id();
      len += 21;
      mon_num++;
    } else if (kind == JVMPI_MONITOR_RAW) {
      char *name = prof_dump_read_id();
      prof_intern_name(name);
      prof_dump_read_id();
      len += 16;
      mon_num++;
    } else {
      if (debug) fprintf(stderr, "PROF ERROR: bad monitor kind: %d\n", kind);
    }
    thread_env = prof_dump_read_id();
    if (thread_env) {
      prof_fetch_thread_info(thread_env);
    }
    prof_dump_read_u4();
    n = prof_dump_read_u4();
    len += 4 * (n + 1);
    for (i = 0; i < n; i++) {
      thread_env = prof_dump_read_id();
      prof_fetch_thread_info(thread_env);
    }
    n = prof_dump_read_u4();
    len += 4 * (n + 1);
    for (i = 0; i < n; i++) {
      thread_env = prof_dump_read_id();
      prof_fetch_thread_info(thread_env);
    }
  }
  len += 6 * event->u.monitor_dump.num_traces + 8;
  prof_write_header(PROF_MONITOR_DUMP, len);
  prof_write_u4(event->u.monitor_dump.num_traces);
  for (i = 0; i < event->u.monitor_dump.num_traces; i++) {
    prof_write_u4(traces[i]->serial_num);
    prof_write_u4(event->u.monitor_dump.threads_status[i]);
  }
  prof_dump_seek(event->u.monitor_dump.begin);
  prof_write_u4(mon_num);
  while (prof_dump_cur() < event->u.monitor_dump.end) {
    int kind = prof_dump_read_u1();
    int n;
    int entry_count;
    JNIEnv *owner;
    prof_write_u1((unsigned char)kind);
    if (kind == JVMPI_MONITOR_JAVA) {
      jobjectID obj = prof_dump_read_id();
      prof_write_object_name(obj);
    } else if (kind == JVMPI_MONITOR_RAW) {
      char *name = prof_dump_read_id();
      JVMPI_RawMonitor mid = prof_dump_read_id();
      prof_write_id(prof_lookup_name(name));
      prof_write_id(mid);
    } else {
      if (debug) fprintf(stderr, "PROF ERROR: bad monitor kind: %d\n", kind);
    }
    owner = prof_dump_read_id();
    if (owner) {
      prof_write_thread_info(owner);
    } else {
      prof_write_u4(0);
    }
    entry_count = prof_dump_read_u4();
    prof_write_u4(entry_count);
    n = prof_dump_read_u4();
    prof_write_u4(n);
    if (n > 0) { /* waiting to enter */
      for (i = 0; i < n; i++) {
        void *thr = prof_dump_read_id();
        prof_write_thread_info(thr);
      }
    }
    n = prof_dump_read_u4();
    prof_write_u4(n);
    if (n > 0) { /* waiting to be notified */
      for (i = 0; i < n; i++) {
        void *thr = prof_dump_read_id();
        prof_write_thread_info(thr);
      }
    }
  }
  if (traces != NULL) prof_free(traces);
  CALL(RawMonitorExit)(data_access_lock);
}

void prof_write_object_name(jobjectID obj)
{
  prof_objmap_t *objmap = prof_fetch_object_info(obj);
  if (objmap == NULL) {
    if (debug) fprintf(stderr, "PROF ERROR: unknown object ID 0x%p\n", obj);
    prof_write_u1(0);
    prof_write_id(NULL);
    prof_write_id(NULL);
    prof_write_u4(0);
    return;
  }
  if (objmap->site == NULL) {
    if (debug) fprintf(stderr, "PROF ERROR: unknown object site 0x%p\n", obj);
    prof_write_u1(0);
    prof_write_id(NULL);
    prof_write_id(objmap);
    prof_write_u4(0);
    return;
  }
  prof_write_u1(objmap->site->is_array);
  if (objmap->site->is_array <= JVMPI_CLASS) {
    prof_class_t *hclass = objmap->site->class;
    if (hclass != NULL) {
      prof_write_id(prof_lookup_name(hclass->name->name));
    } else {
      prof_write_id(NULL);
    }
  } else {
    prof_write_id(NULL);
  }
  prof_write_id(objmap);
  prof_write_u4(objmap->site->trace_serial_num);
}

void prof_write_thread_info(JNIEnv *env)
{
  prof_thread_t *thread = prof_lookup_thread(env);
  if (thread == NULL) {
    prof_write_u4(0);
  } else {
    prof_write_u4(thread->serial_num);
  }
}
