/*
 * prof_object.c
 *
 * Copyright (c) ProSyst Software
 * All rights reserved
 *
 * Written by Vladislav Doshev,
 *            Alexi     Jordanov   (alexi91@yahoo.com),
 *            Ivo       Karabashev
 */

#include "prof.h"

/**
 * @author  Vladislav Doshev
 * @author  Alexi     Jordanov (alexi91@yahoo.com)
 * @author  Ivo       Karabashev
 * @version 1.00, 02/11/2000
 */

/* The table that maps objects to allocation sites */
/* 1009 3001 5003 10007 20011 30011 40009 */
static unsigned int objmap_table_size;
static unsigned int objmap_free_size;
#define HASH_OBJ(obj) ((unsigned int)obj % objmap_table_size)

static prof_objmap_t **prof_objmap_table;
static prof_objmap_t *objmap_free_list = NULL;
static unsigned int objmap_n_entries;

void prof_objmap_init(void)
{
	objmap_n_entries = 0;
	objmap_table_size = 40009;
	objmap_free_size = 0;
  prof_objmap_table = prof_calloc(objmap_table_size * sizeof(prof_objmap_t *));
}

static void add_alloc_stats(prof_site_t *site, unsigned int size)
{
  site->n_alloced_instances += 1;
  site->n_live_instances += 1;
  site->n_alloced_bytes += size;
  site->n_live_bytes += size;
  site->changed = 1;

  total_alloced_bytes = jlong_add(total_alloced_bytes, jint_to_jlong(size));
  total_alloced_instances = jlong_add(total_alloced_instances, jint_to_jlong(1));
  total_live_bytes += size;
  total_live_instances += 1;
}

static void sub_alloc_stats(prof_site_t *site, unsigned int size)
{
  site->n_live_instances--;
  site->n_live_bytes -= size;
  site->changed = 1;
  total_live_instances--;
  total_live_bytes -= size;
}

/* The objmap table maps objects to where they are allocated. Since this
 * is a much bigger table than site or trace tables, we do not use the
 * generic hash table routines for this purpose. Instead, we write
 * specialized code for better performance.
 */
static prof_objmap_t *prof_objmap_add(jobjectID obj_id, jint arena_id,
         prof_site_t *site, unsigned int size)
{
  int index;
  prof_objmap_t *bucket;

  if (objmap_n_entries >= objmap_table_size * 0.5) {
		//prof_objmap_rehash();
	}

  if (objmap_free_list) {
    bucket = objmap_free_list;
    objmap_free_list = bucket->next;
		objmap_free_size--;
  } else {
		/* buffer allocations */
    /* bucket = prof_calloc(sizeof(prof_objmap_t)); */
		int i;
    bucket = prof_calloc(100 * sizeof(prof_objmap_t));
		for (i = 1; i < 99; i++) {
      bucket[i].next = &bucket[i + 1];
		}
    objmap_free_list = &bucket[1];
    objmap_free_size = 99;
  }
  index = HASH_OBJ(obj_id);
  bucket->size = size;
  bucket->site = site;
  bucket->obj_id = obj_id;
  bucket->next = prof_objmap_table[index];
  prof_objmap_table[index] = bucket;

  if (site != NULL) {
    add_alloc_stats(site, size);
  }
	objmap_n_entries++;
  return bucket;
}

static void prof_objmap_del(jobjectID obj_id)
{
  int index = HASH_OBJ(obj_id);
  prof_objmap_t **p = prof_objmap_table + index;
  prof_objmap_t *bucket;
  while ((bucket = *p)) {
    if (bucket->obj_id == obj_id) {
      *p = bucket->next;
      bucket->next = objmap_free_list;
      objmap_free_list = bucket;
      objmap_free_size++;
      if (bucket->site != NULL) {
        sub_alloc_stats(bucket->site, bucket->size);
      }
			objmap_n_entries--;
      return;
    }
    p = &(bucket->next);
  }
  return;
}

static void prof_objmap_move(jobjectID obj_id, jint arena_id,
         jobjectID new_obj_id, jint new_arena_id)
{
  int old_index = HASH_OBJ(obj_id);
  int new_index = HASH_OBJ(new_obj_id);
  prof_objmap_t **p = prof_objmap_table + old_index;
  prof_objmap_t *bucket;

  /* remove it from the old position in the hash table */
  while ((bucket = *p)) {
    if (bucket->obj_id == obj_id) {
      *p = bucket->next;
      break;
    }
    p = &(bucket->next);
  }

  if (bucket != NULL) {
    /* rehash using the new obj_id; NOTE - the bucket remains the same */
    bucket->obj_id = new_obj_id;
    bucket->next = prof_objmap_table[new_index];
    prof_objmap_table[new_index] = bucket;
  } /* we ignore moves for objects with no prior record */
}

prof_objmap_t * prof_objmap_lookup(jobjectID obj_id)
{
  int index = HASH_OBJ(obj_id);
  prof_objmap_t *bucket = prof_objmap_table[index];
  while (bucket) {
    if (bucket->obj_id == obj_id) {
      return bucket;
    }
    bucket = bucket->next;
  }
  return NULL;
}

void prof_obj_alloc_event(JNIEnv *env_id, jobjectID class_id,
         jint is_array, unsigned int size, jobjectID obj_id, jint arena_id, int requested)
{
  prof_site_t *site;
  unsigned int trace_num = 0;

  CALL(RawMonitorEnter)(data_access_lock);
  if (requested) {
    if (prof_objmap_lookup(obj_id)) {
      goto done;
    }
    trace_num = 0;
  } else {
    prof_trace_t *htrace = prof_get_trace(env_id, max_trace_depth);
    if (htrace == NULL) {
      if (debug) fprintf(stderr, "PROF ERROR: got NULL trace in obj_alloc\n");
    } else {
      trace_num = htrace->serial_num;
    }
  }
  site = prof_intern_site(class_id, is_array, trace_num);
  prof_objmap_add(obj_id, arena_id, site, size);
done:
  CALL(RawMonitorExit)(data_access_lock);
}

void prof_obj_free_event(JNIEnv *env_id, jobjectID obj_id)
{
  /* data_access_lock already entered in GC_START event */
  prof_objmap_del(obj_id);
}

void prof_obj_move_event(JNIEnv *env_id, jobjectID obj_id, jint arena_id,
         jobjectID new_obj_id, jint new_arena_id)
{
  /* data_access_lock already entered in GC_START event */
  prof_objmap_move(obj_id, arena_id, new_obj_id, new_arena_id);
}

prof_objmap_t * prof_fetch_object_info(jobjectID obj)
{
  prof_objmap_t * objmap;
  if (obj == NULL) {
    return NULL;
  }
  objmap = prof_objmap_lookup(obj);
  if (objmap == NULL) {
    /* 1.4 problem when calling from monitor event */
    //CALL(RequestEvent)(JVMPI_EVENT_OBJ_ALLOC, obj);
    objmap = prof_objmap_lookup(obj);
  }
  return objmap;
}

void prof_objmap_rehash()
{
  unsigned int i;
  unsigned int index;

	unsigned int old_n_entries = objmap_n_entries;
	unsigned int old_size = objmap_table_size;
	prof_objmap_t **old_entries = prof_objmap_table;

  objmap_n_entries = 0;
  objmap_table_size = objmap_table_size * 2 + 1;
  prof_objmap_table = prof_calloc(objmap_table_size * sizeof(prof_objmap_t *));

  for (i = 0; i < old_size; i++)	{
    prof_objmap_t **p = &(old_entries[i]);
	  prof_objmap_t *bucket;
	  while ((bucket = *p)) {
      *p = bucket->next;
			index = HASH_OBJ(bucket->obj_id);
			bucket->next = prof_objmap_table[index];
			prof_objmap_table[index] = bucket;
			objmap_n_entries++;
		}
  }
	prof_free(old_entries);
	if (old_n_entries != objmap_n_entries) {
    fprintf(stderr, "PROF ERROR: Number of elements after rehashing is different.\n");
    CALL(ProfilerExit)((jint)1);
	}
}

void prof_print_objmap_tbl_hash_stats(void)
{
  unsigned int i;
  unsigned int empties = 0;
  unsigned int max_elems = 0;

  fprintf(stderr, "\nHash Table Statistics for 'OBJMAP'\n");

  for (i = 0; i < objmap_table_size; i++) {
    prof_objmap_t *bucket;
    unsigned int   elems = 0;

    for (bucket = prof_objmap_table[i]; bucket; bucket = bucket->next) elems++;
    if (elems == 0) empties++;
    else if (elems > max_elems) max_elems = elems;
  }

  if (objmap_n_entries == 0) {
    fprintf(stderr, "TABLE IS EMPTY!!\n");
  } else {
    fprintf(stderr, "tbl-size=%u   #-elems=%u   empty-slots=%.2f%%   avg-#-elems=%.2f   max-#-elems=%u\n",
      objmap_table_size,
      objmap_n_entries,
      ((float)empties / objmap_table_size) * 100.0f,
      (float)objmap_n_entries / (objmap_table_size - empties),
      max_elems);
#ifdef _HASH_STAT_
    fprintf(stderr, "mem-size=%u   pool-size=%u\n",
			objmap_n_entries * sizeof(prof_objmap_t),
			objmap_free_size * sizeof(prof_objmap_t));
#endif
  }
  fflush(stderr);
}
