/*
 * prof_setup.c
 *
 * Copyright (c) ProSyst Software
 * All rights reserved
 *
 * Written by Vladislav Doshev,
 *            Alexi     Jordanov   (alexi91@yahoo.com),
 *            Ivo       Karabashev
 */

#include <string.h>
#include "prof.h"

/**
 * @author  Vladislav Doshev
 * @author  Alexi     Jordanov (alexi91@yahoo.com)
 * @author  Ivo       Karabashev
 * @version 1.00, 02/11/2000
 */

static int prof_real_connect(char *hostname, unsigned short port);
static int prof_real_listen(unsigned short port);

typedef struct {
  char *name;
  int  *ptr;
} binary_switch_t;

int lineno_in_traces_unused = TRUE;
int thread_in_traces_unused = TRUE;
int monitor_tracing_unused  = FALSE;

static binary_switch_t binary_switches[] = {
  {"lineno",  &lineno_in_traces_unused},
  {"thread",  &thread_in_traces_unused},
  {"doe",     &dump_on_exit},
  {"monitor", &monitor_tracing_unused}
};

static int get_tok(char **src, char *buf, int buflen, char sep)
{
  int i;
  char *p = *src;
  for (i = 0; i < buflen; i++) {
    if (p[i] == 0 || p[i] == sep) {
      buf[i] = 0;
      if (p[i] == sep) {
        i++;
      }
      *src += i;
      return i;
    }
    buf[i] = p[i];
  }
  /* overflow */
  return 0;
}

static void prof_print_usage(void)
{
  fprintf(stdout,
    "Prof usage: -Xrunprof[:help]|[<option>=<value>, ...]\n\n"
    "Option Name and Value  Description                 Default\n"
    "---------------------  --------------------------  -------\n"
    "net=<host>:<port>      send data over a socket\n"
    "attach=<port>          wait for connection\n"
    "heap=dump|sites|all    heap profiling              all\n"
    "cpu=samples|times      CPU usage                   off\n"
    "depth=<size>           stack trace depth           4\n"
    "cutoff=<value>         output cutoff point         0.0001\n"
    "lineno=y|n             line number in traces?      y\n"
    "thread=y|n             thread in traces?           y\n"
    "doe=y|n                dump on exit?               n\n\n");
}

static void prof_do_setup(char *str)
{
  int output_specified = FALSE;

  prof_is_on = TRUE;

  if (str == 0) str = "";

  if ((strcmp(str, "help")) == 0) {
    prof_print_usage();
    CALL(ProfilerExit)((jint)0);
  }

  while (*str) {
    char buf[1000];
    if (!get_tok(&str, buf, sizeof(buf), '=')) {
      goto bad_option;
    }

    /* file */
    if (strcmp(buf, "file") == 0) {
      fprintf(stderr, "PROF ERROR: only network output is supported\n");
      goto bad_option_no_msg;

    /* net */
    } else if (strcmp(buf, "net") == 0) {
      char port_str[6];
      unsigned short port;
      if (output_specified ||
          !get_tok(&str, buf, sizeof(buf), ':') ||
          !get_tok(&str, port_str, sizeof(port_str), ',')) {
        goto bad_option;
      }
      port = (unsigned short)atoi(port_str);
      prof_fd = prof_real_connect(buf, port);
      if (prof_fd <= 0) {
        fprintf(stderr, "PROF ERROR: can't connect to %s:%u", buf, port);
        goto bad_option_no_msg;
      }
      output_specified = TRUE;

    /* attach */
    } else if (strcmp(buf, "attach") == 0) {
      unsigned short port;
      if (output_specified || !get_tok(&str, buf, sizeof(buf), ',')) {
        goto bad_option;
      }
      port = (unsigned short)atoi(buf);
      prof_fd = prof_real_listen(port);
      if (prof_fd <= 0) {
        fprintf(stderr, "PROF ERROR: can't listen on %u", port);
        goto bad_option_no_msg;
      }
      output_specified = TRUE;

    /* format */
    } else if (strcmp(buf, "format") == 0) {
      if (!get_tok(&str, buf, sizeof(buf), ',')) {
        goto bad_option;
      }
      if (strcmp(buf, "a") == 0) {
        fprintf(stderr, "PROF ERROR: only binary format is supported\n");
        goto bad_option_no_msg;
      } else if (strcmp(buf, "b") != 0) {
        goto bad_option;
      }

    /* heap */
    } else if (strcmp(buf, "heap") == 0) {
      if (!get_tok(&str, buf, sizeof(buf), ',')) {
        goto bad_option;
      }
      if (strcmp(buf, "dump") == 0) {
      } else if (strcmp(buf, "sites") == 0) {
        alloc_sites = TRUE;
      } else if (strcmp(buf, "all") == 0) {
        alloc_sites = TRUE;
      } else {
        goto bad_option;
      }

    /* cpu */
    } else if (strcmp(buf, "cpu") == 0) {
      if (!get_tok(&str, buf, sizeof(buf), ',')) {
        goto bad_option;
      }
      if ((strcmp(buf, "samples") == 0) || (strcmp(buf, "y") == 0)) {
        cpu_sampling = TRUE;
      } else {
        goto bad_option;
      }

    /* depth */
    } else if (strcmp(buf, "depth") == 0) {
      if (!get_tok(&str, buf, sizeof(buf), ',')) {
        goto bad_option;
      }
      max_trace_depth = atol(buf);
      prof_trace_depth = max_trace_depth;

    /* cutoff */
    } else if (strcmp(buf, "cutoff") == 0) {
      if (!get_tok(&str, buf, sizeof(buf), ',')) {
        goto bad_option;
      }
      prof_cutoff_point = (float) atof(buf);

    /* debug */
    } else if (strcmp(buf, "debug") == 0) {
      if (!get_tok(&str, buf, sizeof(buf), ',')) {
        goto bad_option;
      }
      if (strcmp(buf, "true") == 0) {
        debug = TRUE;
      } else if (strcmp(buf, "false") == 0) {
        debug = FALSE;
      } else {
        goto bad_option;
      }
    } else {
      int i;
      int n_switches = sizeof(binary_switches) / sizeof(binary_switch_t);
      for (i = 0; i < n_switches; i++) {
        if (strcmp(binary_switches[i].name, buf) == 0) {
          if (!get_tok(&str, buf, sizeof(buf), ',')) {
            goto bad_option;
          }
          if (strcmp(buf, "y") == 0) {
            *(binary_switches[i].ptr) = TRUE;
          } else if (strcmp(buf, "n") == 0) {
            *(binary_switches[i].ptr) = FALSE;
          } else {
            goto bad_option;
          }
          break;
        }
      }
      if (i >= n_switches) {
        goto bad_option;
      }
    }
  }

  /* By default we turn on alloc_sites and heap_dump */
  if (!cpu_sampling && !alloc_sites) {
    alloc_sites = TRUE;
  }
  return;

bad_option:
  fprintf(stderr, "PROF ERROR: bad prof option\n");
bad_option_no_msg:
  prof_real_disconnect();
  CALL(ProfilerExit)((jint)1);
}

jint prof_enable_monitors(void)
{
  /*if (CALL(EnableEvent)(JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTER, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }*/
  /*if (CALL(EnableEvent)(JVMPI_EVENT_RAW_MONITOR_CONTENDED_ENTERED, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }*/
  /*if (CALL(EnableEvent)(JVMPI_EVENT_RAW_MONITOR_CONTENDED_EXIT, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }*/
  if (CALL(EnableEvent)(JVMPI_EVENT_MONITOR_CONTENDED_ENTER, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }
  if (CALL(EnableEvent)(JVMPI_EVENT_MONITOR_CONTENDED_ENTERED, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }
  /*if (CALL(EnableEvent)(JVMPI_EVENT_MONITOR_CONTENDED_EXIT, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }*/
  if (CALL(EnableEvent)(JVMPI_EVENT_MONITOR_WAIT, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }
  if (CALL(EnableEvent)(JVMPI_EVENT_MONITOR_WAITED, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }

  return JNI_OK;
}

static jint prof_enable_events(void)
{
  if (CALL(EnableEvent)(JVMPI_EVENT_CLASS_LOAD, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }
  if (CALL(EnableEvent)(JVMPI_EVENT_CLASS_UNLOAD, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }
  if (CALL(EnableEvent)(JVMPI_EVENT_THREAD_START, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }
  if (CALL(EnableEvent)(JVMPI_EVENT_THREAD_END, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }
  /*if (CALL(EnableEvent)(JVMPI_EVENT_DUMP_DATA_REQUEST, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }*/
  /*if (CALL(EnableEvent)(JVMPI_EVENT_RESET_DATA_REQUEST, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }*/
  if (CALL(EnableEvent)(JVMPI_EVENT_JVM_INIT_DONE, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }
  if (CALL(EnableEvent)(JVMPI_EVENT_JVM_SHUT_DOWN, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }

  if (CALL(EnableEvent)(JVMPI_EVENT_GC_START, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }
  if (CALL(EnableEvent)(JVMPI_EVENT_GC_FINISH, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }

  if (alloc_sites) {
    if (CALL(EnableEvent)(JVMPI_EVENT_OBJECT_ALLOC, NULL) != JVMPI_SUCCESS) {
      return JNI_ERR;
    }
    if (CALL(EnableEvent)(JVMPI_EVENT_OBJECT_MOVE, NULL) != JVMPI_SUCCESS) {
      return JNI_ERR;
    }
    if (CALL(EnableEvent)(JVMPI_EVENT_OBJECT_FREE, NULL) != JVMPI_SUCCESS) {
      return JNI_ERR;
    }
  }

  /* Problem in J9 */
  /*if (CALL(EnableEvent)(JVMPI_EVENT_ARENA_NEW, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }
  if (CALL(EnableEvent)(JVMPI_EVENT_ARENA_DELETE, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }*/

	/* No need for now */
  /*if (CALL(EnableEvent)(JVMPI_EVENT_JNI_GLOBALREF_ALLOC, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }
  if (CALL(EnableEvent)(JVMPI_EVENT_JNI_GLOBALREF_FREE, NULL) != JVMPI_SUCCESS) {
    return JNI_ERR;
  }*/

  return JNI_OK;
}

void prof_init_setup(char *profiler_options)
{
  jint settings = 0;
  prof_do_setup(profiler_options);

  total_alloced_bytes = jlong_zero;
  total_alloced_instances = jlong_zero;

  prof_io_setup();
  prof_class_table_init();
  prof_thread_table_init();
  prof_method_table_init();
  prof_name_table_init();
  prof_frame_table_init();
  prof_trace_table_init();
  prof_site_table_init();
  prof_objmap_init();
  prof_trigger_table_init();

  /* Create data access lock */
  data_access_lock = CALL(RawMonitorCreate)("_prof_data_access_lock");

  prof_enable_events();

  /* Write the header information */
  prof_write_raw(PROF_HEADER, strlen(PROF_HEADER) + 1);
  prof_write_u4(sizeof(void *));
  {
    jlong t = prof_get_timemillis();
    prof_write_u4(jlong_high(t));
    prof_write_u4(jlong_low(t));
  }
  prof_write_header(PROF_CONTROL_SETTINGS, 4 + 2);
  if (alloc_sites) settings |= 1;
  if (cpu_sampling) settings |= 2;
  prof_write_u4(settings);
  prof_write_u2((unsigned short)max_trace_depth);

  micro_sec_ticks = prof_get_milliticks() * 1000;
}

/**
 * Socket connection
 */
#define PROF_MAX_HOSTADDR_SIZE 512

/*
 * Return a socket descriptor connect()ed to a "hostname" that is
 * accept()ing heap profile data on "port." Return a value <= 0 if
 * such a connection can't be made.
 */
static int prof_real_connect(char *hostname, unsigned short port)
{
#ifdef _WIN32_
  /* Alex's correction {{ */
  static WSADATA wsaData;
  /* }} */
#endif

  struct hostent *hentry = NULL;
  struct sockaddr_in s;
  int fd;

  if (port == 0) {
    fprintf(stderr, "PROF ERROR: bad port number\n");
    return -1;
  }
  if (hostname == NULL) {
    fprintf(stderr, "PROF ERROR: hostname is NULL\n");
    return -1;
  }

#ifdef _WIN32_
  /* Alex Jordanov's correction {{ */
  if (WSAStartup(0x0202, &wsaData) == SOCKET_ERROR) {
    fprintf(stderr, "PROF ERROR: transport error in WSAStartup(): %d\n", WSAGetLastError());
    WSACleanup();
    return -1;
  }
  /* }} */
#endif

  /* create a socket */
  fd = socket(AF_INET, SOCK_STREAM, 0);

  memset((char *)&s, 0, sizeof(s));

  /* find remote host's addr from name */
#ifdef _VXWORKS_
  /* Boryan Yotov's correction - Host resolve for vxWorks */
  if (((s.sin_addr.s_addr = inet_addr(hostname)) == ERROR) &&
      ((s.sin_addr.s_addr = hostGetByName(hostname)) == ERROR)) {
    return -1;
  }
  /* Boryan Yotov's correction end */
#else
  hentry = gethostbyname(hostname);
  if (hentry == NULL) {
    return -1;
  }

  /* set remote host's addr; its already in network byte order */
  memcpy(&s.sin_addr.s_addr, *(hentry->h_addr_list), sizeof(s.sin_addr.s_addr));
#endif

  /* set remote host's port */
  s.sin_port = htons(port);
  s.sin_family = AF_INET;

  /* now try connecting */
  if (-1 == connect(fd, (struct sockaddr*)&s, sizeof(s))) {
    fprintf(stderr, "PROF ERROR: connect doesn't succeed\n");
    return 0;
  } else {
    /* fprintf(stderr, "PROF: connect succeed\n"); */
    return fd;
  }
}

/*
 * Return a socket descriptor connect()ed to a "hostname" that is
 * accept()ing heap profile data on "port." Return a value <= 0 if
 * such a connection can't be made.
 */
static int prof_real_listen(unsigned short port)
{
#ifdef _WIN32_
  /* Alex's correction {{ */
  static WSADATA wsaData;
  /* }} */
#endif

  struct sockaddr_in srv;
  struct sockaddr_in clnt;
  int sock_fd, new_fd;
  int sin_size;

  if (port == 0) {
    fprintf(stderr, "PROF ERROR: bad port number\n");
    return -1;
  }

#ifdef _WIN32_
  /* Alex Jordanov's correction {{ */
  if (WSAStartup(0x0202, &wsaData) == SOCKET_ERROR) {
    fprintf(stderr, "PROF ERROR: transport error in WSAStartup(): %d\n", WSAGetLastError());
    WSACleanup();
    return -1;
  }
  /* }} */
#endif

  /* create a socket */
  sock_fd = socket(AF_INET, SOCK_STREAM, 0);

  srv.sin_family = AF_INET;
  srv.sin_port = htons(port);
  srv.sin_addr.s_addr = INADDR_ANY;

  /* now try connecting */
  bind(sock_fd, (struct sockaddr *)&srv, sizeof(struct sockaddr));

  listen(sock_fd, 1);

  sin_size = sizeof(struct sockaddr_in);
  new_fd = accept(sock_fd, (struct sockaddr *)&clnt, &sin_size);

  if (new_fd == -1) {
    fprintf(stderr, "PROF ERROR: connect doesn't succeed\n");
    return 0;
  } else {
    return new_fd;
  }
}

void prof_real_disconnect(void) {
#ifdef _WIN32_
  #define SHUTCODE 2

  if (prof_fd > 0) {
    Sleep(1000);
    shutdown(prof_fd, SHUTCODE);
    closesocket(prof_fd);
  }
  WSACleanup();
#else
  close(prof_fd);
#endif
}
