/*
 * prof_site.c
 *
 * Copyright (c) ProSyst Software
 * All rights reserved
 *
 * Written by Vladislav Doshev,
 *            Alexi     Jordanov   (alexi91@yahoo.com),
 *            Ivo       Karabashev
 */

#include "prof.h"

/**
 * @author  Vladislav Doshev
 * @author  Alexi     Jordanov (alexi91@yahoo.com)
 * @author  Ivo       Karabashev
 * @version 1.00, 02/11/2000
 */

#define PROF_SITE_TABLE_SIZE 30011

prof_hash_t prof_site_table;

static unsigned int hash_site(void *_site)
{
  prof_site_t *site = _site;
  return (unsigned int)site->trace_serial_num +
         37 * ((unsigned int)site->is_array + 37 * (unsigned int)site->class);
}

static unsigned int size_site(void *_site)
{
  return sizeof(prof_site_t);
}

static int compare_site(void *_site1, void *_site2)
{
  prof_site_t *site1 = _site1;
  prof_site_t *site2 = _site2;
  int result = (int)((unsigned int)site1->class - (unsigned int)site2->class);
  if (result) return result;
  result = (int)((unsigned int)site1->is_array - (unsigned int)site2->is_array);
  if (result) return result;
  return site1->trace_serial_num - site2->trace_serial_num;
}

void prof_site_table_init(void)
{
  prof_hash_init(&prof_site_table, PROF_SITE_TABLE_SIZE, hash_site, size_site, compare_site);
}

prof_site_t * prof_intern_site(jobjectID class_id, jint is_array, unsigned int trace_serial_num)
{
  prof_site_t site_tmp;
  prof_site_t *result;

  site_tmp.class = NULL;
  if (class_id != 0) {
    site_tmp.class = prof_lookup_class(class_id);
  }
  site_tmp.is_array = (unsigned char)is_array;
  site_tmp.trace_serial_num = trace_serial_num;
  site_tmp.n_alloced_instances = 0;
  site_tmp.n_alloced_bytes = 0;
  site_tmp.n_live_instances = 0;
  site_tmp.n_live_bytes = 0;

  result = prof_hash_intern(&prof_site_table, &site_tmp);
  result->changed = 1;
  return result;
}

static void * prof_site_collect(void *_site, void *_arg)
{
  prof_site_iterate_t *arg = _arg;
  prof_site_t *site = _site;
  if (!site->changed) {
    return _arg;
  }
  arg->sites[arg->index++] = _site;
  return _arg;
}

void prof_clear_site_table(void)
{
  prof_hash_removeall(&prof_site_table);
}

static void * prof_site_mark_unchanged(void *_site, void *_arg)
{
  prof_site_t *site = _site;
  site->changed = 0;
  return _arg;
}

/* Output allocation sites */
void prof_output_sites(int flags, float cutoff)
{
  prof_site_iterate_t iterate;
  int i;
  int site_table_size;

  CALL(RawMonitorEnter)(data_access_lock);
  site_table_size = prof_site_table.n_entries;
  iterate.sites = prof_calloc(site_table_size * sizeof (void *));
  iterate.index = 0;
  prof_hash_iterate(&prof_site_table, prof_site_collect, &iterate);

  site_table_size = iterate.index;

  prof_output_unmarked_traces_limited();

  prof_write_header(PROF_ALLOC_SITES, site_table_size * (4 * 6 + 1) + 2 + 32);
  prof_write_u2((unsigned short)flags);
  prof_write_u4(0);
  prof_write_u4(total_live_bytes);
  prof_write_u4(total_live_instances);
  prof_write_u4(jlong_high(total_alloced_bytes));
  prof_write_u4(jlong_low(total_alloced_bytes));
  prof_write_u4(jlong_high(total_alloced_instances));
  prof_write_u4(jlong_low(total_alloced_instances));
  prof_write_u4(site_table_size);

  for (i = 0; i < site_table_size; i++) {
    prof_site_t *site = iterate.sites[i];
    unsigned int class_serial_num;

    if (site->class == NULL) {
      class_serial_num = 0;
    } else {
      class_serial_num = site->class->serial_num;
    }

    prof_write_u1(site->is_array);
    prof_write_u4(class_serial_num);
    prof_write_u4(site->trace_serial_num);
    prof_write_u4(site->n_live_bytes);
    prof_write_u4(site->n_live_instances);
    prof_write_u4(site->n_alloced_bytes);
    prof_write_u4(site->n_alloced_instances);
  }

  prof_hash_iterate(&prof_site_table, prof_site_mark_unchanged, NULL);

  prof_free(iterate.sites);
  CALL(RawMonitorExit)(data_access_lock);
}
