/*
 * prof_thread.c
 *
 * Copyright (c) ProSyst Software
 * All rights reserved
 *
 * Written by Vladislav Doshev,
 *            Alexi     Jordanov   (alexi91@yahoo.com),
 *            Ivo       Karabashev
 */

#include "prof.h"

/**
 * @author  Vladislav Doshev
 * @author  Alexi     Jordanov (alexi91@yahoo.com)
 * @author  Ivo       Karabashev
 * @version 1.00, 02/11/2000
 */

#define PROF_THREAD_TABLE_SIZE 101

static unsigned int thread_serial_number = 1;

prof_hash_t prof_thread_table;

/* list of active threads */
live_thread_t *live_thread_list = NULL;
int num_live_threads = 0;

/* list of nonsystem threads */
nonsys_thread_t *nonsys_thread_list = NULL;
int num_nonsys_threads = 0;

static unsigned int hash_thread(void *_thread)
{
  prof_thread_t *thread = _thread;
  return (unsigned int)thread->env_id;
}

static unsigned int size_thread(void *_thread)
{
  return sizeof(prof_thread_t);
}

static int compare_thread(void *_thread1, void *_thread2)
{
  prof_thread_t *thread1 = _thread1;
  prof_thread_t *thread2 = _thread2;

  return (int)((unsigned int)thread1->env_id - (unsigned int)thread2->env_id);
}

void prof_thread_table_init(void)
{
  prof_hash_init(&prof_thread_table, PROF_THREAD_TABLE_SIZE, hash_thread, size_thread, compare_thread);
}

prof_thread_t * prof_intern_thread(JNIEnv *env_id)
{
  prof_thread_t thread_tmp;
  prof_thread_t *result;

  thread_tmp.env_id = env_id;
  result = prof_hash_lookup(&prof_thread_table, &thread_tmp);
  if (result == NULL) {
    thread_tmp.serial_num = thread_serial_number++;
    /* set thread id to NULL, it'll be set to the right value in thread_start_event */
    thread_tmp.thread_id = NULL;
    result = prof_hash_put(&prof_thread_table, &thread_tmp);
  }
  return result;
}

static int prof_remove_thread_helper(void *_thread, void *_env)
{
  prof_thread_t *thread = _thread;
  JNIEnv *env_id = _env;
  return (thread->env_id == env_id);
}

void prof_remove_thread(JNIEnv *env_id)
{
  prof_hash_remove(&prof_thread_table, prof_remove_thread_helper, (void *)env_id);
}

prof_thread_t *prof_lookup_thread(JNIEnv *env_id)
{
  prof_thread_t thread_tmp;
  prof_thread_t *result;
  thread_tmp.env_id = env_id;
  result = prof_hash_lookup(&prof_thread_table, &thread_tmp);
  return result;
}

void prof_thread_start_event(JNIEnv *env_id,
         char *t_name, char *g_name, char *p_name,
         jobjectID thread_id, int requested)
{
  prof_thread_t *result;
  prof_name_t *thread_name;
  prof_name_t *group_name;
  prof_name_t *parent_name;
  prof_objmap_t *objmap;
  int trace_num;

  CALL(RawMonitorEnter)(data_access_lock);

  objmap = prof_fetch_object_info(thread_id);

  if (objmap == NULL) {
    if (debug) fprintf(stderr, "PROF ERROR: unable to map JVMPI thread ID to "
                       "prof thread ID in thread_start \n");
    goto threadstart_done;
  }

  result = prof_intern_thread(env_id);
  if (result->thread_id != NULL) {
    goto threadstart_done;
  }
  {
    live_thread_t *newthread = prof_calloc(sizeof(live_thread_t));
    /* add to the list of live threads */
    newthread->next = live_thread_list;
    newthread->tid = objmap;
    newthread->env = env_id;
    newthread->cpu_sampled = 1;
    live_thread_list = newthread;
    num_live_threads++;
  }
  if (p_name != NULL) {
    nonsys_thread_t *newthread = prof_calloc(sizeof(nonsys_thread_t));
    /* add to the list of live threads */
    newthread->next = nonsys_thread_list;
    newthread->env = env_id;
    nonsys_thread_list = newthread;
    num_nonsys_threads++;
  }
  result = prof_intern_thread(env_id);
  if (result->thread_id != NULL) {
    if (debug) fprintf(stderr, "PROF ERROR: thread ID already in use\n");
    goto threadstart_done;
  }
  result->thread_id = objmap;
  thread_name = prof_intern_name(t_name);
  group_name = prof_intern_name(g_name);
  parent_name = prof_intern_name(p_name);

  if (requested) {
    trace_num = 0;
  } else {
    prof_trace_t *htrace = prof_get_trace(env_id, max_trace_depth);
    if (htrace == NULL) {
      if (debug) fprintf(stderr, "PROF ERROR: got NULL trace in thread_start\n");
      goto threadstart_done;
    }
    trace_num = htrace->serial_num;
  }
  prof_write_header(PROF_START_THREAD, sizeof(void *) * 4 + 8);
  prof_write_u4(result->serial_num);
  prof_write_id(objmap);
  prof_write_u4(trace_num);
  prof_write_id(thread_name);
  prof_write_id(group_name);
  prof_write_id(parent_name);
threadstart_done:
  CALL(RawMonitorExit)(data_access_lock);
}

void prof_thread_end_event(JNIEnv *env_id)
{
  prof_thread_t thread_tmp;
  prof_thread_t *thread;

  CALL(RawMonitorEnter)(data_access_lock);
  {
    /* remove from list of live threads */
    live_thread_t **p;
    p = &live_thread_list;
    while (*p) {
      live_thread_t *t = *p;
      if (t->env == env_id) {
        *p = t->next;
        prof_free(t);
        break;
      }
      p = &(t->next);
    }
    num_live_threads--;
  }
  {
    /* remove from list of live threads */
    nonsys_thread_t **p;
    p = &nonsys_thread_list;
    while (*p) {
      nonsys_thread_t *t = *p;
      if (t->env == env_id) {
        *p = t->next;
        prof_free(t);
        break;
      }
      p = &(t->next);
    }
    num_nonsys_threads--;
  }
  thread_tmp.env_id = env_id;
  thread = prof_hash_lookup(&prof_thread_table, &thread_tmp);
  if (thread == NULL) {
    if (debug) fprintf(stderr, "PROF ERROR: unknown thread ID in thread_end\n");
  } else {
    prof_write_header(PROF_END_THREAD, 4);
    prof_write_u4(thread->serial_num);
    prof_remove_thread(env_id);
  }
  CALL(RawMonitorExit)(data_access_lock);
}

prof_thread_t * prof_fetch_thread_info(JNIEnv *env)
{
  prof_thread_t *thread = prof_lookup_thread(env);
  if (thread == NULL) {
    jobjectID tobj = CALL(GetThreadObject)(env);
    if (tobj) {
      CALL(RequestEvent)(JVMPI_EVENT_THREAD_START, tobj);
    }
    thread = prof_lookup_thread(env);
  }
  return thread;
}
