/*
 * prof_trace.c
 *
 * Copyright (c) ProSyst Software
 * All rights reserved
 *
 * Written by Vladislav Doshev,
 *            Alexi     Jordanov   (alexi91@yahoo.com),
 *            Ivo       Karabashev
 */

#include <string.h>
#include "prof.h"

/**
 * @author  Vladislav Doshev
 * @author  Alexi     Jordanov (alexi91@yahoo.com)
 * @author  Ivo       Karabashev
 * @version 1.00, 02/11/2000
 */

#define PROF_TRACE_TABLE_SIZE 30011
#define PROF_FRAME_TABLE_SIZE 10007

static unsigned int trace_serial_number = 1;
prof_hash_t prof_trace_table;

static unsigned int hash_trace(void * _trace)
{
  prof_trace_t *trace = _trace;
  unsigned int hash = trace->thread_serial_num + 37 * trace->n_frames;
  unsigned int i;
  for (i = 0; i < trace->n_frames; i++) {
    hash = hash * 37 + (unsigned int)trace->frames[i];
  }
  return hash;
}

static unsigned int size_trace(void * _trace)
{
  prof_trace_t *trace = _trace;
  return ((char *)&(trace->frames) - (char *)trace) + trace->n_frames * sizeof(prof_frame_t *);
}

static int compare_trace(void *_trace1, void *_trace2)
{
  prof_trace_t *trace1 = _trace1;
  prof_trace_t *trace2 = _trace2;
  unsigned int i;
  int result = trace1->thread_serial_num - trace2->thread_serial_num;
  if (result) return result;
  result = trace1->n_frames - trace2->n_frames;
  if (result) return result;
  for (i = 0; i < trace1->n_frames; i++) {
    result = (int)((unsigned int)trace1->frames[i] - (unsigned int)trace2->frames[i]);
    if (result) return result;
  }
  return 0;
}

void prof_trace_table_init(void)
{
  prof_hash_init(&prof_trace_table, PROF_TRACE_TABLE_SIZE, hash_trace, size_trace, compare_trace);
}

prof_trace_t *prof_alloc_tmp_trace(unsigned int n_frames, JNIEnv *env_id)
{
  prof_trace_t *trace_tmp = NULL;
  int thread_serial_num = 0;
  trace_tmp = prof_calloc(n_frames * sizeof(prof_frame_t *) +
              ((char *)&(trace_tmp->frames) - (char *)trace_tmp));
  trace_tmp->n_frames = (unsigned short) n_frames;
  trace_tmp->serial_num = 0;
  trace_tmp->marked = 0;
  trace_tmp->num_hits = 0;

  /* include thread info only if thread_in_traces is true and env_id is not NULL. */
  if ((!thread_in_traces) || (env_id == NULL)) {
    thread_serial_num = 0;
  } else {
    prof_thread_t *thread = prof_intern_thread(env_id);
    thread_serial_num = thread->serial_num;
  }
  trace_tmp->thread_serial_num = thread_serial_num;

  return trace_tmp;
}

prof_trace_t *prof_intern_tmp_trace(prof_trace_t *trace_tmp)
{
  prof_trace_t *result = prof_hash_lookup(&prof_trace_table, trace_tmp);
  if (result == NULL) {
    trace_tmp->serial_num = trace_serial_number++;
    result = prof_hash_put(&prof_trace_table, trace_tmp);
  }
  prof_free(trace_tmp);

  return result;
}

prof_trace_t * prof_intern_jvmpi_trace(JVMPI_CallFrame *frames, unsigned int n_frames, JNIEnv *env_id)
{
  unsigned int i;
  prof_trace_t *trace_tmp = prof_alloc_tmp_trace(n_frames, env_id);

  /* intern all the frames in the trace */
  for (i = 0; i < n_frames; i++) {
    prof_frame_t *frame = prof_intern_jvmpi_frame(frames[i].method_id, frames[i].lineno);
    if (frame == NULL) {
      prof_free(trace_tmp);
      return NULL;
    }
    trace_tmp->frames[i] = frame;
  }
  return prof_intern_tmp_trace(trace_tmp);
}

prof_trace_t *prof_get_trace(JNIEnv *env_id, unsigned int depth)
{
  prof_trace_t *htrace;
  JVMPI_CallTrace jtrace;

  /* allocate space for the trace */
  jtrace.frames = (JVMPI_CallFrame *)prof_calloc(depth * sizeof(JVMPI_CallFrame));

  /* Get the trace from the JVM */
  jtrace.env_id = env_id;
  CALL(GetCallTrace)(&jtrace, depth);

  /* intern the trace */
  htrace = prof_intern_jvmpi_trace(jtrace.frames, jtrace.num_frames, jtrace.env_id);

  /* Free up the space */
  prof_free(jtrace.frames);

  return htrace;
}

prof_hash_t prof_frame_table;

static unsigned int hash_frame(void *_frame)
{
  prof_frame_t *frame = _frame;
  return ((unsigned int)frame->method + 31 * (unsigned int)frame->lineno);
}

static unsigned int size_frame(void *_frame)
{
  return sizeof(prof_frame_t);
}

static int compare_frame(void *_frame1, void *_frame2)
{
  prof_frame_t *frame1 = _frame1;
  prof_frame_t *frame2 = _frame2;

  int result = frame1->lineno - frame2->lineno;
  if (result) return result;
  return (int)((unsigned int)frame1->method - (unsigned int)frame2->method);
}

void prof_frame_table_init(void)
{
  prof_hash_init(&prof_frame_table, PROF_FRAME_TABLE_SIZE, hash_frame, size_frame, compare_frame);
}

prof_frame_t * prof_intern_jvmpi_frame(jmethodID method_id, int lineno)
{
  prof_frame_t frame_tmp;

  frame_tmp.marked = 0;
  if (lineno_in_traces) {
    frame_tmp.lineno = lineno;
  } else {
    frame_tmp.lineno = -1;
  }

  frame_tmp.method = prof_lookup_method(method_id);
  if (frame_tmp.method == NULL) {
    if (debug) fprintf(stderr, "PROF ERROR: unable to resolve a method id\n");
    return NULL;
  }

  return prof_hash_intern(&prof_frame_table, &frame_tmp);
}

/* Write out a stack trace. */
static void prof_output_trace(prof_trace_t *trace)
{
  unsigned int i;
  if (trace->marked == 2) return;

  if (trace->marked == 1 && trace->n_frames < 2) return;

  trace->marked = 2;

  for (i = 0; i < trace->n_frames; i++) {
    if (trace->frames[i]->marked == 0) {
      prof_frame_t *frame = trace->frames[i];
      prof_method_t *method = frame->method;
      prof_name_t *method_name = method->method_name;
      prof_class_t *class = method->class;
      prof_name_t *src_name = class->src_name;

      frame->marked = 1;
      prof_write_header(PROF_FRAME, sizeof(void *) * 4 + 8);
      prof_write_id(frame);
      prof_write_id(method_name);
      prof_write_id(NULL); /* method_signature removed */
      prof_write_id(src_name);
      prof_write_u4(class->serial_num);
      prof_write_u4(frame->lineno);
    }
  }
  prof_write_header(PROF_TRACE, sizeof(void *) * trace->n_frames + 12);
  prof_write_u4(trace->serial_num);
  prof_write_u4(trace->thread_serial_num);
  prof_write_u4(trace->n_frames);
  for (i = 0; i < trace->n_frames; i++) {
    prof_write_id(trace->frames[i]);
  }
}

static void prof_output_trace_limited(prof_trace_t *trace)
{
  if (trace->marked) return;
  trace->marked = 1;

  if (trace->n_frames > 0) {
    if (trace->frames[0]->marked == 0) {
      prof_frame_t *frame = trace->frames[0];
      prof_method_t *method = frame->method;
      prof_name_t *method_name = method->method_name;
      prof_class_t *class = method->class;
      prof_name_t *src_name = class->src_name;

      frame->marked = 1;
      prof_write_header(PROF_FRAME, sizeof(void *) * 4 + 8);
      prof_write_id(frame);
      prof_write_id(method_name);
      prof_write_id(NULL); /* method_signature removed */
      prof_write_id(src_name);
      prof_write_u4(class->serial_num);
      prof_write_u4(frame->lineno);
    }      
    prof_write_header(PROF_TRACE, sizeof(void *) + 12);
    prof_write_u4(trace->serial_num);
    prof_write_u4(trace->thread_serial_num);
    prof_write_u4(1);
    prof_write_id(trace->frames[0]);
  }
}

static void *prof_output_unmarked_trace(void *_trace, void *arg)
{
  prof_trace_t *trace = _trace;
  prof_output_trace(trace);
  return arg;
}

static void *prof_output_unmarked_trace_limited(void *_trace, void *arg)
{
  prof_trace_t *trace = _trace;
  prof_output_trace_limited(trace);
  return arg;
}

void prof_output_unmarked_traces(void)
{
  prof_hash_iterate(&prof_trace_table, prof_output_unmarked_trace, 0);
}

void prof_output_unmarked_traces_limited(void)
{
  prof_hash_iterate(&prof_trace_table, prof_output_unmarked_trace_limited, 0);
}

static void * prof_trace_collect(void *_trace, void *_arg)
{
  prof_trace_iterate_t *arg = _arg;
  prof_trace_t *trace = _trace;
  arg->traces[arg->index++] = trace;
  arg->total_count += (int) trace->num_hits;
  return _arg;
}

static int prof_trace_compare_cost(const void *p_trace1, const void *p_trace2)
{
  prof_trace_t *trace1 = *(prof_trace_t **)p_trace1;
  prof_trace_t *trace2 = *(prof_trace_t **)p_trace2;
  return (int)(trace2->num_hits - trace1->num_hits);
}

/* output info on the cost associated with traces  */
void prof_output_trace_cost(float cutoff)
{
  prof_trace_iterate_t iterate;
  int i, trace_table_size, n_items;
  float percent;

  CALL(RawMonitorEnter)(data_access_lock);

  /* First write all trace we might refer to. */
  prof_output_unmarked_traces();

  iterate.traces = prof_calloc(prof_trace_table.n_entries * sizeof(void *));
  iterate.total_count = iterate.index = 0;
  prof_hash_iterate(&prof_trace_table, prof_trace_collect, &iterate);

  trace_table_size = iterate.index;

  /* sort all the traces according to the cost */
  qsort(iterate.traces, trace_table_size, sizeof(void *), prof_trace_compare_cost);
  n_items = 0;
  if (iterate.total_count != 0) for (i = 0; i < trace_table_size; i++) {
    prof_trace_t *trace = iterate.traces[i];
    percent = (float)trace->num_hits / (float)iterate.total_count;
    if (percent < cutoff) {
      break;
    }
    n_items++;
  }
  /* output the info */
  prof_write_header(PROF_CPU_SAMPLES, n_items * 8 + 4 + 4);
  prof_write_u4(iterate.total_count);
  prof_write_u4(n_items);
  for (i = 0; i < n_items; i++) {
    prof_trace_t *trace = iterate.traces[i];
    prof_write_u4((unsigned int)trace->num_hits);
    prof_write_u4(trace->serial_num);
  }
  prof_free(iterate.traces);

  CALL(RawMonitorExit)(data_access_lock);
}

static void * prof_clear_trace_cost_helper(void *_trace, void *_arg)
{
  prof_trace_t *trace = _trace;
  trace->num_hits = 0;
  return _arg;
}

void prof_clear_trace_cost(void)
{
  prof_hash_iterate(&prof_trace_table, prof_clear_trace_cost_helper, NULL);
}
