/*
 * Copyright (c) 2000 by Prosyst Software AG.,
 * http://www.prosyst.com
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Prosyst Software AG. You shall not disclose such Confidential
 * Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Prosyst.
 */

package com.prosyst.jpd.profiler;

import com.prosyst.util.log.Dumpable; // from Util project
import com.prosyst.util.log.Log;      // from Util project

import java.io.IOException;

/**
 * <PRE>
 *
 * header     "JAVA PROFILER AGENT 1.2" (0-terminated)
 * u4         size of identifiers. Identifiers are used to represent
 *            UTF8 strings, objects, stack traces, etc. They usually
 *            have the same size as host pointers. For example, on
 *            Solaris and Win32, the size is 4.
 * u4         high word
 * u4         low word    number of milliseconds since 0:00 GMT, 1/1/70
 */
public class BinHeader implements Dumpable {
  
  public static final String PROF_HEADER = "JAVA PROFILER AGENT";
  public static final String PROF_HEADER_VER = "JAVA PROFILER AGENT 1.2";
  public static final String HPROF_HEADER = "JAVA PROFILE 1.0.1";
  
  /**
   * Size of identifiers.
   * Identifiers are used to represent UTF8 strings, objects,
   * stack traces, etc.
   * They usually have the same size as host pointers.
   * For example, on Solaris and Win32, the size is 4.
   */
  public int sizeOfIdentifiers = 4;
  
  /**
   * Number of milliseconds since 0:00 GMT, 01/01/1970
   */
  public long millisecondsSince = 0;
  
  /**
   * Input data source for reading.
   */
  protected BinInputImpl in = null;
  
  public BinHeader(BinInputImpl in) {
    init(in);
  }
  
  public void init(BinInputImpl in) {
    this.in = in;
  }
  
  public void read() throws IOException {
    String agent_name = readHeader();
    if (!agent_name.equals(PROF_HEADER_VER)) {
      if (agent_name.startsWith(PROF_HEADER)) {
        Log.print("This is an old version of the profiler agent, please use the current version");
        throw new IOException("Old version of the profiler agent");
      }
      in.isSupported = false;
      if (agent_name.equals(HPROF_HEADER)) in.isHprofAgent = true;
      Log.printDebug("BIN_HEADER.read", agent_name);
    }
    this.sizeOfIdentifiers = in.read_u4();
    this.millisecondsSince = in.read_u8();
  }
  
  protected String readHeader() throws IOException {
    StringBuffer sb = new StringBuffer();
    int chr;
    while ((chr = in.read_u1()) != 0) {
      sb.append((char) chr);
    }
    return sb.toString();
  }
  
  // Dumpable
  
  public void dump() {
    dump(System.out);
  }
  
  public void dump(java.io.PrintStream out) {
    out.println("HEADER");
    out.println("  size of identifiers = " + sizeOfIdentifiers);
    out.println("  milliseconds since  = " + millisecondsSince);
    out.println();
  }
}