/*
 * Copyright (c) 2000 by Prosyst Software AG.,
 * http://www.prosyst.com
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Prosyst Software AG. You shall not disclose such Confidential
 * Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Prosyst.
 */

package com.prosyst.jpd.profiler;

import java.io.IOException;
import java.io.EOFException;
import java.io.DataInputStream;
import java.net.Socket;

import com.prosyst.jpd.profiler.binbody.BinRecord;
import com.prosyst.jpd.profiler.info.AllocSiteInfo;
import com.prosyst.jpd.profiler.info.CpuSampleInfo;
import com.prosyst.jpd.profiler.info.GraphInfo;
import com.prosyst.jpd.profiler.info.MonitorsDump;
import com.prosyst.jpd.profiler.storage.ProfilerNamespace;
import com.prosyst.jpd.profiler.storage.ProfilerNamespaceDefaultImpl;

import com.prosyst.util.log.Log;      // from Util project
import com.prosyst.util.log.Dumpable; // from Util project

/**
 * <PRE>
 *
 * (!) Prof Agent binary format:
 *
 * header     "JAVA PROFILER AGENT 1.2" (0-terminated)
 * u4         size of identifiers. Identifiers are used to represent
 *            UTF8 strings, objects, stack traces, etc. They usually
 *            have the same size as host pointers. For example, on
 *            Solaris and Win32, the size is 4.
 * u4         high word
 * u4         low word    number of milliseconds since 0:00 GMT, 1/1/70
 * [record]*  a sequence of records.
 *
 *
 * (!) Record format:
 *
 * u1         a TAG denoting the type of the record
 * u4         number of *microseconds* since the time stamp in the
 *            Record. (wraps around in a little more than an hour)
 * u4         number of bytes *remaining* in the record. Note that
 *            this number excludes the tag and the length field itself.
 * [u1]*      BODY of the record (a sequence of bytes)
 *
 *
 * (!) The following TAGs are supported:
 *
 * TAG           BODY      notes
 * -------------------------------------------------------------------
 * PROF_UTF8               a UTF8-encoded name
 *
 *               id         name ID
 *              [u1]*       UTF8 characters (no trailing zero)
 *
 * PROF_LOAD_CLASS         a newly loaded class
 *
 *               u4         class serial number (> 0)
 *               id         class object ID
 *               u4         stack trace serial number
 *               id         class name ID
 *
 * PROF_UNLOAD_CLASS       an unloading class
 *
 *               u4         class serial_number
 *
 * PROF_FRAME              a Java stack frame
 *
 *               id         stack frame ID
 *               id         method name ID
 *               id         method signature ID
 *               id         source file name ID
 *               u4         class serial number
 *               u4         line number >0: normal
 *                                      -1: unknown
 *                                      -2: compiled method
 *                                      -3: native method
 *
 * PROF_TRACE              a Java stack trace
 *
 *               u4         stack trace serial number
 *               u4         thread serial number
 *               u4         number of frames
 *              [id]*       stack frame IDs
 *
 * PROF_ALLOC_SITES        a set of heap allocation sites, obtained after GC
 *
 *               u2         flags 0x0001: incremental vs. complete
 *                                0x0002: sorted by allocation vs. live
 *                                0x0004: whether to force a GC
 *               u4         cutoff ratio
 *               u4         total live bytes
 *               u4         total live instances
 *               u8         total bytes allocated
 *               u8         total instances allocated
 *               u4         number of sites that follow
 *              [u1         is_array: 0:  normal object
 *                                    2:  object array
 *                                    4:  boolean array
 *                                    5:  char array
 *                                    6:  float array
 *                                    7:  double array
 *                                    8:  byte array
 *                                    9:  short array
 *                                    10: int array
 *                                    11: long array
 *               u4         class serial number (may be zero during startup)
 *               u4         stack trace serial number
 *               u4         number of bytes alive
 *               u4         number of instances alive
 *               u4         number of bytes allocated
 *               u4]*       number of instance allocated
 *
 * PROF_START_THREAD       a newly started thread.
 *
 *               u4         thread serial number (> 0)
 *               id         thread object ID
 *               u4         stack trace serial number
 *               id         thread name ID
 *               id         thread group name ID
 *               id         thread group parent name ID
 *
 * PROF_END_THREAD         a terminating thread.
 *
 *               u4         thread serial number
 *
 * PROF_HEAP_SUMMARY       heap summary
 *
 *               u4         total live bytes
 *               u4         total live instances
 *               u8         total bytes allocated
 *               u8         total instances allocated
 *
 * PROF_HEAP_DUMP          denote a heap dump
 *
 *               [heap dump sub-records]*
 *
 *                          There are four kinds of heap dump sub-records:
 *
 *               u1         sub-record type
 *
 *               PROF_GC_ROOT_UNKNOWN         unknown root
 *
 *                          id         object ID
 *
 *               PROF_GC_ROOT_THREAD_OBJ      thread object
 *
 *                          id         thread object ID  (may be 0 for a
 *                                     thread newly attached through JNI)
 *                          u4         thread sequence number
 *                          u4         stack trace sequence number
 *
 *               PROF_GC_ROOT_JNI_GLOBAL      JNI global ref root
 *
 *                          id         object ID
 *                          id         JNI global ref ID
 *
 *               PROF_GC_ROOT_JNI_LOCAL       JNI local ref
 *
 *                          id         object ID
 *                          u4         thread serial number
 *                          u4         frame # in stack trace (-1 for empty)
 *
 *               PROF_GC_ROOT_JAVA_FRAME      Java stack frame
 *
 *                          id         object ID
 *                          u4         thread serial number
 *                          u4         frame # in stack trace (-1 for empty)
 *
 *               PROF_GC_ROOT_NATIVE_STACK    Native stack
 *
 *                          id         object ID
 *                          u4         thread serial number
 *
 *               PROF_GC_ROOT_STICKY_CLASS    System class
 *
 *                          id         object ID
 *
 *               PROF_GC_ROOT_THREAD_BLOCK    Reference from thread block
 *
 *                          id         object ID
 *                          u4         thread serial number
 *
 *               PROF_GC_ROOT_MONITOR_USED    Busy monitor
 *
 *                          id         object ID
 *
 *               PROF_GC_CLASS_DUMP           dump of a class object
 *
 *                          id         class object ID
 *                          u4         stack trace serial number
 *                          id         super class object ID
 *                          id         class loader object ID
 *                          id         signers object ID
 *                          id         protection domain object ID
 *                          id         reserved
 *                          id         reserved
 *
 *                          u4         instance size (in bytes)
 *
 *                          u2         size of constant pool
 *                         [u2,        constant pool index,
 *                          ty,        type
 *                                       2:  object
 *                                       4:  boolean
 *                                       5:  char
 *                                       6:  float
 *                                       7:  double
 *                                       8:  byte
 *                                       9:  short
 *                                       10: int
 *                                       11: long
 *                          vl]*       and value
 *
 *                          u2         number of static fields
 *                         [id,        static field name,
 *                          ty,        type,
 *                          vl]*       and value
 *
 *                          u2         number of inst. fields (not inc. super)
 *                         [id,        instance field name,
 *                          ty]*       type
 *
 *               PROF_GC_INSTANCE_DUMP        dump of a normal object
 *
 *                          id         object ID
 *                          u4         stack trace serial number
 *                          id         class object ID
 *                          u4         number of bytes that follow
 *                         [vl]*       instance field values (class, followed
 *                                     by super, super's super ...)
 *
 *               PROF_GC_OBJ_ARRAY_DUMP       dump of an object array
 *
 *                          id         array object ID
 *                          u4         stack trace serial number
 *                          u4         number of elements
 *                          id         element class ID
 *                         [id]*       elements
 *
 *               PROF_GC_PRIM_ARRAY_DUMP      dump of a primitive array
 *
 *                          id         array object ID
 *                          u4         stack trace serial number
 *                          u4         number of elements
 *                          u1         element type
 *                                       4:  boolean array
 *                                       5:  char array
 *                                       6:  float array
 *                                       7:  double array
 *                                       8:  byte array
 *                                       9:  short array
 *                                       10: int array
 *                                       11: long array
 *                         [u1]*       elements
 *
 * PROF_CPU_SAMPLES        a set of sample traces of running threads
 *
 *                u4        total number of samples
 *                u4        # of traces
 *               [u4        # of samples
 *                u4]*      stack trace serial number
 *
 * PROF_CONTROL_SETTINGS   the settings of on/off switches
 *
 *                u4        0x00000001: alloc traces on/off
 *                          0x00000002: cpu sampling on/off
 *                u2        stack trace depth
 *
 */
public class BinInputImpl extends ProfilerInput implements BinInput, ProfilerListener {
  
  protected ProfilerNamespace namespace = new ProfilerNamespaceDefaultImpl();
  
  protected BinHeader header = new BinHeader(this);
  
  public boolean isSupported  = true;
  public boolean isHprofAgent = false;
  public boolean allocTraces  = false;
  public boolean cpuSampling  = false;
  public int stackTraceDepth  = 3;
  public int sizeofID         = 4;
  
  private ProfilerListener listener = null;
  
  private boolean dump_bin_input = false;
  private boolean dump_namespace = false;
  
  public BinInputImpl(DataInputStream in, ProfilerListener listener) {
    super(in);
    this.listener = listener;
  }
  
  public BinInputImpl(Socket socket, ProfilerListener listener) throws IOException {
    super(socket);
    this.listener = listener;
  }
  
  public void enableLogging(boolean dump_bin_input, boolean dump_namespace) {
    this.dump_bin_input = dump_bin_input;
    this.dump_namespace = dump_namespace;
  }
  
  public ProfilerNamespace getNamespace() {
    return namespace;
  }
  
  // ProfilerListener implementation
  
  public void profilerStarted() {
    if (listener != null) listener.profilerStarted();
  }
  
  public void profilerSupport(boolean isSupported, boolean heapSupport) {
    if (listener != null) listener.profilerSupport(isSupported, heapSupport);
  }
  
  public void profilerStopped() {
    if (listener != null) listener.profilerStopped();
  }
  
  public void profilerError() {
    if (listener != null) listener.profilerError();
  }
  
  public void profilerErrorMsg(Throwable ex) {
    if (listener != null) listener.profilerErrorMsg(ex);
  }
  
  public void eventAllocSites(AllocSiteInfo[] allocSites) {
    if (listener != null) listener.eventAllocSites(allocSites);
  }
  
  public void eventCpuSamples(int ticks, int totalNumberOfSamples, CpuSampleInfo[] cpuSamples) {
    if (listener != null) listener.eventCpuSamples(ticks, totalNumberOfSamples, cpuSamples);
  }
  
  public void eventHeapDump(ProfilerNamespace namespace) {
    if (listener != null) listener.eventHeapDump(namespace);
  }
  
  public void eventGraphInfo(GraphInfo graphInfo) {
    if (listener != null) listener.eventGraphInfo(graphInfo);
  }
  
  public void eventMonitorsDump(MonitorsDump monitorsDump) {
    if (listener != null) listener.eventMonitorsDump(monitorsDump);
  }
  
  public void eventCommon(Object event) {
    if (listener != null) listener.eventCommon(event);
  }
  
  // BinInput implementation
  
  byte[] bytes = new byte[2048];
  int[]  ids   = new int[128];
  
  public int sizeofID() {
    return sizeofID;
  }
  
  public void skip(int length) throws IOException {
    in.skipBytes(length);
  }
  
  public void skip(int length, boolean warn) throws IOException {
    if (length < 0) throw new IOException("Negative skip bytes size");
    if (warn) Log.printDebug("BIN_INPUT.skip", length + " bytes skipped");
    in.skipBytes(length);
  }
  
  public byte read_u1() throws IOException {
    return in.readByte();
  }
  
  public short read_u2() throws IOException {
    return in.readShort();
  }
  
  public int read_u4() throws IOException {
    return in.readInt();
  }
  
  public long read_u8() throws IOException {
    return in.readLong();
  }
  
  public int read_id() throws IOException {
    if (sizeofID == 4) {
      return in.readInt();
    } else if (sizeofID == 2) {
      return in.readShort();
    } else {
      throw new IOException("Illegal identifier size: " + sizeofID);
    }
  }
  
  public int[] read_ids(int length) throws IOException {
    if ((length * sizeofID) > bytes.length) {
      bytes = new byte[length * sizeofID];
    }
    if (length > ids.length) {
      ids = new int[length];
    }
    in.readFully(bytes, 0, length * sizeofID);
    if (sizeofID == 4) {
      for (int i = 0, j = 0; i < length; i++) {
        ids[i] = (bytes[j++] << 24) + ((bytes[j++] & 0xff) << 16)
               + ((bytes[j++] & 0xff) << 8) + (bytes[j++] & 0xff);
      }
    } else if (sizeofID == 2) {
      for (int i = 0, j = 0; i < length; i++) {
        ids[i] = (bytes[j++] << 8) | (bytes[j++] & 0xff);
      }
    } else {
      throw new IOException("Illegal identifier size: " + sizeofID);
    }
    return ids;
  }
  
  public boolean read_boolean() throws IOException {
    return in.readBoolean();
  }
  
  public char read_char() throws IOException {
    return in.readChar();
  }
  
  public float read_float() throws IOException {
    return in.readFloat();
  }
  
  public double read_double() throws IOException {
    return in.readDouble();
  }
  
  public byte read_byte() throws IOException {
    return in.readByte();
  }
  
  public short read_short() throws IOException {
    return in.readShort();
  }
  
  public int read_int() throws IOException {
    return in.readInt();
  }
  
  public long read_long() throws IOException {
    return in.readLong();
  }
  
  public byte[] read_bytes(int length) throws IOException {
    byte[] bytes = new byte[length];
    in.readFully(bytes, 0, length);
    return bytes;
  }
  
  public void read_bytes(byte[] bytes, int length) throws IOException {
    in.readFully(bytes, 0, length);
  }
  
  public String read_utf8(int length) throws IOException {
    if (length > bytes.length) {
      bytes = new byte[length];
    }
    in.readFully(bytes, 0, length);
    //return new String(bytes, 0); // for speeding the conversion from bytes to chars !!
    return new String(bytes, 0, length, "UTF8"); // use unicode !!
  }
  
  //  Runnable implementation
  
  public void run() {
    profilerStarted();
    try {
      do_prof_header();
      profilerSupport(isSupported, isSupported || isHprofAgent);
      do_prof_records();
    } catch (Throwable ex) {
      if (!(ex instanceof java.net.SocketException
            && (ex.getLocalizedMessage().startsWith("Connection reset by peer")
             || ex.getLocalizedMessage().startsWith("Connection reset")))) {
        Log.printDebug("BIN_INPUT.run", ex);
        profilerError();
        profilerErrorMsg(ex);
      }
    }
    close();
    profilerStopped();
    
    if (dump_namespace) {
      if (namespace instanceof Dumpable) {
        ((Dumpable) namespace).dump();
      }
    }
  }
  
  protected void do_prof_header() throws IOException {
    header.read();
    sizeofID = header.sizeOfIdentifiers;
    if (dump_bin_input) {
      header.dump();
    }
  }
  
  protected void do_prof_records() throws IOException {
    record = new BinRecord(this);
    for (;;) {
      try {
        do_prof_record();
      } catch (EOFException _) {
        break;
      }
    }
  }
  
  private BinRecord record;
  
  protected void do_prof_record() throws IOException {
    record.read();
    if (dump_bin_input) {
      record.dump();
      record.getBody().dump();
    }
  }
}