/*
 * Copyright (c) 2000 by Prosyst Software AG.,
 * http://www.prosyst.com
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Prosyst Software AG. You shall not disclose such Confidential
 * Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Prosyst.
 */

package com.prosyst.jpd.profiler;

import java.io.IOException;
import java.io.DataOutputStream;
import java.io.BufferedOutputStream;
import java.net.Socket;

/**
 * @author Vladislav Doshev (VladiSD)
 */
public class BinOutput {
  
  public static final int PROF_NULL_ID = 0x0000;
  
  // command tags
  
  public static final int PROF_CMD_GC           = 0x01;
  public static final int PROF_CMD_DUMP_HEAP    = 0x02;
  public static final int PROF_CMD_ALLOC_SITES  = 0x03;
  public static final int PROF_CMD_HEAP_SUMMARY = 0x04;
  public static final int PROF_CMD_EXIT         = 0x05;
  public static final int PROF_CMD_DUMP_TRACES  = 0x06;
  public static final int PROF_CMD_CPU_SAMPLES  = 0x07;
  public static final int PROF_CMD_CONTROL      = 0x08;
  
  // our tags
  
  public static final int PROF_CMD_SUSPEND_ALL    = 0x40;
  public static final int PROF_CMD_RESUME_ALL     = 0x41;
  public static final int PROF_CMD_GRAPH_INFO     = 0x42;
  public static final int PROF_CMD_DUMP_MONITORS  = 0x43;
  public static final int PROF_CMD_DUMP_ON_EXIT   = 0x44;
  public static final int PROF_CMD_MONITOR_EVENTS = 0x45;
  
  public static final int PROF_CMD_SET_TRIGGER    = 0x50;
  
  // flags used in output of allocation sites
  
  public static final int PROF_SITE_DUMP_INCREMENTAL = 0x01;
  public static final int PROF_SITE_SORT_BY_ALLOC    = 0x02;
  public static final int PROF_SITE_FORCE_GC         = 0x04;
  
  // flags used in output of cpu samples
  
  public static final int PROF_CPU_NONE              = 0x00;
  
  protected DataOutputStream out = null;
  protected Socket socket = null;
  
  public BinOutput(DataOutputStream out) {
    this.out = out;
  }
  
  public BinOutput(Socket socket) throws IOException {
    this(new DataOutputStream(new BufferedOutputStream(socket.getOutputStream())));
    this.socket = socket;
  }
  
  public void close() {
    // close output stream and clear reference
    if (out != null) {
      try {
        out.close();
      } catch (Exception _) { }
      out = null;
    }
    // close socket and clear reference
    if (socket != null) {
      try {
        socket.close();
      } catch (Exception _) { }
      socket = null;
    }
  }
  
  public void write_u1(int u1) throws IOException {
    out.writeByte(u1);
  }
  
  public void write_u2(int u2) throws IOException {
    out.writeShort(u2);
  }
  
  public void write_u4(int u4) throws IOException {
    out.writeInt(u4);
  }
  
  public void write_u8(long u8) throws IOException {
    throw new IOException("write_u8: Not implemented, yet");
  }
  
  public void write_id(int id) throws IOException {
    out.writeInt(id);
  }
  
  public void write_ids(int[] ids) throws IOException {
    for (int i = 0; i < ids.length; i++) {
      write_id(ids[i]);
    }
  }
  
  public void write_utf8(String utf8) throws IOException {
    throw new IOException("write_utf8: Not implemented, yet");
  }
  
  protected void write_ready() throws IOException {
    out.flush();
  }
  
  /**
   * The client can send the VM a set of commands.
   * The commands have the following format:
   *
   * u1         a TAG denoting the type of the record
   * u4         a serial number
   * u4         number of bytes *remaining* in the record. Note that
   *            this number excludes the tag and the length field itself.
   * [u1]*      BODY of the record (a sequence of bytes)
   */
  protected void write_command(int tag, int length) throws IOException {
    write_u1(tag);
    write_u4(++serialNumber);
    write_u4(length);
  }
  
  protected void write_command(int tag) throws IOException {
    write_command(tag, 0);
  }
  
  protected static int serialNumber = 0;
  
  /**
   * <PRE>
   * PROF_CMD_GC             force a GC.
   */
  public void command_GC() throws IOException {
    write_command(PROF_CMD_GC);
    write_ready();
  }
  
  /**
   * <PRE>
   * PROF_CMD_DUMP_HEAP      obtain a heap dump
   */
  public void command_DumpHeap() throws IOException {
    write_command(PROF_CMD_DUMP_HEAP);
    write_ready();
  }
  
  /**
   * <PRE>
   * PROF_CMD_ALLOC_SITES    obtain allocation sites
   *
   *               u2         flags 0x0001: incremental vs. complete
   *                                0x0002: sorted by allocation vs. live
   *                                0x0004: whether to force a GC
   *               u4         cutoff ratio (0.0 ~ 1.0)
   */
  public void command_AllocSites(int flags, int cutoffRatio) throws IOException {
    write_command(PROF_CMD_ALLOC_SITES, 2 + 4);
    write_u2(flags);
    write_u4(cutoffRatio);
    write_ready();
  }
  
  public void command_AllocSites(int flags, float cutoffRatio) throws IOException {
    command_AllocSites(flags, Float.floatToIntBits(cutoffRatio));
  }
  
  /**
   * <PRE>
   * PROF_CMD_HEAP_SUMMARY   obtain heap summary
   */
  public void command_HeapSummary() throws IOException {
    write_command(PROF_CMD_HEAP_SUMMARY);
    write_ready();
  }
  
  /**
   * <PRE>
   * PROF_CMD_EXIT
   */
  public void command_Exit() throws IOException {
    write_command(PROF_CMD_EXIT);
    write_ready();
  }
  
  /**
   * <PRE>
   * PROF_CMD_DUMP_TRACES    obtain all newly created traces
   */
  public void command_DumpTraces() {
    try {
      write_command(PROF_CMD_DUMP_TRACES);
      write_ready();
    } catch (IOException _) { }
  }
  
  /**
   * <PRE>
   * PROF_CMD_CPU_SAMPLES    obtain a PROF_CPU_SAMPLES record
   *
   *               u2         ignored for now
   *               u4         cutoff ratio (0.0 ~ 1.0)
   */
  public void command_CpuSamples(int flags, int ratio) throws IOException {
    write_command(PROF_CMD_CPU_SAMPLES, 2 + 4);
    write_u2(flags);
    write_u4(ratio);
    write_ready();
  }
  
  public void command_CpuSamples(int flags, float ratio) throws IOException {
    command_CpuSamples(flags, Float.floatToIntBits(ratio));
  }
  
  /**
   * <PRE>
   * PROF_CMD_CONTROL        changing settings
   *
   *               u2         0x0001: alloc traces on
   *                          0x0002: alloc traces off
   *
   *                          0x0003: CPU sampling on
   *
   *                                  id:   thread object id (NULL for all)
   *
   *                          0x0004: CPU sampling off
   *
   *                                  id:   thread object id (NULL for all)
   *
   *                          0x0005: clear CPU sampling info
   *
   *                          0x0006: clear alloc sites info
   *
   *                          0x0007: set max stack depth in CPU samples
   *                                  and alloc traces
   *
   *                                  u2:   new depth
   */
  public void command_SetAllocTraces(boolean traces) throws IOException {
    write_command(PROF_CMD_CONTROL, 2);
    write_u2(traces ? 0x0001 : 0x0002);
    write_ready();
  }
  
  public void command_SetCpuSampling(boolean sampling, int threadID) throws IOException {
    write_command(PROF_CMD_CONTROL, 2 + 4);
    write_u2(sampling ? 0x0003 : 0x0004);
    write_id(threadID);
    write_ready();
  }
  
  public void command_SetCpuSamplingAll(boolean sampling) throws IOException {
    command_SetCpuSampling(sampling, PROF_NULL_ID);
  }
  
  public void command_ClearCpuSampling() throws IOException {
    write_command(PROF_CMD_CONTROL, 2);
    write_u2(0x0005);
    write_ready();
  }
  
  public void command_ClearAllocSitesInfo() throws IOException {
    write_command(PROF_CMD_CONTROL, 2);
    write_u2(0x0006);
    write_ready();
  }
  
  public void command_SetMaxStackDepth(int depth) throws IOException {
    write_command(PROF_CMD_CONTROL, 4);
    write_u2(0x0007);
    write_u2(depth);
    write_ready();
  }
  
  /**
   * <PRE>
   * PROF_CMD_SUSPEND_ALL
   */
  public void command_Suspend() throws IOException {
    write_command(PROF_CMD_SUSPEND_ALL);
    write_ready();
  }
  
  /**
   * <PRE>
   * PROF_CMD_RESUME_ALL
   */
  public void command_Resume() throws IOException {
    write_command(PROF_CMD_RESUME_ALL);
    write_ready();
  }
  
  /**
   * <PRE>
   * PROF_CMD_GRAPH_INFO
   */
  public void command_GraphInfo() throws IOException {
    write_command(PROF_CMD_GRAPH_INFO);
    write_ready();
  }
  
  /**
   * <PRE>
   * PROF_CMD_DUMP_MONITORS
   */
  public void command_DumpMonitors() throws IOException {
    write_command(PROF_CMD_DUMP_MONITORS);
    write_ready();
  }
  
  /**
   * <PRE>
   * PROF_CMD_DUMP_ON_EXIT
   */
  public void command_DumpOnExit() throws IOException {
    write_command(PROF_CMD_DUMP_ON_EXIT);
    write_ready();
  }
  
  /**
   * <PRE>
   * PROF_CMD_MONITOR_EVENTS
   */
  public void command_MonitorEvents(boolean enable) throws IOException {
    write_command(PROF_CMD_MONITOR_EVENTS, 2);
    write_u2(enable ? 0x0001 : 0x0000);
    write_ready();
  }
  
  /**
   * <PRE>
   * PROF_CMD_SET_TRIGGER
   */
  public void command_SetTrigger(int type, int action, String className, String methodName) throws IOException {
    if (methodName == null) methodName = "";
    className = className.replace('.', '/');
    byte[] classBytes  = className.getBytes("UTF8");
    byte[] methodBytes = methodName.getBytes("UTF8");
    int clen = classBytes.length;
    int mlen = methodBytes.length;
    write_command(PROF_CMD_SET_TRIGGER, 5 + clen + mlen);
    write_u1(type);
    write_u1(action);
    write_u2(clen);
    out.write(classBytes);
    write_u2(mlen);
    out.write(methodBytes);
    write_ready();
  }
}