/*
 * Copyright (c) 2000 by Prosyst Software AG.,
 * http://www.prosyst.com
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Prosyst Software AG. You shall not disclose such Confidential
 * Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Prosyst.
 */

package com.prosyst.jpd.profiler;

import java.io.BufferedReader;
import java.io.File;
import java.io.InputStreamReader;

import com.prosyst.util.exec.Env;
import com.prosyst.util.exec.EnvException;
import com.prosyst.util.exec.Executor;
import com.prosyst.util.string.CommandLine;
import com.prosyst.util.log.Log;

public class Parser {
  
  public static final String OS_NAME = System.getProperty("os.name");
  private static final boolean dump_ver = Boolean.getBoolean("dump.java.ver");
  
  public static final int VER_UNKNOWN     = -1;
  public static final int VER_UNSUPPORTED = -2;
  public static final int VER_J9          = -3;
  public static final int VER_JEODE       = -4;
  public static final int VER_CVM         = -5;
  public static final int VER_SUNW130     = -10;
  
  private static final String JAVA_VERSION  = "java version \"1.";
  private static final String JEODE_VERSION = "Jeode EVM";
  private static final String SOL_122       = "Solaris VM (build Solaris_JDK_1.2.2";
  private static final String SOL_130       = "(build 1.3.0";
  
  public static int checkJavaVersion(String javaPath) {
    if (javaPath != null) {
      if (javaPath.endsWith("j9.exe") || javaPath.endsWith("j9")) {
        if (dump_ver) Log.printDebug("Java version: j9");
        return VER_J9;
      }
      if (javaPath.endsWith("evm.exe") || javaPath.endsWith("evm")) {
        if (dump_ver) Log.printDebug("Java version: Jeode");
        return VER_JEODE;
      }
      if (javaPath.endsWith("cvm.exe") || javaPath.endsWith("cvm")) {
        if (dump_ver) Log.printDebug("Java version: CVM");
        return VER_CVM;
      }
      try {
        Process javaPrc = Runtime.getRuntime().exec(javaPath + " -version");
        BufferedReader in = new BufferedReader(new InputStreamReader(javaPrc.getErrorStream()));
        String line = in.readLine();
        String subline = in.readLine();
        //while (in.readLine() != null);
        javaPrc.destroy();
        if (line != null && line.startsWith(JEODE_VERSION)) {
          if (dump_ver) Log.printDebug("Java version: Jeode");
          return VER_JEODE;
        }
        if (line != null && line.startsWith(JAVA_VERSION)) {
          if (subline != null) {
            if (OS_NAME.startsWith("SunOS") || OS_NAME.startsWith("Solaris")) {
              if (subline.indexOf(SOL_122) != -1 || subline.indexOf(SOL_130) != -1) {
                if (dump_ver) Log.printDebug("Java version: unsupported");
                return VER_UNSUPPORTED;
              }
            }
          }
          int jver = line.charAt(JAVA_VERSION.length()) - '0';
          if (jver == 3 && line.indexOf("1.3.0") != -1 && OS_NAME.startsWith("Windows")) jver = VER_SUNW130;
          if (dump_ver) Log.printDebug("Java version: " + jver);
          return jver;
        }
      } catch (Throwable ex) { }
    }
    if (dump_ver) Log.printDebug("Java version: unknown");
    return VER_UNKNOWN;
  }
  
  public static String parseCommandWithClassic(String profCmd, int ver) {
    if (ver == VER_SUNW130) profCmd = "-classic " + profCmd; // classic fix !!
    return parseCommand(profCmd, ver);
  }
  
  public static String parseCommand(String profCmd, int ver) {
    if (ver == VER_JEODE) profCmd = "-Xprofile " + profCmd;
    if (ver == VER_CVM) profCmd = "-Dsun.boot.library.path=" + System.getProperty("user.dir") + " " + profCmd;
    
    int index = profCmd.indexOf("-Xrunprof:");
    if (index != -1) {
      if (OS_NAME.startsWith("SunOS") || OS_NAME.startsWith("Solaris")) {
        profCmd = profCmd.substring(0, index + 9) + 's' + profCmd.substring(index + 9);
      } else if (OS_NAME.startsWith("Linux")) {
        profCmd = profCmd.substring(0, index + 9) + 'l' + profCmd.substring(index + 9);
      }
    }
    return profCmd;
  }

  /** Runs process through Executor or Runtime and sets environment variables. */
  public static Process runProcess(String proflibDir, String workDir, String command) throws Exception {
    Process process;
    if (workDir != null && workDir.length() > 0) {
      String[] env = null;
      if (OS_NAME.startsWith("Windows")) {
        env = getWinProperties(proflibDir);
      } else if (OS_NAME.startsWith("SunOS") || OS_NAME.startsWith("Solaris")) {
        env = getUnixProperties(proflibDir);
      } else if (OS_NAME.startsWith("Linux")) {
        env = getUnixProperties(proflibDir);
      }
      process = Executor.exec(workDir, command, env);
    } else {
      process = Runtime.getRuntime().exec(CommandLine.createArgs(command));
    }
    return process;
  }
  
  private static final String WINPATH = "PATH=";
  
  public static String[] getWinProperties(String proflibDir) throws EnvException {
    String[] env = Env.getProperties();
    for (int i = 0; i < env.length; i++) {
      if (env[i].toUpperCase().startsWith(WINPATH)) {
        env[i] = WINPATH + proflibDir
               + File.pathSeparator + env[i].substring(WINPATH.length());
        return env;
      }
    }
    String[] env_new = new String[env.length + 1];
    System.arraycopy(env, 0, env_new, 0, env.length);
    env_new[env.length] = WINPATH + proflibDir;
    return env_new;
  }
  
  private static final String LDLIBPATH = "LD_LIBRARY_PATH=";
  
  public static String[] getUnixProperties(String proflibDir) throws EnvException {
    String[] env = Env.getProperties();
    for (int i = 0; i < env.length; i++) {
      if (env[i].toUpperCase().startsWith(LDLIBPATH)) {
        env[i] = LDLIBPATH + proflibDir
               + File.pathSeparator + env[i].substring(LDLIBPATH.length());
        return env;
      }
    }
    String[] env_new = new String[env.length + 1];
    System.arraycopy(env, 0, env_new, 0, env.length);
    env_new[env.length] = LDLIBPATH + proflibDir;
    return env_new;
  }
}
