/*
 * Copyright (c) 2000 by Prosyst Software AG.,
 * http://www.prosyst.com
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Prosyst Software AG. You shall not disclose such Confidential
 * Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Prosyst.
 */

package com.prosyst.jpd.profiler;

import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;

import com.prosyst.jpd.profiler.gui.AllocSamplerImpl;
import com.prosyst.jpd.profiler.gui.CpuSamplerImpl;
import com.prosyst.jpd.profiler.info.AllocSiteInfo;
import com.prosyst.jpd.profiler.info.CpuSampleInfo;
import com.prosyst.jpd.profiler.info.GraphInfo;
import com.prosyst.jpd.profiler.info.MonitorsDump;
import com.prosyst.jpd.profiler.storage.ProfilerNamespace;

import com.prosyst.util.log.Log;      // from Util project

/**
 * Creates input and output; registers listener and starts server.
 */
public class ProfilerServer implements ProfilerListener {
  
  /** default server port */
  public static final int DEFAULT_PORT   = 5115;
  /** server socket timeout */
  public static final int SERVER_TIMEOUT = 5000;
  
  /** ignores input data */
  public static final int MODE_NONE    = 0;
  /** ascii input mode, input data is dumped to System.out */
  public static final int MODE_ASCII   = 1;
  /** binary input mode */
  public static final int MODE_BINARY  = 2;
  
  private boolean dump_bin_input = Boolean.getBoolean("dump.bin.input");
  private boolean dump_namespace = Boolean.getBoolean("dump.namespace");
  
  private int agentPort = DEFAULT_PORT;
  private int inputMode = MODE_ASCII;
  
  private ProfilerListener listener;
  
  private ProfilerInput input;
  private BinOutput     output;
  
  private AllocSamplerImpl allocSampler;
  private CpuSamplerImpl   cpuSampler;
  
  /**
   * Constructs a new profile server, waiting on the specified port,
   * with ascii input mode.
   *
   * @param   port      the port number
   */
  public ProfilerServer(int port) {
    this(port, MODE_ASCII);
  }
  
  /**
   * Constructs a new profile server, waiting on the specified port,
   * with binary input mode and registers the listener.
   *
   * @param   port      the port number
   * @param   listener  that receives profiler events
   */
  public ProfilerServer(ProfilerListener listener) {
    this(DEFAULT_PORT, MODE_BINARY);
    this.listener = listener;
  }
  
  /**
   * Constructs a new profile server, waiting on the specified port,
   * with the specified input mode.
   *
   * @param   port      the port number
   * @param   mode      ascii, binary, or none
   */
  public ProfilerServer(int port, int mode) {
    this.agentPort = port;
    this.inputMode = mode;
  }
  
  /**
   * Creates profile input and output depending on the mode,
   * starts the server and waits for connection from the client.
   *
   * @exception   IOException  if an I/O error occurs when opening the socket
   */
  public void start() throws IOException {
    start(agentPort);
  }
  
  /**
   * Creates profile input and output depending on the mode,
   * starts the server and waits for connection from the client.
   *
   * @param       port         agentPort
   * @exception   IOException  if an I/O error occurs when opening the socket
   */
  public void start(int port) throws IOException {
    start(port, SERVER_TIMEOUT);
  }
  
  ServerSocket ss;
  
  /**
   * Creates profile input and output depending on the mode,
   * starts the server and waits for connection from the client.
   *
   * @param       port         agentPort
   * @param       timeout      server socket timeout
   * @exception   IOException  if an I/O error occurs when opening the socket
   */
  public void start(int port, int timeout) throws IOException {
    ss = new ServerSocket(port);
    if (timeout > 0) ss.setSoTimeout(timeout);
    Socket agentSocket;
    try {
      agentSocket = ss.accept();
    } finally {
      ss.close();
      ss = null;
    }
    
    output = createOutput(agentSocket);
    input  = createInput(agentSocket);
    
    if (input != null) {
      input.setPriority(Thread.NORM_PRIORITY);
      input.start();
    }
  }
  
  /**
   * Creates profile input and output depending on the mode,
   * and connects to the client.
   *
   * @param       host         agentHost
   * @param       port         agentPort
   * @exception   IOException  if an I/O error occurs when opening the socket
   */
  public void start(String host, int port) throws IOException {
    Socket agentSocket = new Socket(host, port);
    
    output = createOutput(agentSocket);
    input  = createInput(agentSocket);
    
    if (input != null) {
      input.setPriority(Thread.NORM_PRIORITY - 1);
      input.start();
    }
  }
  
  /**
   * Gets the output to the profiling client.
   * Used to execute commands like 'run garbage collector',
   * 'obtain alloc sites', 'obtain cpu samples', etc.
   *
   * @return     the profiler output
   */
  public BinOutput getOutput() {
    return output;
  }
  
  /**
   * Gets the namespace of the profiling client.
   *
   * @return     the profiler namespace
   */
  public ProfilerNamespace getNamespace() {
    if (input != null && input instanceof BinInputImpl) return ((BinInputImpl) input).getNamespace();
    return null;
  }
  
  /**
   * Closes the input from the profiling client.
   */
  public boolean shutDown() {
    try {
      if (ss     != null) ss.close();
      if (output != null) {
        output.command_Exit();
        return true;
      }
    } catch (IOException ex) {
      if (!(ex instanceof java.net.SocketException
            && (ex.getLocalizedMessage().startsWith("Connection reset by peer")
             || ex.getLocalizedMessage().startsWith("Connection reset")))) {
        Log.printDebug("SERVER.shutDown", ex);
      }
    }
    return false;
  }
  
  /**
   * Clears the input and output.
   */
  public void clear() {
    input  = null;
    output = null;
  }
  
  private ProfilerInput createInput(Socket agentSocket) throws IOException {
    switch (inputMode) {
      case MODE_BINARY:
        if (listener == null) {
          allocSampler = new AllocSamplerImpl();
          cpuSampler   = new CpuSamplerImpl();
          listener     = this;
        }
        BinInputImpl binInput = new BinInputImpl(agentSocket, listener);
        binInput.enableLogging(dump_bin_input, dump_namespace);
        return binInput;
      case MODE_ASCII:
        return new ProfilerInputAscii(agentSocket);
      case MODE_NONE:
      default:
        return new ProfilerInputNone(agentSocket);
    }
  }
  
  private BinOutput createOutput(Socket agentSocket) throws IOException {
    switch (inputMode) {
      case MODE_BINARY:
        return new BinOutput(agentSocket);
      case MODE_ASCII:
      case MODE_NONE:
      default:
        return null;
    }
  }
  
  //
  //  ProfilerListener
  //
  
  /** Event meaning that profiler has started. */
  public void profilerStarted() {
  }
  
  /** Event with profiler support capability. */
  public void profilerSupport(boolean isSupported, boolean heapSupport) {
  }
  
  /** Event meaning that profiler has stopped. */
  public void profilerStopped() {
    allocSampler.showAllocSites();
    cpuSampler.showCpuSamples();
  }
  
  /** Event meaning that profiler error has occurred. */
  public void profilerError() {
    shutDown();
  }
  
  /** Event meaning that profiler error has occurred. */
  public void profilerErrorMsg(Throwable ex) {
  }
  
  /**
   * Event meaning that alloc sites info has arrived.
   *
   * @param   allocSites  array of info containing allocated bytes and insts
   */
  public void eventAllocSites(AllocSiteInfo[] allocSites) {
    allocSampler.eventAllocSites(allocSites);
  }
  
  /**
   * Event meaning that cpu samples info has arrived.
   *
   * @param   ticks                 the time in ms of the event
   * @param   totalNumberOfSamples  total number of samples after the preceding cpu samples event
   * @param   cpuSamples            array of info containing cost and trace
   */
  public void eventCpuSamples(int ticks, int totalNumberOfSamples, CpuSampleInfo[] cpuSamples) {
    cpuSampler.eventCpuSamples(ticks, totalNumberOfSamples, cpuSamples);
  }
  
  /**
   * Event meaning that heap dump info has arrived.
   *
   * @param   namespace            namespace with all heap classes and objects
   */
  public void eventHeapDump(ProfilerNamespace namespace) {
    // Never happens
  }
  
  /**
   * Event meaning that graph info has arrived.
   *
   * @param   graphInfo  info containing live and alloc bytes and instances, thread, class and gc info
   */
  public void eventGraphInfo(GraphInfo graphInfo) {
    // Never happens
  }
  
  /**
   * Event meaning that monitors dump has arrived.
   *
   * @param   monitorsDump  info containing stack traces of live threads and acquired monitors
   */
  public void eventMonitorsDump(MonitorsDump monitorsDump) {
    // Never happens
  }
  
  /**
   * Common event notification.
   *
   * @param   event  object containing event information
   */
  public void eventCommon(Object event) {
    // Never happens
  }
  
  public static void main(String[] args) throws Exception {
    ProfilerServer profilerServer =
               new ProfilerServer(parsePort(args, 0),
                                  parseMode(args, 1));
    profilerServer.start(profilerServer.agentPort, 0);
  }
  
  private static int parsePort(String args[], int i) {
    if (args.length > i) {
      return Integer.parseInt(args[i]);
    }
    return DEFAULT_PORT;
  }
  
  private static int parseMode(String args[], int i) {
    if (args.length > i) {
      String s = args[i].trim();
      if (s.length() >= 1) {
        switch (s.charAt(0)) {
          case 'A': case 'a':
            return MODE_ASCII;
          case 'B': case 'b':
            return MODE_BINARY;
          case 'N': case 'n':
            return MODE_NONE;
        }
      }
    }
    return MODE_BINARY; // default mode
  }
}