/*
 * Copyright (c) 2000 by Prosyst Software AG.,
 * http://www.prosyst.com
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Prosyst Software AG. You shall not disclose such Confidential
 * Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Prosyst.
 */

package com.prosyst.jpd.profiler.binbody;

import java.io.IOException;

import com.prosyst.jpd.profiler.heapdump.*;
import com.prosyst.jpd.profiler.info.ClassDumpInfo;
import com.prosyst.jpd.profiler.info.ObjectDumpInfo;
import com.prosyst.jpd.profiler.info.ObjArrayDumpInfo;
import com.prosyst.jpd.profiler.info.PrimArrayDumpInfo;

import com.prosyst.util.log.Log;

/**
 * <PRE>
 * TAG           BODY      notes
 * ------------------------------------------
 * PROF_HEAP_DUMP          denote a heap dump
 *
 *               [heap dump sub-records]*
 *
 *                          There are four kinds of heap dump sub-records:
 *
 *               u1         sub-record type
 *
 *               PROF_GC_ROOT_UNKNOWN         unknown root
 *
 *                          id         object ID
 *
 *               PROF_GC_ROOT_THREAD_OBJ      thread object
 *
 *                          id         thread object ID  (may be 0 for a
 *                                     thread newly attached through JNI)
 *                          u4         thread sequence number
 *                          u4         stack trace sequence number
 *
 *               PROF_GC_ROOT_JNI_GLOBAL      JNI global ref root
 *
 *                          id         object ID
 *                          id         JNI global ref ID
 *
 *               PROF_GC_ROOT_JNI_LOCAL       JNI local ref
 *
 *                          id         object ID
 *                          u4         thread serial number
 *                          u4         frame # in stack trace (-1 for empty)
 *
 *               PROF_GC_ROOT_JAVA_FRAME      Java stack frame
 *
 *                          id         object ID
 *                          u4         thread serial number
 *                          u4         frame # in stack trace (-1 for empty)
 *
 *               PROF_GC_ROOT_NATIVE_STACK    Native stack
 *
 *                          id         object ID
 *                          u4         thread serial number
 *
 *               PROF_GC_ROOT_STICKY_CLASS    System class
 *
 *                          id         object ID
 *
 *               PROF_GC_ROOT_THREAD_BLOCK    Reference from thread block
 *
 *                          id         object ID
 *                          u4         thread serial number
 *
 *               PROF_GC_ROOT_MONITOR_USED    Busy monitor
 *
 *                          id         object ID
 *
 *               PROF_GC_CLASS_DUMP           dump of a class object
 *
 *                          id         class object ID
 *                          u4         stack trace serial number
 *                          id         super class object ID
 *                          id         class loader object ID
 *                          id         signers object ID
 *                          id         protection domain object ID
 *                          id         reserved
 *                          id         reserved
 *
 *                          u4         instance size (in bytes)
 *
 *                          u2         size of constant pool
 *                         [u2,        constant pool index,
 *                          ty,        type
 *                                       2:  object
 *                                       4:  boolean
 *                                       5:  char
 *                                       6:  float
 *                                       7:  double
 *                                       8:  byte
 *                                       9:  short
 *                                       10: int
 *                                       11: long
 *                          vl]*       and value
 *
 *                          u2         number of static fields
 *                         [id,        static field name,
 *                          ty,        type,
 *                          vl]*       and value
 *
 *                          u2         number of inst. fields (not inc. super)
 *                         [id,        instance field name,
 *                          ty]*       type
 *
 *               PROF_GC_INSTANCE_DUMP        dump of a normal object
 *
 *                          id         object ID
 *                          u4         stack trace serial number
 *                          id         class object ID
 *                          u4         number of bytes that follow
 *                         [vl]*       instance field values (class, followed
 *                                     by super, super's super ...)
 *
 *               PROF_GC_OBJ_ARRAY_DUMP       dump of an object array
 *
 *                          id         array object ID
 *                          u4         stack trace serial number
 *                          u4         number of elements
 *                          id         element class ID
 *                         [id]*       elements
 *
 *               PROF_GC_PRIM_ARRAY_DUMP      dump of a primitive array
 *
 *                          id         array object ID
 *                          u4         stack trace serial number
 *                          u4         number of elements
 *                          u1         element type
 *                                       4:  boolean array
 *                                       5:  char array
 *                                       6:  float array
 *                                       7:  double array
 *                                       8:  byte array
 *                                       9:  short array
 *                                       10: int array
 *                                       11: long array
 *                         [u1]*       elements
 */
public class BinBody_HeapDump extends BinBody {
  
  /** Heap dump constants */
  public static final byte PROF_GC_ROOT_UNKNOWN      = -1;
  public static final byte PROF_GC_ROOT_JNI_GLOBAL   = 0x01;
  public static final byte PROF_GC_ROOT_JNI_LOCAL    = 0x02;
  public static final byte PROF_GC_ROOT_JAVA_FRAME   = 0x03;
  public static final byte PROF_GC_ROOT_NATIVE_STACK = 0x04;
  public static final byte PROF_GC_ROOT_STICKY_CLASS = 0x05;
  public static final byte PROF_GC_ROOT_THREAD_BLOCK = 0x06;
  public static final byte PROF_GC_ROOT_MONITOR_USED = 0x07;
  public static final byte PROF_GC_ROOT_THREAD_OBJ   = 0x08;
  
  public static final byte PROF_GC_CLASS_DUMP        = 0x20;
  public static final byte PROF_GC_INSTANCE_DUMP     = 0x21;
  public static final byte PROF_GC_OBJ_ARRAY_DUMP    = 0x22;
  public static final byte PROF_GC_PRIM_ARRAY_DUMP   = 0x23;
  
  private boolean dump_heap_input = Boolean.getBoolean("dump.heap.input");
  
  private HeapDump_ClassDump     hd_ClassDump;
  private HeapDump_InstanceDump  hd_InstanceDump;
  private HeapDump_ObjArrayDump  hd_ObjArrayDump;
  private HeapDump_PrimArrayDump hd_PrimArrayDump;
  
  public BinBody_HeapDump(BinRecord parent) {
    super(parent);
    hd_ClassDump     = new HeapDump_ClassDump(this);
    hd_InstanceDump  = new HeapDump_InstanceDump(this);
    hd_ObjArrayDump  = new HeapDump_ObjArrayDump(this);
    hd_PrimArrayDump = new HeapDump_PrimArrayDump(this);
  }
  
  public void read() throws IOException {
    byte     subRecordType;
    HeapDump hdump;
    int      lengthRemain = parent.getLength();
    
    namespace.getClsdumpTable().clear();
    namespace.getObjdumpTable().clear();
    
    while (lengthRemain > 0) {
      subRecordType = in.read_u1();
      
      hdump = createDump(subRecordType);
      if (hdump == null) {
        in.skip(lengthRemain - 1);
        lengthRemain = 0;
        in.profilerErrorMsg(new IOException("Unknown heap dump sub-record type = " + subRecordType));
        continue;
      }
      lengthRemain -= hdump.read() + 1;
      if (dump_heap_input) hdump.dump();
    }
    
    namespace.heap_primarr = new int[12];
    Object[] objects = namespace.getClsdumpTable().elements();
    for (int i = 0; i < objects.length; i++) {
      ClassDumpInfo clsDumpInfo = (ClassDumpInfo) objects[i];
      if (clsDumpInfo != null) {
        clsDumpInfo.superClassInfo  = namespace.lookupClassDump(clsDumpInfo.superClassID);
        clsDumpInfo.classLoaderInfo = namespace.lookupObjectDump(clsDumpInfo.classLoaderID);
      }
    }
    objects = namespace.getObjdumpTable().elements();
    for (int i = 0; i < objects.length; i++) {
      if (objects[i] instanceof ObjectDumpInfo) {
        ObjectDumpInfo objDumpInfo = (ObjectDumpInfo) objects[i];
        objDumpInfo.classDumpInfo = namespace.lookupClassDump(objDumpInfo.classObjectID);
        if (objDumpInfo.classDumpInfo != null) objDumpInfo.classDumpInfo.numberOfInstances++;
        objDumpInfo.readValues();
      }
      if (objects[i] instanceof ObjArrayDumpInfo) {
        ObjArrayDumpInfo objDumpInfo = (ObjArrayDumpInfo) objects[i];
        objDumpInfo.classDumpInfo = namespace.lookupClassDump(objDumpInfo.classObjectID);
        if (objDumpInfo.classDumpInfo != null) objDumpInfo.classDumpInfo.numberOfInstances++;
      }
      if (objects[i] instanceof PrimArrayDumpInfo) {
        PrimArrayDumpInfo objDumpInfo = (PrimArrayDumpInfo) objects[i];
        namespace.heap_primarr[objDumpInfo.type]++;
      }
    }
    in.eventHeapDump(namespace);
  }
  
  protected HeapDump createDump(byte subRecordType) throws IOException {
    switch (subRecordType) {
      case PROF_GC_ROOT_UNKNOWN:
        return new HeapDump_RootUnknown(this);
      
      case PROF_GC_ROOT_THREAD_OBJ:
        return new HeapDump_RootThreadObj(this);
      
      case PROF_GC_ROOT_JNI_GLOBAL:
        return new HeapDump_RootJniGlobal(this);
      
      case PROF_GC_ROOT_JNI_LOCAL:
        return new HeapDump_RootJniLocal(this);
      
      case PROF_GC_ROOT_JAVA_FRAME:
        return new HeapDump_RootJavaFrame(this);
      
      case PROF_GC_ROOT_NATIVE_STACK:
        return new HeapDump_RootNativeStack(this);
      
      case PROF_GC_ROOT_STICKY_CLASS:
        return new HeapDump_RootStickyClass(this);
      
      case PROF_GC_ROOT_THREAD_BLOCK:
        return new HeapDump_RootThreadBlock(this);
      
      case PROF_GC_ROOT_MONITOR_USED:
        return new HeapDump_RootMonitorUsed(this);
      
      case PROF_GC_CLASS_DUMP:
        return hd_ClassDump;
      
      case PROF_GC_INSTANCE_DUMP:
        return hd_InstanceDump;
      
      case PROF_GC_OBJ_ARRAY_DUMP:
        return hd_ObjArrayDump;
      
      case PROF_GC_PRIM_ARRAY_DUMP:
        return hd_PrimArrayDump;
      
      default:
        Log.printDebug("Unknown heap dump sub-record type = " + subRecordType);
        return null;
        //throw new IOException("Unknown heap dump sub-record type = " + subRecordType);
    }
  }
  
  // Dumpable
  
  public void dump(java.io.PrintStream out) {
    // dump is removed
  }
}