/*
 * Copyright (c) 2000 by Prosyst Software AG.,
 * http://www.prosyst.com
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Prosyst Software AG. You shall not disclose such Confidential
 * Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Prosyst.
 */

package com.prosyst.jpd.profiler.binbody;

import com.prosyst.jpd.profiler.BinInputImpl;
import com.prosyst.jpd.profiler.storage.ProfilerNamespace;

import com.prosyst.util.log.Dumpable; // from Util project

import java.io.IOException;

/**
 * <PRE>
 *
 * u1         a TAG denoting the type of the record
 * u4         number of *microseconds* since the time stamp in the
 *            Record. (wraps around in a little more than an hour)
 * u4         number of bytes *remaining* in the record. Note that
 *            this number excludes the tag and the length field itself.
 * [u1]*      BODY of the record (a sequence of bytes)
 *
 */
public class BinRecord implements Dumpable {
  
  /** The following tags supported */
  public static final int PROF_UTF8                = 0x01;
  public static final int PROF_LOAD_CLASS          = 0x02;
  public static final int PROF_UNLOAD_CLASS        = 0x03;
  public static final int PROF_FRAME               = 0x04;
  public static final int PROF_TRACE               = 0x05;
  public static final int PROF_ALLOC_SITES         = 0x06;
  public static final int PROF_HEAP_SUMMARY        = 0x07;
  public static final int PROF_START_THREAD        = 0x0a;
  public static final int PROF_END_THREAD          = 0x0b;
  public static final int PROF_HEAP_DUMP           = 0x0c;
  public static final int PROF_CPU_SAMPLES         = 0x0d;
  public static final int PROF_CONTROL_SETTINGS    = 0x0e;
  public static final int PROF_LOCKSTATS_WAIT_TIME = 0x10;
  public static final int PROF_LOCKSTATS_HOLD_TIME = 0x11;
  
  /** Our tags */
  public static final int PROF_GRAPH_INFO          = 0x40;
  public static final int PROF_MONITOR_DUMP        = 0x41;
  public static final int PROF_TRIGGER_EVENT       = 0x42;
  public static final int PROF_MONITOR_ENTER       = 0x50;
  public static final int PROF_MONITOR_ENTERED     = 0x51;
  public static final int PROF_MONITOR_EXIT        = 0x52;
  public static final int PROF_MONITOR_WAIT        = 0x53;
  public static final int PROF_MONITOR_WAITED      = 0x54;
  
  /** Monitor dump constants */
  public static final int JVMPI_MONITOR_JAVA       = 0x01;
  public static final int JVMPI_MONITOR_RAW        = 0x02;
  
  /** A Tag denoting the type of the record. */
  public byte type;
  
  /** Number of microseconds since the time stamp in the Record. */
  public int ticks;
  
  /** Number of bytes remaining in the record. */
  public int length;
  
  /** Body of the record */
  public BinBody body;
  
  protected BinInputImpl in = null;
  
  protected ProfilerNamespace namespace = null;
  
  private BinBody_Utf8            bb_Utf8;
  private BinBody_LoadClass       bb_LoadClass;
  private BinBody_UnloadClass     bb_UnloadClass;
  private BinBody_Frame           bb_Frame;
  private BinBody_Trace           bb_Trace;
  private BinBody_AllocSites      bb_AllocSites;
  //private BinBody_HeapSummary     bb_HeapSummary; // rarely used event
  private BinBody_StartThread     bb_StartThread;
  private BinBody_EndThread       bb_EndThread;
  private BinBody_HeapDump        bb_HeapDump;
  private BinBody_CpuSamples      bb_CpuSamples;
  //private BinBody_ControlSettings bb_ControlSettings; // rarely used event
  private BinBody_GraphInfo       bb_GraphInfo;
  
  public BinRecord(BinInputImpl in) {
    init(in);
  }
  
  public void init(BinInputImpl in) {
    this.in = in;
    this.namespace = in.getNamespace();
    
    bb_Utf8            = new BinBody_Utf8(this);
    bb_LoadClass       = new BinBody_LoadClass(this);
    bb_UnloadClass     = new BinBody_UnloadClass(this);
    bb_Frame           = new BinBody_Frame(this);
    bb_Trace           = new BinBody_Trace(this);
    bb_AllocSites      = new BinBody_AllocSites(this);
    //bb_HeapSummary     = new BinBody_HeapSummary(this); // rarely used event
    bb_StartThread     = new BinBody_StartThread(this);
    bb_EndThread       = new BinBody_EndThread(this);
    bb_HeapDump        = new BinBody_HeapDump(this);
    bb_CpuSamples      = new BinBody_CpuSamples(this);
    //bb_ControlSettings = new BinBody_ControlSettings(this); // rarely used event
    bb_GraphInfo       = new BinBody_GraphInfo(this);
  }
  
  public void read() throws IOException {
    type   = in.read_u1();
    ticks  = in.read_u4() / 1000;
    length = in.read_u4();
    
    body = createBody();
    body.read();
  }
  
  public BinBody getBody() {
    return body;
  }
  
  public byte getType() {
    return type;
  }
  
  public int getLength() {
    return length;
  }
  
  public ProfilerNamespace getNamespace() {
    return namespace;
  }
  
  protected BinBody createBody() throws IOException {
    switch (type) {
      case PROF_UTF8:
        return bb_Utf8;
      
      case PROF_LOAD_CLASS:
        return bb_LoadClass;
      
      case PROF_UNLOAD_CLASS:
        return bb_UnloadClass;
      
      case PROF_FRAME:
        return bb_Frame;
      
      case PROF_TRACE:
        return bb_Trace;
      
      case PROF_ALLOC_SITES:
        return bb_AllocSites;
      
      case PROF_HEAP_SUMMARY:
        return new BinBody_HeapSummary(this); // rarely used event
      
      case PROF_START_THREAD:
        return bb_StartThread;
      
      case PROF_END_THREAD:
        return bb_EndThread;
      
      case PROF_HEAP_DUMP:
        return bb_HeapDump;
      
      case PROF_CPU_SAMPLES:
        return bb_CpuSamples;
      
      case PROF_CONTROL_SETTINGS:
        return new BinBody_ControlSettings(this); // rarely used event
      
      case PROF_LOCKSTATS_WAIT_TIME:
        // return new BinBody_LockstatsWaitTime(this); // unused event
        throw new IOException("PROF_LOCKSTATS_WAIT_TIME");
      
      case PROF_LOCKSTATS_HOLD_TIME:
        // return new BinBody_LockstatsHoldTime(this); // unused event
        throw new IOException("PROF_LOCKSTATS_HOLD_TIME");
      
      case PROF_GRAPH_INFO:
        return bb_GraphInfo;
      
      case PROF_MONITOR_DUMP:
        return new BinBody_MonitorDump(this); // rarely used event
      
      case PROF_TRIGGER_EVENT:
        return new BinBody_TriggerEvent(this); // rarely used event
      
      case PROF_MONITOR_ENTER:
      case PROF_MONITOR_ENTERED:
      case PROF_MONITOR_EXIT:
      case PROF_MONITOR_WAIT:
      case PROF_MONITOR_WAITED:
        return new BinBody_MonitorEvent(this, type); // rarely used event
      
      default:
        throw new IOException("Illegal record type = " + type);
    }
  }
  
  // Dumpable
  
  public void dump() {
    dump(System.out);
  }
  
  public void dump(java.io.PrintStream out) {
    out.println("type=" + type + ", ticks=" + ticks + ", length=" + length);
  }
}