/*
 * Copyright (c) 2000 by Prosyst Software AG.,
 * http://www.prosyst.com
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Prosyst Software AG. You shall not disclose such Confidential
 * Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Prosyst.
 */

package com.prosyst.jpd.profiler.gui;

import java.io.File;
import java.util.Locale;
import java.util.ResourceBundle;

import com.prosyst.jpd.SecurityManager;
import com.prosyst.gui.resource.PropertyHolder;
import com.prosyst.util.log.Log;                 // from Util project
import com.prosyst.util.io.IniFile;              // from Util project
import com.prosyst.util.io.IniSection;           // from Util project

/**
 * ResourceManager manages resource strings, ints, icons
 * and persistent properties in ini files.
 *
 * @author Ivo Karabashev
 * @version 1.0
 */
public class ResourceManager {
  
  private static final String propsName = "profiler";
  private static final String iniName   = SecurityManager.getJPDHome() + File.separator + "profiler.ini";
  
  private static ResourceBundle resBundle;
  private static PropertyHolder propHolder;
  private static IniFile        ini = new IniFile();
  
  static {
    loadProps();
    loadIni();
  }
  
  private static void loadProps() {
    try {
      String internationalization = System.getProperty("internationalization.enabled");
      if ((internationalization != null) && (internationalization.equalsIgnoreCase("true"))) {
        resBundle  = ResourceBundle.getBundle(propsName, Locale.getDefault());
        if (resBundle == null)  Log.print("Resource file 'profiler.properties' not found!");
      } else {
        propHolder = PropertyHolder.getBundle(propsName, ResourceManager.class.getClassLoader());
        if (propHolder == null) Log.print("Resource file 'profiler.properties' not found!");
      }
    } catch (Exception _) {
      Log.print("Error reading properties file.");
    }
  }
  
  private static void loadIni() {
    try {
      ini.load(new File(iniName));
    } catch (Exception _) {
      Log.printDebug("Error reading ini file.");
    }
  }
  
  public static void save() throws Exception {
    ini.save(new File(iniName));
  }
  
  public static String getProp(String key) {
    try {
             if (resBundle  != null) {
        return resBundle.getString(key);
      } else if (propHolder != null) {
        return propHolder.getString(key);
      }
    } catch (Exception _) { }
    return null;
  }
  
  //
  //  String
  //
  
  /**
   * Searches for the string with the specified key in the property file.
   * The method returns null if the property is not found.
   *
   * @param   key        The property key
   * @return             The string in the property file with the specified key
   */
  public static String getString(String key) {
    return getString(key, null);
  }
  
  /**
   * Searches for the string with the specified key in the property file.
   * The method returns the default value argument if the property is not found.
   *
   * @param   key        The property key
   * @param   defValue   A default value
   * @return             The string in the property file with the specified key
   */
  public static String getString(String key, String defValue) {
    String value = getProp(key);
    if (value == null) {
      Log.printDebug("RES.getString", "Error in " + key);
      return defValue;
    }
    return value;
  }
  
  /**
   * Searches for the string with the specified key in the ini file.
   * The method returns null if the property is not found.
   *
   * @param   section    The section
   * @param   key        The property key
   * @return             The string in the ini file with the specified key
   */
  public static String getPersistentString(String section, String key) {
    return ini.getValue(section, key);
  }
  
  /**
   * Searches for the string with the specified key in the ini file.
   * The method returns the default value argument if the property is not found.
   *
   * @param   section    The section
   * @param   key        The property key
   * @param   defValue   A default value
   * @return             The string in the ini file with the specified key
   */
  public static String getPersistentString(String section, String key, String defValue) {
    String s = ini.getValue(section, key);
    if (s != null) return s;
    else           return defValue;
  }
  
  /**
   * Sets the specified value to the specified key.
   * Neither the key nor the value can be null.
   *
   * @param   section    The section
   * @param   key        The property key
   * @param   value      The value
   */
  public static void setPersistentString(String section, String key, String value) {
    ini.setValue(section, key, value);
  }
  
  //
  //  Int
  //
  
  /**
   * Searches for the integer with the specified key in the property file.
   * The method returns "0" if the property is not found.
   *
   * @param   key        The property key
   * @return             The integer in the property file with the specified key
   */
  public static int getInt(String key) {
    return getInt(key, 0);
  }
  
  /**
   * Searches for the integer with the specified key in the property file.
   * The method returns the default value argument if the property is not found.
   *
   * @param   key        The property key
   * @param   defValue   A default value
   * @return             The integer in the property file with the specified key
   */
  public static int getInt(String key, int defValue) {
    try {
      return parseInt(getProp(key));
    } catch (NumberFormatException ex) {
      Log.printDebug("RES.getInt", "Error in " + key);
    }
    return defValue;
  }
  
  /**
   * Searches for the integer with the specified key in the ini file.
   * The method returns "0" if the property is not found.
   *
   * @param   section    The section
   * @param   key        The property key
   * @return             The integer in the ini file with the specified key
   */
  public static int getPersistentInt(String section, String key) {
    return getPersistentInt(section, key, 0);
  }
  
  /**
   * Searches for the integer with the specified key in the ini file.
   * The method returns the default value argument if the property is not found.
   *
   * @param   section    The section
   * @param   key        The property key
   * @param   defValue   A default value
   * @return             The integer in the ini file with the specified key
   */
  public static int getPersistentInt(String section, String key, int defValue) {
    try {
      return parseInt(getPersistentString(section, key));
    } catch (NumberFormatException ex) {
      Log.printDebug("RES.getPersistentInt", "Error in " + key);
    }
    return defValue;
  }
  
  /**
   * Sets the specified integer to the specified key.
   * Neither the key nor the value can be null.
   *
   * @param   section    The section
   * @param   key        The property key
   * @param   value      The integer
   */
  public static void setPersistentInt(String section, String key, int value) {
    ini.setValue(section, key, Integer.toString(value, 10));
  }
  
  public static int parseInt(String value) throws NumberFormatException {
    if (value != null && value.startsWith("0x")) {
      return Integer.parseInt(value.substring(2), 16);
    } else {
      return Integer.parseInt(value);
    }
  }
  
  //
  //  Boolean
  //
  
  /**
   * Searches for the boolean with the specified key in the ini file.
   * The method returns "false" if the property is not found.
   *
   * @param   key        The property key
   * @return             The boolean in the property file with the specified key
   */
  public static boolean getPersistentBoolean(String section, String key) {
    return getPersistentBoolean(section, key, false);
  }
  
  /**
   * Searches for the boolean with the specified key in the ini file.
   * The method returns the default value argument if the property is not found.
   *
   * @param   section    The section
   * @param   key        The property key
   * @param   defValue   A default value
   * @return             The boolean in the ini file with the specified key
   */
  public static boolean getPersistentBoolean(String section, String key, boolean defValue) {
    String s = ini.getValue(section, key);
    if (s != null) return convertBoolean(s);
    Log.printDebug("RES.getPersistentBoolean", "Error in " + key);
    return defValue;
  }
  
  /**
   * Sets the specified boolean to the specified key.
   * Neither the key nor the value can be null.
   *
   * @param   section    The section
   * @param   key        The property key
   * @param   value      The boolean
   */
  public static void setPersistentBoolean(String section, String key, boolean value) {
    ini.setValue(section, key, String.valueOf(value));
  }
  
  public static boolean convertBoolean(String value) {
    return "true".equalsIgnoreCase(value);
  }
  
  public static boolean convertBoolean(String value, boolean defValue) {
    if (value == null) return defValue;
    return "true".equalsIgnoreCase(value);
  }
  
  public static void deletePersistentKey(String section, String key) {
    IniSection iniSection = ini.getSection(section, false);
    if (iniSection != null) iniSection.deleteKey(key);
  }
}