/*
 * Copyright (c) 2000 by Prosyst Software AG.,
 * http://www.prosyst.com
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Prosyst Software AG. You shall not disclose such Confidential
 * Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Prosyst.
 */

package com.prosyst.jpd.profiler.gui;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.event.AdjustmentListener;
import java.awt.event.AdjustmentEvent;

import com.prosyst.pgui.PComponent;
import com.prosyst.pgui.PScrollBar;

public class VMInfoComponent extends PComponent implements AdjustmentListener {
  
  /** Shows the whole graph in the component. */
  public static final int SCALE_ALL  = 0;
  /** Shows one minute from the graph in the component. */
  public static final int SCALE_1MIN = 60;
  /** Shows two minutes from the graph in the component. */
  public static final int SCALE_2MIN = 120;
  /** Shows five minutes from the graph in the component. */
  public static final int SCALE_5MIN = 300;
  
  /** Views live and total bytes graph. */
  public static final int VIEW_BYTES   = 0;
  /** Views live and total insts graph. */
  public static final int VIEW_INSTS   = 1;
  /** Views all threads and active threads graph. */
  public static final int VIEW_THREADS = 2;
  /** Views classes graph. */
  public static final int VIEW_CLASSES = 3;
  /** Views gc activity graph. */
  public static final int VIEW_GC      = 4;
  
  private static final String labelEmpty         = "";
  private static final String labelLiveBytes     = ResourceManager.getString("VMInfoPanel.label.livebytes");
  private static final String labelAllocBytes    = ResourceManager.getString("VMInfoPanel.label.allocbytes");
  private static final String labelLiveInsts     = ResourceManager.getString("VMInfoPanel.label.liveinsts");
  private static final String labelAllocInsts    = ResourceManager.getString("VMInfoPanel.label.allocinsts");
  private static final String labelThreads       = ResourceManager.getString("VMInfoPanel.label.threads");
  private static final String labelActiveThreads = ResourceManager.getString("VMInfoPanel.label.activethreads");
  private static final String labelClasses       = ResourceManager.getString("VMInfoPanel.label.classes");
  private static final String labelGCActivity    = ResourceManager.getString("VMInfoPanel.label.gcactivity");
  private static final String labelLastGCTime    = ResourceManager.getString("VMInfoPanel.label.lastgctime");
  
  private static final int      POINTS_INCREASE  = 120;
  
  public static Color BKG_COLOR        = new Color(ResourceManager.getInt("ColorBkgrnd", 0xF7F7FB));
  public static Color SCALE_COLOR      = new Color(ResourceManager.getInt("ColorScale",  0xD0E0E8));
  public static Color SCALE_COLOR_DARK = new Color(ResourceManager.getInt("ColorScaleD", 0x7080A0));
  public static Color SCALE_TIME_COLOR = new Color(ResourceManager.getInt("ColorScaleT", 0x000000));
  
  public  static final int      GRAPHS_NUM       = 8;
  private static final Color[]  GRAPHS_COLORS    = { Color.red,  Color.blue,    Color.green,  Color.cyan,
                                                     Color.pink, Color.magenta, Color.orange, Color.yellow };
  
  private PScrollBar scrollBar;
  
  // size dependant variables
  private int        view_mode  = VIEW_BYTES;
  private int        scale_mode = SCALE_1MIN;
  private int        scale_step_w;
  private int        scale_step_time;
  private int        width;
  private int        height;
  private int        extent;
  
  // graph points coherent variables
  private int        count;
  private int        index;
  private VMPoint[]  points;
  private float[]    points_max;
  private long       last_gc_time;
  
  // Constructor
  
  public VMInfoComponent(PScrollBar sb) {
    scrollBar = sb;
    _clear();
  }
  
  // Properties
  
  /* Clears variables and repaints. */
  public void clear() {
    _clear();
    repaintGraph();
  }
  
  /* Only clears variables. */
  public void _clear() {
    count = 0;
    index = 0;
    points = new VMPoint[POINTS_INCREASE];
    points_max = new float[GRAPHS_NUM];
    last_gc_time = 0;
    scrollBar.setValues(0, 1, 0, 1);
    scrollBar.setEnabled(false);
    long[] p = { 0, 0, 0, 0, 0, 0, 0, 0 };
    addPoint(new VMPoint(p));
  }
  
  /* Changes the display mode between SCALE_ALL, SCALE_1MIN, SCALE_2MIN, SCALE_5MIN. */
  public void setScale(int sm) {
    switch (sm) {
      case 0:
        scale_mode = SCALE_ALL;  break;
      case 1:
        scale_mode = SCALE_1MIN; break;
      case 2:
        scale_mode = SCALE_2MIN; break;
      case 3:
        scale_mode = SCALE_5MIN; break;
    }
    index = 0;
    scrollBar.setValues(0, 1, 0, 1);
    adjustHScale();
    adjustScroll();
    repaintGraph();
  }
  
  /* Changes the display mode between BYTES, INSTS, THREADS, CLASSES. */
  public void setView(int v) {
    view_mode = v;
    repaintGraph();
  }
  
  /* Adds a new point to the graph. */
  public void addPoint(VMPoint vmp) {
    if (count == points.length) {
      VMPoint[] newpoints = new VMPoint[points.length + POINTS_INCREASE];
      System.arraycopy(points, 0, newpoints, 0, points.length);
      points = newpoints;
    }
    points[count++] = vmp;
    
    for (int i = 0; i < GRAPHS_NUM; i++) {
      if (vmp.points[i] > points_max[i]) {
        int n = 0;
        long j = vmp.points[i];
        if (j < 10) points_max[i] = 10F;
        else {
          while (j > 9) { j /= 10; n++; }
          j++;
          while (n > 0) { j *= 10; n--; }
          points_max[i] = (float) j;
        }
      }
    }
    
    if (scale_mode == SCALE_ALL) adjustHScale();
    if (count > 1) scrollBar.setEnabled(true);
    adjustScroll();
  }
  
  /* Sets the first visible point of the graph. */
  public void setIndexPoint(int ip) {
    index = ip;
    repaintGraph();
  }
  
  /* Gets the label according to view. */
  public String getLabel() {
    if (count == 0) return labelEmpty;
    switch (view_mode) {
      case VIEW_BYTES:
        return new String(labelAllocBytes    + points[count - 1].points[2] + "  -  "
                        + labelLiveBytes     + points[count - 1].points[0]);
      case VIEW_INSTS:
        return new String(labelAllocInsts    + points[count - 1].points[3] + "  -  "
                        + labelLiveInsts     + points[count - 1].points[1]);
      case VIEW_THREADS:
        return new String(labelThreads       + points[count - 1].points[4] + "  -  "
                        + labelActiveThreads + points[count - 1].points[5]);
      case VIEW_CLASSES:
        return new String(labelClasses       + points[count - 1].points[6]);
      case VIEW_GC:
        if (points[count - 1].points[7] != 0) last_gc_time = points[count - 1].points[7];
        return new String(labelGCActivity    + points[count - 1].points[7]
                        + labelLastGCTime + 10 * last_gc_time + "ms");
    }
    return labelEmpty;
  }
  
  /* Intercepts changes in width and height and adjusts according to them. */
  public void setBounds(int x, int y, int w, int h) {
    if (height != h) {
      height = h;
    }
    if (width != w) {
      width = w;
      adjustHScale();
      adjustScroll();
    }
    super.setBounds(x, y, w, h);
  }
  
  private void adjustHScale() {
    if (scale_mode == SCALE_ALL) {
      if (count > 1) scale_step_w = width / count;
      else           scale_step_w = width;
      scale_step_time = count / 4;
    } else {
      scale_step_w = width / scale_mode;
      if (scale_mode == SCALE_5MIN) scale_step_time = 60;
      else                          scale_step_time = scale_mode / 6;
    }
    if (scale_step_w < 1) scale_step_w = 1;
    extent = (width - 2) / scale_step_w;
  }
  
  private void adjustScroll() {
    int max = scrollBar.getMaximum();
    int sbv = count - extent;
    if (sbv > 1) {
      if (scrollBar.getValue() >= max - extent) {
        scrollBar.setValues(count - extent - 1, extent, 0, count - 1);
      } else {
        scrollBar.setMaximum(count - 1);
      }
    } else {
      index = 0;
      scrollBar.setValues(0, 1, 0, 1);
      repaintGraph();
    }
  }
  
  // Paint
  
  public void repaintGraph() {
    if (!isShowing()) return;
    repaint(0, 0, width, height);
  }
  
  public void paint(Graphics g) {
    Rectangle r = g.getClipBounds();
    
    // clear background
    g.setColor(BKG_COLOR);
    g.fillRect(r.x, r.y, r.width, r.height);
    
    // draw scale
    drawScale(g);
    
    // draw graph
    int ip = index;
    for (int i = 0; i < width && ip < count - 1; i += scale_step_w) {
      switch (view_mode) {
        case VIEW_BYTES:
          g.setColor(GRAPHS_COLORS[2]);
          g.drawLine(i,                height - (int) (points[ip    ].points[2] / points_max[2] * height),
                     i + scale_step_w, height - (int) (points[ip + 1].points[2] / points_max[2] * height));
          g.setColor(GRAPHS_COLORS[0]);
          g.drawLine(i,                height - (int) (points[ip    ].points[0] / points_max[2] * height),
                     i + scale_step_w, height - (int) (points[ip + 1].points[0] / points_max[2] * height));
          break;
        case VIEW_INSTS:
          g.setColor(GRAPHS_COLORS[3]);
          g.drawLine(i,                height - (int) (points[ip    ].points[3] / points_max[3] * height),
                     i + scale_step_w, height - (int) (points[ip + 1].points[3] / points_max[3] * height));
          g.setColor(GRAPHS_COLORS[1]);
          g.drawLine(i,                height - (int) (points[ip    ].points[1] / points_max[3] * height),
                     i + scale_step_w, height - (int) (points[ip + 1].points[1] / points_max[3] * height));
          break;
        case VIEW_THREADS:
          g.setColor(GRAPHS_COLORS[4]);
          g.drawLine(i,                height - (int) (points[ip    ].points[4] / points_max[4] * height),
                     i + scale_step_w, height - (int) (points[ip + 1].points[4] / points_max[4] * height));
          g.setColor(GRAPHS_COLORS[5]);
          g.drawLine(i,                height - (int) (points[ip    ].points[5] / points_max[4] * height),
                     i + scale_step_w, height - (int) (points[ip + 1].points[5] / points_max[4] * height));
          break;
        case VIEW_CLASSES:
          g.setColor(GRAPHS_COLORS[6]);
          g.drawLine(i,                height - (int) (points[ip    ].points[6] / points_max[6] * height),
                     i + scale_step_w, height - (int) (points[ip + 1].points[6] / points_max[6] * height));
          break;
        case VIEW_GC:
          g.setColor(GRAPHS_COLORS[7]);
          g.drawLine(i,                height - (int) (points[ip    ].points[7] / points_max[7] * height),
                     i + scale_step_w, height - (int) (points[ip + 1].points[7] / points_max[7] * height));
          break;
      }
      ip++;
    }
  }
  
  private void drawScale(Graphics g) {
    g.setColor(SCALE_COLOR);
    for (int i = 0; i < width; i += scale_step_w) {
      g.drawLine(i, 0, i, height);
    }
    int z = 10;
    if (height < 200) z = 5;
    if (height < 100) z = 2;
    if (height < 50)  z = 1;
    for (int i = 0; i < z; i ++) {
      int j = i * height / z;
      g.drawLine(0, j, width, j);
    }
    g.setColor(SCALE_COLOR_DARK);
    for (int i = (10 - index % 10) * scale_step_w; i < width; i += scale_step_w * 10) {
      g.drawLine(i, 0, i, height - 18);
    }
    g.setColor(SCALE_TIME_COLOR);
    int bs = scale_step_time - index % scale_step_time;
    for (int i = bs * scale_step_w; i < width;
         i += scale_step_w * scale_step_time, bs += scale_step_time) {
      g.drawLine(i, height - 15, i, height - 12);
      g.drawString(convertTime(index + bs), i, height);
    }
    float max = 0;
    switch (view_mode) {
      case VIEW_BYTES:
        max = points_max[2]; break;
      case VIEW_INSTS:
        max = points_max[3]; break;
      case VIEW_THREADS:
        max = points_max[4]; break;
      case VIEW_CLASSES:
        max = points_max[6]; break;
      case VIEW_GC:
        max = points_max[7]; break;
    }
    if (max > 0) for (int i = 0; i < z; i ++) {
      int j = 10 + height * i / z;
      if (j < height - 20) g.drawString(convertSize(max - max * i / z), 0, j);
    }
  }
  
  private String convertTime(int time) {
    StringBuffer sb = new StringBuffer();
    sb.append(Integer.toString(time / 3600, 10));
    sb.append(":");
    time %= 3600;
    int t = time / 60;
    if (t < 10) sb.append('0');
    sb.append(Integer.toString(t, 10));
    sb.append(":");
    time %= 60;
    if (time < 10) sb.append('0');
    sb.append(Integer.toString(time, 10));
    return sb.toString();
  }
  
  private String convertSize(float fsize) {
    int size = (int) fsize;
    StringBuffer sb = new StringBuffer();
    if (size > 999) {
      sb.append(size / 100);
      sb.insert(sb.length() - 1, '.');
      sb.append("K");
    } else {
      sb.append(size);
    }
    return sb.toString();
  }
  
  // AdjustmentListener
  
  public void adjustmentValueChanged(AdjustmentEvent ev) {
    setIndexPoint(ev.getValue());
  }
}