/*
 * Copyright (c) 2000 by Prosyst Software AG.,
 * http://www.prosyst.com
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Prosyst Software AG. You shall not disclose such Confidential
 * Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Prosyst.
 */

package com.prosyst.jpd.profiler.heapdump;

import com.prosyst.jpd.profiler.binbody.BinBody_HeapDump;
import com.prosyst.jpd.profiler.info.ClassDumpInfo;
import com.prosyst.jpd.profiler.info.InstDumpInfo;

/**
 * Denote a heap dump -- Dump of a class object.
 *
 * <PRE>
 * TAG           BODY      notes
 * ---------------------------------------------------
 * PROF_GC_CLASS_DUMP           dump of a class object
 *
 *                          id         class object ID
 *                          u4         stack trace serial number
 *                          id         super class object ID
 *                          id         class loader object ID
 *                          id         signers object ID
 *                          id         protection domain object ID
 *                          id         reserved
 *                          id         reserved
 *
 *                          u4         instance size (in bytes)
 *
 *                          u2         size of constant pool
 *                          [u2,       constant pool index,
 *                           ty,       type
 *                                     2:  object
 *                                     4:  boolean
 *                                     5:  char
 *                                     6:  float
 *                                     7:  double
 *                                     8:  byte
 *                                     9:  short
 *                                     10: int
 *                                     11: long
 *                           vl]*      and value
 *
 *                          u2         number of static fields
 *                          [id,       static field name,
 *                           ty,       type,
 *                           vl]*      and value
 *
 *                          u2         number of inst. fields (not inc. super)
 *                          [id,       instance field name,
 *                           ty]*      type
 */
public class HeapDump_ClassDump extends HeapDump {

  public int classObjectID;
  public int numberOfStaticFields;
  public int numberOfInstFields; // ! not including super fields
  
  public ClassDumpInfo clsDumpInfo;
  
  public HeapDump_ClassDump(BinBody_HeapDump parent) {
    super(parent);
  }
  
  public int read() throws java.io.IOException {
    int bytesRead = 0;
    clsDumpInfo = new ClassDumpInfo();
    
                classObjectID       = in.read_id();
    clsDumpInfo.className           = namespace.lookupClassFromObjectID(classObjectID);
    
    int         stackTraceSerialNo  = in.read_u4();
    clsDumpInfo.traceInfo           = namespace.lookupTrace(stackTraceSerialNo);
    
    clsDumpInfo.superClassID        = in.read_id();
    
    clsDumpInfo.classLoaderID       = in.read_id();
    
                                      in.read_id(); // signersObjectID
                                      in.read_id(); // protectionDomainObjectID
                                      in.read_id(); // reserved
                                      in.read_id(); // reserved
    
    clsDumpInfo.instanceSizeInBytes = in.read_u4();
    
    int sizeOfConstantPool = in.read_u2();
    in.skip((3 + in.sizeofID()) * sizeOfConstantPool); // ! constants are ignored
    
    numberOfStaticFields = in.read_u2();
    clsDumpInfo.staticFields = new ClassDumpInfo.StaticField[numberOfStaticFields];
    for (int i = 0; i < numberOfStaticFields; i++) {
      ClassDumpInfo.StaticField statf = new ClassDumpInfo.StaticField();
      statf.name  = namespace.lookupName(in.read_id());
      statf.type  = in.read_u1();
      statf.value = InstDumpInfo.readInstance(statf.type, in);
      bytesRead += InstDumpInfo.sizeof(statf.type);
      clsDumpInfo.staticFields[i] = statf;
    }
    
    numberOfInstFields = in.read_u2(); // ! not including super fields
    clsDumpInfo.instanceFields = new ClassDumpInfo.InstanceField[numberOfInstFields];
    for (int i = 0; i < numberOfInstFields; i++) {
      ClassDumpInfo.InstanceField instf = new ClassDumpInfo.InstanceField();
      instf.name = namespace.lookupName(in.read_id());
      instf.type = in.read_u1();
      clsDumpInfo.instanceFields[i] = instf;
    }
    
    bind();
    
    // id + u4 + 6 * id + u4 + 3 * u2 + sizeOfConstantPool * (2 + 1 + id)
    // + numberOfStaticFields * (id + 1) + numberOfInstFields * (id + 1) + bytesRead
    return in.sizeofID() * (7 + sizeOfConstantPool + numberOfStaticFields + numberOfInstFields)
         + 3 * sizeOfConstantPool + numberOfStaticFields + numberOfInstFields + 14 + bytesRead;
  }
  
  public void bind() {
    namespace.getClsdumpTable().bind(classObjectID, clsDumpInfo);
  }
  
  // Dumpable
  
  public void dump(java.io.PrintStream out) {
    out.println("  sub-record type = PROF_GC_CLASS_DUMP");
    out.println("  class object ID           = " + Integer.toHexString(classObjectID));
    out.println("  class name                = " + clsDumpInfo.className);
    out.println("  super class id            = " + Integer.toHexString(clsDumpInfo.superClassID));
    
    out.println("  number of static fields   = " + numberOfStaticFields);
    for (int i = 0; i < numberOfStaticFields; i++) {
      ClassDumpInfo.StaticField statf = clsDumpInfo.staticFields[i];
      out.println("  static field[" + i + "]");
      out.println("    " + InstDumpInfo.toStringType(statf.type) + " " + statf.name
                + " = " + Integer.toHexString(statf.value));
    }
    
    out.println("  number of instance fields = " + numberOfInstFields);
    for (int i = 0; i < numberOfInstFields; i++) {
      ClassDumpInfo.InstanceField instf = clsDumpInfo.instanceFields[i];
      out.println("  instance field[" + i + "]");
      out.println("    " + InstDumpInfo.toStringType(instf.type) + " " + instf.name);
    }
    out.println();
  }
}