/*
 * Copyright (c) 2000 by Prosyst Software AG.,
 * http://www.prosyst.com
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Prosyst Software AG. You shall not disclose such Confidential
 * Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Prosyst.
 */

package com.prosyst.jpd.profiler.info;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import com.prosyst.jpd.profiler.storage.ObjectStorageInt;
import com.prosyst.jpd.profiler.gui.ResourceManager;
import com.prosyst.jpd.profiler.gui.dialogs.PreferencesDialog;

/**
 * Sorts alloc sites information
 * according to the order induced by the sortedBy value.
 *
 * @author Ivo Karabashev
 */
public class AllocSiteSorter {
  
  public static final int CLASS_COLUMN       = 0;
  public static final int BYTES_ALLOC_COLUMN = 1;
  public static final int INSTS_ALLOC_COLUMN = 2;
  public static final int BYTES_MARK_COLUMN  = 3;
  public static final int INSTS_MARK_COLUMN  = 4;
  public static final int THREAD_COLUMN      = 5;
  public static final int FRAME_COLUMN       = 6;
  
  public static final int ALLOCTABLE_ISIZE   = 101;
  
  private static final String INI_SMP_GC     = "sampler.disable_gc";
  
  private int      sortedBy      = BYTES_ALLOC_COLUMN;
  private boolean  ascending     = true;
  private boolean  isGCDisabled  = ResourceManager.getPersistentBoolean(PreferencesDialog.PROF_SECTION, INI_SMP_GC, false);
  private boolean  showZeroInsts = false;
  private boolean  isMarking     = true;
  private boolean  isGChanged    = true;
  private boolean  isFChanged    = true;
  private boolean  isSChanged    = true;
  private int      oldGLines     = 0;
  private int      oldFLines     = 0;
  private String[] cfiltersI, cfiltersE, mfiltersI, mfiltersE;
  private AllocSiteGroup lastGroup;
  
  private int              sitesNumber = 0;
  private Hashtable        allGroups = new Hashtable(ALLOCTABLE_ISIZE);
  private AllocSiteGroup[] sortedGroups;
  private AllocSiteInfo[]  sortedSites;
  
  /**
   * Adds array of allocation sites info in the object storage.
   *
   * @param   allocSites  array of allocation sites info
   */
  public void addAllocSites(AllocSiteInfo[] allocSites) {
    if (allocSites.length == 0) return;
    synchronized (allGroups) {
      isGChanged = true; isFChanged = true; isSChanged = true;
      AllocSiteInfo  oldsite,  newsite;
      AllocSiteGroup oldgroup, newgroup;
      int bl, il, ba, ia;
      for (int j = 0; j < allocSites.length; j++) {
        newsite = allocSites[j];
        oldgroup = (AllocSiteGroup) allGroups.get(newsite.className);
        if (oldgroup != null) {
          oldsite = (AllocSiteInfo) oldgroup.storage.lookup(newsite.stackTraceSerialNo);
          if (oldsite != null) {
            bl = newsite.bytesLive  - oldsite.bytesLive;
            il = newsite.instsLive  - oldsite.instsLive;
            ba = newsite.bytesAlloc - oldsite.bytesAlloc;
            ia = newsite.instsAlloc - oldsite.instsAlloc;
            if (isMarking) {
              oldgroup.bytesLiveMark  += bl;
              oldgroup.instsLiveMark  += il;
              oldgroup.bytesAllocMark += ba;
              oldgroup.instsAllocMark += ia;
              oldsite.bytesLiveMark  += bl;
              oldsite.instsLiveMark  += il;
              oldsite.bytesAllocMark += ba;
              oldsite.instsAllocMark += ia;
            }
            oldgroup.bytesLive  += bl;
            oldgroup.instsLive  += il;
            oldgroup.bytesAlloc += ba;
            oldgroup.instsAlloc += ia;
            oldsite.copyFrom(newsite);
          } else {
            oldgroup.addFrom(newsite, isMarking);
            if (isMarking) newsite.copyMark();
            oldgroup.storage.bind(newsite.stackTraceSerialNo, newsite);
            sitesNumber++;
          }
        } else {
          newgroup = new AllocSiteGroup(newsite, isMarking);
          if (isMarking) newsite.copyMark();
          allGroups.put(newgroup.className, newgroup);
          newgroup.storage.bind(newsite.stackTraceSerialNo, newsite);
          sitesNumber++;
        }
      }
    }
  }
  
  /**
   * Gets the unsorted array of all allocation sites info.
   *
   * @return     all sites
   */
  public AllocSiteInfo[] getAllSites() {
    synchronized (allGroups) {
      Enumeration     groups   = allGroups.elements();
      AllocSiteInfo[] allSites = new AllocSiteInfo[sitesNumber];
      
      int i = 0;
      while (groups.hasMoreElements()) {
        Object[] values = ((AllocSiteGroup) groups.nextElement()).storage.elements();
        AllocSiteInfo info;
        for (int j = 0; j < values.length; j++) {
          if (values[j] != null) {
            info = (AllocSiteInfo) values[j];
            if (showZeroInsts || isGCDisabled || info.instsLive != 0) allSites[i++] = info;
          }
        }
      }
      if (i != allSites.length) {
        AllocSiteInfo[] newAllSites = new AllocSiteInfo[i];
        System.arraycopy(allSites, 0, newAllSites, 0, i);
        return newAllSites;
      }
      return allSites;
    }
  }
  
  /**
   * Gets the sorted array of allocation sites groups that match the given filters.
   *
   * @param   lines  number of lines that should be displayed
   * @return     sorted groups
   */
  public AllocSiteGroup[] getFilteredGroups(int lines) {
    synchronized (allGroups) {
      if (isGChanged || oldGLines != lines) {
        isGChanged = false;
        oldGLines  = lines;
        
        Enumeration      values = allGroups.elements();
        AllocSiteGroup[] groups = new AllocSiteGroup[allGroups.size()];
        
        AllocSiteGroup group;
        int i = 0;
        if (cfiltersI == null && cfiltersE == null) {
          while (values.hasMoreElements()) {
            group = (AllocSiteGroup) values.nextElement();
            if (showZeroInsts || isGCDisabled || group.instsLive != 0) groups[i++] = group;
          }
        } else {
          while (values.hasMoreElements()) {
            group = (AllocSiteGroup) values.nextElement();
            if ((showZeroInsts || isGCDisabled || group.instsLive != 0) && matchesFilters(group)) groups[i++] = group;
          }
        }
        
        sort(groups, 0, i - 1);
        if (lines == -1 || lines > i) lines = i;
        if (lines == groups.length) {
          sortedGroups = groups;
        } else {
          sortedGroups = new AllocSiteGroup[lines];
          System.arraycopy(groups, 0, sortedGroups, 0, lines);
        }
      }
      return sortedGroups;
    }
  }
  
  /**
   * Gets the sorted array of allocation sites info that match the given filters.
   *
   * @param   lines  number of lines that should be displayed
   * @return     sorted sites
   */
  public AllocSiteInfo[] getFilteredSites(int lines) {
    synchronized (allGroups) {
      if (isFChanged || oldFLines != lines) {
        isFChanged = false;
        isSChanged = true;
        oldFLines  = lines;
        
        Enumeration     groups = allGroups.elements();
        AllocSiteInfo[] sites  = new AllocSiteInfo[sitesNumber];
        
        AllocSiteInfo info;
        int i = 0;
        while (groups.hasMoreElements()) {
          Object[] values = ((AllocSiteGroup) groups.nextElement()).storage.elements();

          if (cfiltersI == null && cfiltersE == null && mfiltersI == null && mfiltersE == null) {
            for (int j = 0; j < values.length; j++) {
              if (values[j] == null) continue;
              info = (AllocSiteInfo) values[j];
              if (showZeroInsts || isGCDisabled || info.instsLive != 0) sites[i++] = info;
            }
          } else {
            for (int j = 0; j < values.length; j++) {
              if (values[j] == null) continue;
              info = (AllocSiteInfo) values[j];
              if ((showZeroInsts || isGCDisabled || info.instsLive != 0)
                && matchesClassFilters(info) && matchesMethodFilters(info)) sites[i++] = info;
            }
          }
        }
        
        sort(sites, 0, i - 1);
        if (lines == -1 || lines > i) lines = i;
        if (lines == sites.length) {
          sortedSites = sites;
        } else {
          sortedSites = new AllocSiteInfo[lines];
          System.arraycopy(sites, 0, sortedSites, 0, lines);
        }
      }
      return sortedSites;
    }
  }
  
  /**
   * Gets the sorted array of allocation sites info of given group that match the given filters.
   *
   * @param   group  the allocation sites group to which they belong
   * @return     sorted sites
   */
  public AllocSiteInfo[] getFilteredSites(AllocSiteGroup group) {
    synchronized (allGroups) {
      if (isSChanged || lastGroup != group) {
        isSChanged = false;
        isFChanged = true;
        lastGroup = group;
        
        ObjectStorageInt storage = group.storage;
        Object[]         values  = storage.elements();
        AllocSiteInfo[]  sites   = new AllocSiteInfo[storage.size()];
        
        AllocSiteInfo info;
        int i = 0;
        if (mfiltersI == null && mfiltersE == null) {
          for (int j = 0; j < values.length; j++) {
            if (values[j] == null) continue;
            info = (AllocSiteInfo) values[j];
            if (showZeroInsts || isGCDisabled || info.instsLive != 0) sites[i++] = info;
          }
        } else {
          for (int j = 0; j < values.length; j++) {
            if (values[j] == null) continue;
            info = (AllocSiteInfo) values[j];
            if ((showZeroInsts || isGCDisabled || info.instsLive != 0) && matchesMethodFilters(info)) sites[i++] = info;
          }
        }
        
        sort(sites, 0, i - 1);
        if (i == sites.length) {
          sortedSites = sites;
        } else {
          sortedSites = new AllocSiteInfo[i];
          System.arraycopy(sites, 0, sortedSites, 0, i);
        }
      }
      return sortedSites;
    }
  }
  
  /**
   * Gets the filtered array of all allocation sites info that match the given trace and depth.
   *
   * @return     all sites
   */
  public Vector getAllSitesFromRoot() {
    synchronized (allGroups) {
      Vector allNodes = new Vector();
      if (initEnumeration()) return allNodes;
      
      AllocSiteInfo info;
      AllocSiteData asd;
      ThreadInfo    thread;
      next:
      while ((info = nextElement()) != null) {
        thread = info.getThread();
        for (int i = 0; i < allNodes.size(); i++) {
          asd = (AllocSiteData) allNodes.elementAt(i);
          if (asd.obj == thread) {
            asd.bytes += info.bytesAlloc;
            asd.insts += info.instsAlloc;
            continue next;
          }
        }
        allNodes.addElement(new AllocSiteData(thread, info.bytesAlloc, info.instsAlloc, null, true));
      }
      groupsG = null;
      valuesG = null;
      Misc.sort(allNodes, 0, allNodes.size() - 1);
      return allNodes;
    }
  }
  
  /**
   * Gets the filtered array of all allocation sites info that match the given trace and depth.
   *
   * @return     all sites
   */
  public Vector getAllSitesByTrace(TraceInfo trace) {
    synchronized (allGroups) {
      Vector allNodes = new Vector();
      if (initEnumeration()) return allNodes;
      
      int        depth  = trace.numberOfFrames;
      ThreadInfo thread = trace.threadInfo;
      
      AllocSiteInfo info;
      AllocSiteData asd;
      TraceInfo     infoTrace;
      FrameInfo     frame;
      int           d;
      next:
      while ((info = nextElement()) != null) {
        if (info.getThread() != thread) continue next;
        if (info.traceInfo == null || info.traceInfo.numberOfFrames == 0) {
          if (depth != 0) continue next;
          frame = null;
          d = 1;
        } else {
          infoTrace = info.traceInfo;
          if (infoTrace.numberOfFrames <= depth) continue next;
          d = infoTrace.numberOfFrames - depth;
          for (int i = 0; i < depth; i++) {
            if (infoTrace.frameInfo[d + i] != trace.frameInfo[i]) continue next;
          }
          frame = infoTrace.frameInfo[d - 1];
        }
        for (int i = 0; i < allNodes.size(); i++) {
          asd = (AllocSiteData) allNodes.elementAt(i);
          if (asd.obj == frame) {
            asd.bytes += info.bytesAlloc;
            asd.insts += info.instsAlloc;
            if (d == 1) asd.allocSite = info;
            if (!asd.hasChildren) asd.hasChildren = (d > 1);
            continue next;
          }
        }
        allNodes.addElement(new AllocSiteData(frame, info.bytesAlloc, info.instsAlloc, (d == 1 ? info : null), d > 1));
      }
      groupsG = null;
      valuesG = null;
      Misc.sort(allNodes, 0, allNodes.size() - 1);
      return allNodes;
    }
  }
  
  /**
   * Gets the filtered array of all allocation sites info that match the given trace and depth.
   *
   * @return     all sites
   */
  public Vector getSitesFromNode(TraceInfo trace, boolean down) {
    synchronized (allGroups) {
      Vector allNodes = new Vector();
      if (initEnumeration()) return allNodes;
      
      AllocSiteInfo info;
      next:
      while ((info = nextElement()) != null) {
        if (info.traceInfo == null) continue next;
        if (down) {
          if (info.traceInfo.compareTo(trace) < 0) continue next;
        } else {
          if (info.traceInfo.compareTo(trace) != 0) continue next;
        }
        if ((showZeroInsts || isGCDisabled || info.instsLive != 0)
          && matchesClassFilters(info) && matchesMethodFilters(info)) allNodes.addElement(info);
      }
      groupsG = null;
      valuesG = null;
      sort(allNodes, 0, allNodes.size() - 1);
      return allNodes;
    }
  }
  
  Enumeration groupsG;
  Object[]    valuesG;
  int         nextIndex;
  
  private boolean initEnumeration() {
    groupsG = allGroups.elements();
    if (groupsG.hasMoreElements()) {
      valuesG   = ((AllocSiteGroup) groupsG.nextElement()).storage.elements();
      nextIndex = 0;
      return false;
    } else {
      groupsG = null;
      valuesG = null;
      return true;
    }
  }
  
  private AllocSiteInfo nextElement() {
    for (;;) {
      if (nextIndex >= valuesG.length) {
        if (groupsG.hasMoreElements()) {
          valuesG   = ((AllocSiteGroup) groupsG.nextElement()).storage.elements();
          nextIndex = 0;
          continue;
        } else {
          return null;
        }
      }
      if (valuesG[nextIndex] == null) { nextIndex++; continue; }
      AllocSiteInfo info = (AllocSiteInfo) valuesG[nextIndex++];
      if (showZeroInsts || isGCDisabled || info.instsLive != 0) return info;
    }
  }
  
  private boolean matchesFilters(AllocSiteGroup info) {
    if (cfiltersI == null) {
      for (int i = 0; i < cfiltersE.length; i++) {
        if (info.className.indexOf(cfiltersE[i]) != -1) return false;
      }
      return true;
    } else {
      if (cfiltersE != null) {
        for (int i = 0; i < cfiltersE.length; i++) {
          if (info.className.indexOf(cfiltersE[i]) != -1) return false;
        }
      }
      for (int i = 0; i < cfiltersI.length; i++) {
        if (info.className.indexOf(cfiltersI[i]) != -1) return true;
      }
      return false;
    }
  }
  
  private boolean matchesClassFilters(AllocSiteInfo info) {
    if (cfiltersI == null) {
      if (cfiltersE != null) {
        for (int i = 0; i < cfiltersE.length; i++) {
          if (info.className.indexOf(cfiltersE[i]) != -1) return false;
        }
      }
      return true;
    } else {
      if (cfiltersE != null) {
        for (int i = 0; i < cfiltersE.length; i++) {
          if (info.className.indexOf(cfiltersE[i]) != -1) return false;
        }
      }
      for (int i = 0; i < cfiltersI.length; i++) {
        if (info.className.indexOf(cfiltersI[i]) != -1) return true;
      }
      return false;
    }
  }
  
  private boolean matchesMethodFilters(AllocSiteInfo info) {
    if (mfiltersI == null) {
      if (mfiltersE != null) {
        for (int i = 0; i < mfiltersE.length; i++) {
          if (info.getFrameName().indexOf(mfiltersE[i]) != -1) return false;
        }
      }
      return true;
    } else {
      if (mfiltersE != null) {
        for (int i = 0; i < mfiltersE.length; i++) {
          if (info.getFrameName().indexOf(mfiltersE[i]) != -1) return false;
        }
      }
      for (int i = 0; i < mfiltersI.length; i++) {
        if (info.getFrameName().indexOf(mfiltersI[i]) != -1) return true;
      }
      return false;
    }
  }
  
  /**
   * Gets the alloc site group from alloc site info.
   *
   * @param   site  allocation site info
   * @return        allocation site group
   */
  public AllocSiteGroup getAllocGroup(AllocSiteInfo site) {
    synchronized (allGroups) {
      return (AllocSiteGroup) allGroups.get(site.className);
    }
  }
  
  /**
   * Gets the row index in the table of the site.
   *
   * @param   site  allocation site info
   * @return        index of the row in the table of the site
   */
  public int getAllocSiteRow(AllocSiteInfo site) {
    synchronized (allGroups) {
      for (int i = 0; i < sortedSites.length; i++) {
        if (site == sortedSites[i]) return i;
      }
      return -1;
    }
  }
  
  /** Clears the allocation site data. */
  public void clear() {
    synchronized (allGroups) {
      isGChanged = true; isFChanged = true; isSChanged = true;
      isMarking = true;
      allGroups = new Hashtable(ALLOCTABLE_ISIZE);
      sitesNumber = 0;
    }
  }
  
  /** Starts marking the current state so that you can view the difference in mem. */
  public void startMark() {
    synchronized (allGroups) {
      isMarking = true;
      AllocSiteGroup group;
      Enumeration groups = allGroups.elements();
      while (groups.hasMoreElements()) {
        group = (AllocSiteGroup) groups.nextElement();
        group.startMark();
        Object[] values = group.storage.elements();
        for (int i = 0; i < values.length; i++) {
          if (values[i] != null) ((AllocSiteInfo) values[i]).startMark();
        }
      }
    }
  }
  
  /** Continues marking the current state so that you can view the difference in mem. */
  public void continueMark() {
    synchronized (allGroups) {
      isMarking = true;
    }
  }
  
  /** Stops marking the current state so that you can view the difference in mem. */
  public void stopMark() {
    synchronized (allGroups) {
      isMarking = false;
    }
  }
  
  /** Sets new filters. */
  public void setClassFilters(String[] newFiltersI, String[] newFiltersE) {
    synchronized (allGroups) {
      if (cfiltersI != newFiltersI || cfiltersE != newFiltersE) {
        isGChanged = true; isFChanged = true; isSChanged = true;
        cfiltersI = newFiltersI;
        cfiltersE = newFiltersE;
      }
    }
  }
  
  /** Sets new filters. */
  public void setMethodFilters(String[] newFiltersI, String[] newFiltersE) {
    synchronized (allGroups) {
      if (mfiltersI != newFiltersI || mfiltersE != newFiltersE) {
        isGChanged = true; isFChanged = true; isSChanged = true;
        mfiltersI = newFiltersI;
        mfiltersE = newFiltersE;
      }
    }
  }
  
  /**
   * Sets the state of the garbage collector.
   *
   * @param     if true the garbage collector is disabled
   */
  public void setGCDisabled(boolean state) {
    synchronized (allGroups) {
      isGChanged = true; isFChanged = true; isSChanged = true;
      isGCDisabled = state;
      ResourceManager.setPersistentBoolean(PreferencesDialog.PROF_SECTION, INI_SMP_GC, state);
      try {
        ResourceManager.save();
      } catch (Exception _) { }
    }
  }
  
  /**
   * Checks whether the garbage collector is disabled.
   *
   * @return     true if the garbage collector is disabled
   */
  public boolean isGCDisabled() {
    synchronized (allGroups) {
      return isGCDisabled;
    }
  }
  
  /**
   * Sets the state of the show zero insts.
   *
   * @param     if true zero insts are displayed
   */
  public void setShowZeroInsts(boolean state) {
    synchronized (allGroups) {
      isGChanged = true; isFChanged = true; isSChanged = true;
      showZeroInsts = state;
    }
  }
  
  /**
   * Sorts the allocation site data according to the specified column index.
   *
   * @param   column  column index
   */
  public void setSort(int column) {
    synchronized (allGroups) {
      isGChanged = true; isFChanged = true; isSChanged = true;
      if (sortedBy == column) {
        ascending = !ascending;
      } else {
        sortedBy  = column;
        ascending = true;
      }
    }
  }
  
  /**
   * Gets the column according to which the allocation site data is sorted.
   *
   * @return     index of the column
   */
  public int getColumn() {
    synchronized (allGroups) {
      return sortedBy;
    }
  }
  
  /**
   * Checks whether the allocation site data is in ascending order.
   *
   * @return     true if allocation site data is in ascending order
   */
  public boolean isAscending() {
    synchronized (allGroups) {
      return ascending;
    }
  }
  
  public void showAllocSites(java.io.OutputStream os, int column, int mode) {
    synchronized (allGroups) {
      java.io.PrintWriter out = new java.io.PrintWriter(os);
      int     sortBak = sortedBy;
      boolean ascBak  = ascending;
      boolean gcBak   = isGCDisabled;
      isGCDisabled = true;
      AllocSiteInfo[] allSites = getAllSites();
      if (column > 20) {
        column -= 20;
        isGCDisabled = false;
      }
      sortedBy     = column;
      ascending    = true;
      sort(allSites, 0, allSites.length - 1);
      sortedBy     = sortBak;
      ascending    = ascBak;
      isGCDisabled = gcBak;
      out.println("ALLOC SITES BEGIN");
      if (mode == AllocSiteInfo.BOTH_INFO) {
        out.println("                                    diff    diff    diff    diff");
        out.println("   alloc   alloc    live    live   alloc   alloc    live    live");
        out.println("   bytes   insts   bytes   insts   bytes   insts   bytes   insts  class, thread & method");
      }
      if (mode == AllocSiteInfo.ALLOC_INFO) {
        out.println("   alloc   alloc    live    live");
        out.println("   bytes   insts   bytes   insts  class, thread & method");
      }
      if (mode == AllocSiteInfo.DIFF_INFO) {
        out.println("    diff    diff    diff    diff");
        out.println("   alloc   alloc    live    live");
        out.println("   bytes   insts   bytes   insts  class, thread & method");
      }
      for (int j = 0; j < allSites.length; j++) {
        if (mode != AllocSiteInfo.DIFF_INFO || allSites[j].bytesAllocMark != 0)
          out.println(allSites[j].toTabbedString(mode));
      }
      out.println("ALLOC SITES END");
      out.println();
      out.println();
      out.flush();
      if (os != System.out) out.close();
    }
  }
  
  private void sort(AllocSiteGroup[] array, int lo0, int hi0) {
    if (lo0 >= hi0) return;
    int lo = lo0;
    int hi = hi0;
    AllocSiteGroup mid = array[(lo + hi) / 2];
    while (lo <= hi) {
      if (ascending) {
        if (isGCDisabled) {
          while ((lo < hi0) && (compareWithoutGC(array[lo], mid) < 0)) lo++;
          while ((lo0 < hi) && (compareWithoutGC(array[hi], mid) > 0)) hi--;
        } else {
          while ((lo < hi0) && (compareWithGC(array[lo], mid) < 0)) lo++;
          while ((lo0 < hi) && (compareWithGC(array[hi], mid) > 0)) hi--;
        }
      } else {
        if (isGCDisabled) {
          while ((lo < hi0) && (compareWithoutGC(array[lo], mid) > 0)) lo++;
          while ((lo0 < hi) && (compareWithoutGC(array[hi], mid) < 0)) hi--;
        } else {
          while ((lo < hi0) && (compareWithGC(array[lo], mid) > 0)) lo++;
          while ((lo0 < hi) && (compareWithGC(array[hi], mid) < 0)) hi--;
        }
      }
      if (lo <= hi) {
        AllocSiteGroup temp = array[lo];
        array[lo] = array[hi];
        array[hi] = temp;
        lo++;
        hi--;
      }
    }
    if (lo0 < hi) sort(array, lo0, hi);
    if (lo < hi0) sort(array, lo, hi0);
  }
  
  private int compareWithGC(AllocSiteGroup group1, AllocSiteGroup group2) {
    switch (sortedBy) {
      case BYTES_ALLOC_COLUMN:
        return group2.bytesLive      - group1.bytesLive;
      case INSTS_ALLOC_COLUMN:
        return group2.instsLive      - group1.instsLive;
      case BYTES_MARK_COLUMN:
        return group2.bytesLiveMark  - group1.bytesLiveMark;
      case INSTS_MARK_COLUMN:
        return group2.instsLiveMark  - group1.instsLiveMark;
      case CLASS_COLUMN:
        return Misc.compareToIgnoreCase(group1.className, group2.className);
      case THREAD_COLUMN:
      case FRAME_COLUMN:
        if (ascending) {
          return Misc.compareToIgnoreCase(group1.className, group2.className);
        } else {
          return Misc.compareToIgnoreCase(group2.className, group1.className);
        }
    }
    return 0;
  }
  
  private int compareWithoutGC(AllocSiteGroup group1, AllocSiteGroup group2) {
    switch (sortedBy) {
      case BYTES_ALLOC_COLUMN:
        return group2.bytesAlloc     - group1.bytesAlloc;
      case INSTS_ALLOC_COLUMN:
        return group2.instsAlloc     - group1.instsAlloc;
      case BYTES_MARK_COLUMN:
        return group2.bytesAllocMark - group1.bytesAllocMark;
      case INSTS_MARK_COLUMN:
        return group2.instsAllocMark - group1.instsAllocMark;
      case CLASS_COLUMN:
        return Misc.compareToIgnoreCase(group1.className, group2.className);
      case THREAD_COLUMN:
      case FRAME_COLUMN:
        if (ascending) {
          return Misc.compareToIgnoreCase(group1.className, group2.className);
        } else {
          return Misc.compareToIgnoreCase(group2.className, group1.className);
        }
    }
    return 0;
  }
  
  public void sort(AllocSiteInfo[] array, int lo0, int hi0) {
    if (lo0 >= hi0) return;
    int lo = lo0;
    int hi = hi0;
    AllocSiteInfo mid = array[(lo + hi) / 2];
    while (lo <= hi) {
      if (ascending) {
        if (isGCDisabled) {
          while ((lo < hi0) && (compareWithoutGC(array[lo], mid) < 0)) lo++;
          while ((lo0 < hi) && (compareWithoutGC(array[hi], mid) > 0)) hi--;
        } else {
          while ((lo < hi0) && (compareWithGC(array[lo], mid) < 0)) lo++;
          while ((lo0 < hi) && (compareWithGC(array[hi], mid) > 0)) hi--;
        }
      } else {
        if (isGCDisabled) {
          while ((lo < hi0) && (compareWithoutGC(array[lo], mid) > 0)) lo++;
          while ((lo0 < hi) && (compareWithoutGC(array[hi], mid) < 0)) hi--;
        } else {
          while ((lo < hi0) && (compareWithGC(array[lo], mid) > 0)) lo++;
          while ((lo0 < hi) && (compareWithGC(array[hi], mid) < 0)) hi--;
        }
      }
      if (lo <= hi) {
        AllocSiteInfo temp = array[lo];
        array[lo] = array[hi];
        array[hi] = temp;
        lo++;
        hi--;
      }
    }
    if (lo0 < hi) sort(array, lo0, hi);
    if (lo < hi0) sort(array, lo, hi0);
  }
  
  public void sort(Vector v, int lo0, int hi0) {
    if (lo0 >= hi0) return;
    int lo = lo0;
    int hi = hi0;
    AllocSiteInfo mid = (AllocSiteInfo) v.elementAt((lo + hi) / 2);
    while (lo <= hi) {
      if (ascending) {
        if (isGCDisabled) {
          while ((lo < hi0) && (compareWithoutGC((AllocSiteInfo) v.elementAt(lo), mid) < 0)) lo++;
          while ((lo0 < hi) && (compareWithoutGC((AllocSiteInfo) v.elementAt(hi), mid) > 0)) hi--;
        } else {
          while ((lo < hi0) && (compareWithGC((AllocSiteInfo) v.elementAt(lo), mid) < 0)) lo++;
          while ((lo0 < hi) && (compareWithGC((AllocSiteInfo) v.elementAt(hi), mid) > 0)) hi--;
        }
      } else {
        if (isGCDisabled) {
          while ((lo < hi0) && (compareWithoutGC((AllocSiteInfo) v.elementAt(lo), mid) > 0)) lo++;
          while ((lo0 < hi) && (compareWithoutGC((AllocSiteInfo) v.elementAt(hi), mid) < 0)) hi--;
        } else {
          while ((lo < hi0) && (compareWithGC((AllocSiteInfo) v.elementAt(lo), mid) > 0)) lo++;
          while ((lo0 < hi) && (compareWithGC((AllocSiteInfo) v.elementAt(hi), mid) < 0)) hi--;
        }
      }
      if (lo <= hi) {
        Object temp = v.elementAt(lo);
        v.setElementAt(v.elementAt(hi), lo);
        v.setElementAt(temp, hi);
        lo++;
        hi--;
      }
    }
    if (lo0 < hi) sort(v, lo0, hi);
    if (lo < hi0) sort(v, lo, hi0);
  }
  
  private int compareWithGC(AllocSiteInfo info1, AllocSiteInfo info2) {
    switch (sortedBy) {
      case CLASS_COLUMN:
        int i = Misc.compareToIgnoreCase(info1.className, info2.className);
        if (i == 0) return info2.bytesLive - info1.bytesLive;
        else return i;
      case BYTES_ALLOC_COLUMN:
        return info2.bytesLive      - info1.bytesLive;
      case INSTS_ALLOC_COLUMN:
        return info2.instsLive      - info1.instsLive;
      case BYTES_MARK_COLUMN:
        return info2.bytesLiveMark  - info1.bytesLiveMark;
      case INSTS_MARK_COLUMN:
        return info2.instsLiveMark  - info1.instsLiveMark;
      case THREAD_COLUMN:
        return Misc.compareToIgnoreCase(info1.getThreadName(), info2.getThreadName());
      case FRAME_COLUMN:
        return Misc.compareToIgnoreCase(info1.getFrameName(),  info2.getFrameName());
    }
    return 0;
  }
  
  private int compareWithoutGC(AllocSiteInfo info1, AllocSiteInfo info2) {
    switch (sortedBy) {
      case CLASS_COLUMN:
        int i = Misc.compareToIgnoreCase(info1.className, info2.className);
        if (i == 0) return info2.bytesAlloc - info1.bytesAlloc;
        else return i;
      case BYTES_ALLOC_COLUMN:
        return info2.bytesAlloc     - info1.bytesAlloc;
      case INSTS_ALLOC_COLUMN:
        return info2.instsAlloc     - info1.instsAlloc;
      case BYTES_MARK_COLUMN:
        return info2.bytesAllocMark - info1.bytesAllocMark;
      case INSTS_MARK_COLUMN:
        return info2.instsAllocMark - info1.instsAllocMark;
      case THREAD_COLUMN:
        return Misc.compareToIgnoreCase(info1.getThreadName(), info2.getThreadName());
      case FRAME_COLUMN:
        return Misc.compareToIgnoreCase(info1.getFrameName(),  info2.getFrameName());
    }
    return 0;
  }
}