/*
 * Copyright (c) 2000 by Prosyst Software AG.,
 * http://www.prosyst.com
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Prosyst Software AG. You shall not disclose such Confidential
 * Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Prosyst.
 */

package com.prosyst.jpd.profiler.info;

import com.prosyst.jpd.profiler.storage.ObjectStorageInt;
import com.prosyst.jpd.profiler.storage.ObjectStorageIntDefaultImpl;

/**
 * Sorts cpu samples information
 * according to the order induced by the sortedBy value.
 *
 * @author Ivo Karabashev
 */
public class CpuSampleSorter {
  
  public static final int THREAD_COLUMN = 0;
  public static final int FRAME_COLUMN  = 1;
  public static final int COST_COLUMN_P = 2;
  public static final int COST_COLUMN_T = 3;
  
  private int      sortedBy  = COST_COLUMN_P;
  private boolean  ascending = true;
  private boolean  isChanged = true;
  private String[] filtersI, filtersE;
  
  private ObjectStorageInt storage = new ObjectStorageIntDefaultImpl();
  private CpuSampleInfo[]  sortedSamples;
  
  /**
   * Adds array of cpu samples info in the object storage.
   *
   * @param   cpuSamples  array of cpu samples info
   */
  public void addCpuSamples(CpuSampleInfo[] cpuSamples) {
    synchronized (storage) {
      isChanged = true;
      CpuSampleInfo oldsample, newsample;
      for (int j = 0; j < cpuSamples.length; j++) {
        newsample = cpuSamples[j];
        oldsample = (CpuSampleInfo) storage.lookup(newsample.stackTraceSerialNo);
        if (oldsample != null) {
          newsample.stackTraceCost += oldsample.stackTraceCost;
        }
        storage.bind(newsample.stackTraceSerialNo, newsample);
      }
    }
  }
  
  /**
   * Gets the sorted array of all cpu samples.
   *
   * @return     sorted samples
   */
  public CpuSampleInfo[] getAllSamples() {
    synchronized (storage) {
      Object[]        values     = storage.elements();
      CpuSampleInfo[] allSamples = new CpuSampleInfo[storage.size()];
      
      int i = 0;
      for (int j = 0; j < values.length; j++) {
        if (values[j] != null) allSamples[i++] = (CpuSampleInfo) values[j];
      }
      return allSamples;
    }
  }
  
  /**
   * Gets the sorted array of cpu samples info that match the given filters.
   *
   * @return     sorted samples
   */
  public CpuSampleInfo[] getFilteredSamples() {
    synchronized (storage) {
      if (isChanged) {
        isChanged = false;
        
        Object[]        values  = storage.elements();
        CpuSampleInfo[] samples = new CpuSampleInfo[storage.size()];
        
        int i = 0;
        if (filtersI == null && filtersE == null) {
          for (int j = 0; j < values.length; j++) {
            if (values[j] != null) samples[i++] = (CpuSampleInfo) values[j];
          }
          sort(samples, 0, i - 1);
          sortedSamples = samples;
        } else {
          CpuSampleInfo info;
          for (int j = 0; j < values.length; j++) {
            if (values[j] == null) continue;
            info = (CpuSampleInfo) values[j];
            if (matchesFilters(info)) samples[i++] = info;
          }
          sort(samples, 0, i - 1);
          if (i == samples.length) {
            sortedSamples = samples;
          } else {
            sortedSamples = new CpuSampleInfo[i];
            System.arraycopy(samples, 0, sortedSamples, 0, i);
          }
        }
      }
      return sortedSamples;
    }
  }
  
  private boolean matchesFilters(CpuSampleInfo info) {
    if (filtersI == null) {
      for (int i = 0; i < filtersE.length; i++) {
        if (info.frameName.indexOf(filtersE[i]) != -1) return false;
      }
      return true;
    } else {
      if (filtersE != null) {
        for (int i = 0; i < filtersE.length; i++) {
          if (info.frameName.indexOf(filtersE[i]) != -1) return false;
        }
      }
      for (int i = 0; i < filtersI.length; i++) {
        if (info.frameName.indexOf(filtersI[i]) != -1) return true;
      }
      return false;
    }
  }
  
  /**
   * Gets the sample from the row index in the table.
   *
   * @param   i  row in the table of the sample
   * @return     cpu sample info
   */
  public CpuSampleInfo getCpuSample(int i) {
    synchronized (storage) {
      return sortedSamples[i];
    }
  }
  
  /**
   * Gets the row index in the table of the sample.
   *
   * @param   sample  cpu sample info
   * @return          index of the row in the table of the sample
   */
  public int getCpuSampleRow(CpuSampleInfo sample) {
    synchronized (storage) {
      for (int i = 0; i < sortedSamples.length; i++) {
        if (sample == sortedSamples[i]) return i;
      }
      return -1;
    }
  }
  
  /** Clears the table data. */
  public void clear() {
    synchronized (storage) {
      isChanged = true;
      storage = new ObjectStorageIntDefaultImpl();
    }
  }
  
  /** Sets new filters. */
  public void setFilters(String[] newFiltersI, String[] newFiltersE) {
    synchronized (storage) {
      if (filtersI != newFiltersI || filtersE != newFiltersE) {
        isChanged = true;
        filtersI = newFiltersI;
        filtersE = newFiltersE;
      }
    }
  }
  
  /**
   * Sorts the table according to the specified column index.
   *
   * @param   column  column index
   */
  public void setSort(int column) {
    synchronized (storage) {
      isChanged = true;
      if (sortedBy == column) {
        ascending = !ascending;
      } else {
        sortedBy  = column;
        ascending = true;
      }
    }
  }
  
  /**
   * Gets the column according to which the table is sorted.
   *
   * @return     index of the column
   */
  public int getColumn() {
    synchronized (storage) {
      return sortedBy;
    }
  }
  
  /**
   * Checks whether the table is in ascending order.
   *
   * @return     true if table is in ascending order
   */
  public boolean isAscending() {
    synchronized (storage) {
      return ascending;
    }
  }
  
  public void showCpuSamples(java.io.OutputStream os, int column, int totalNumberOfSamples, int totalTime) {
    synchronized (storage) {
      java.io.PrintWriter out = new java.io.PrintWriter(os);
      CpuSampleInfo[] allSamples = getAllSamples();
      int     sortBak = sortedBy;
      boolean ascBak  = ascending;
      sortedBy  = column;
      ascending = true;
      sort(allSamples, 0, allSamples.length - 1);
      sortedBy  = sortBak;
      ascending = ascBak;
      out.println("CPU SAMPLES BEGIN");
      out.println("  time % time ms  thread & method");
      for (int j = 0; j < allSamples.length; j++) {
        out.println(allSamples[j].toTabbedString(totalNumberOfSamples, totalTime));
      }
      out.println("CPU SAMPLES END");
      out.println();
      out.println();
      out.flush();
      if (os != System.out) out.close();
    }
  }
  
  public void sort(CpuSampleInfo[] array, int lo0, int hi0) {
    if (lo0 >= hi0) return;
    int lo = lo0;
    int hi = hi0;
    CpuSampleInfo mid = array[(lo + hi) / 2];
    while (lo <= hi) {
      if (ascending) {
        while ((lo < hi0) && (compare(array[lo], mid) < 0)) lo++;
        while ((lo0 < hi) && (compare(array[hi], mid) > 0)) hi--;
      } else {
        while ((lo < hi0) && (compare(array[lo], mid) > 0)) lo++;
        while ((lo0 < hi) && (compare(array[hi], mid) < 0)) hi--;
      }
      if (lo <= hi) {
        CpuSampleInfo temp = array[lo];
        array[lo] = array[hi];
        array[hi] = temp;
        lo++;
        hi--;
      }
    }
    if (lo0 < hi) sort(array, lo0, hi);
    if (lo < hi0) sort(array, lo, hi0);
  }
  
  private int compare(CpuSampleInfo info1, CpuSampleInfo info2) {
    switch (sortedBy) {
      case COST_COLUMN_P:
      case COST_COLUMN_T:
        return info2.stackTraceCost - info1.stackTraceCost;
      case THREAD_COLUMN:
        return Misc.compareToIgnoreCase(info1.threadName, info2.threadName);
      case FRAME_COLUMN:
        return Misc.compareToIgnoreCase(info1.frameName,  info2.frameName);
    }
    return 0;
  }
}
