/*
 * Copyright (c) 2000 by Prosyst Software AG.,
 * http://www.prosyst.com
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Prosyst Software AG. You shall not disclose such Confidential
 * Information and shall use it only in accordance with the terms
 * of the license agreement you entered into with Prosyst.
 */

package com.prosyst.jpd.profiler.storage;

public class HashIntObjNS {
  
  static final int   DEFAULT_CAPACITY    = 101;
  static final float DEFAULT_LOAD_FACTOR = 0.95f;
  
  // count of elements available in table
  private int count = 0;
  // used for computation of next position
  private int step = 499979;
  
  public int[] keys;
  public Object[] values;
  public int[] next;
  
  private int limit;
  private double loadFactor;
  
  /**
   * Constructs an empty hash table with int keys and Object values.
   * Uses default load factor (0.75) and default capacity (101)
   */
  public HashIntObjNS() {
    this(DEFAULT_CAPACITY, DEFAULT_LOAD_FACTOR);
  }
  
  /**
   * Constructs an empty hash table with int keys and Object values.
   * Uses default load factor (0.75) and given capacity.
   *
   * @param   capacity  initial capacity of the table
   *
   * @exception IllegalArgumentException if <code>capacity</code> < 0.
   */
  public HashIntObjNS(int capacity) {
    this(capacity, DEFAULT_LOAD_FACTOR);
  }
  
  /**
   * Constructs an empty hash table with int keys and Object values.
   * Uses given load factor and given capacity.
   *
   * @param   capacity  initial capacity of the table
   * @param   lf        load factor ot the table
   *
   * @exception IllegalArgumentException if <code>capacity</code> < 0 or <code>lf</code> <= 0.0
   */
  public HashIntObjNS(int capacity, double lf) {
    if (capacity < 0) {
      throw new IllegalArgumentException("Invalid hashtable capacity: " + capacity);
    }
    if (lf <= 0) {
      throw new IllegalArgumentException("Invalid load factor: " + lf);
    }
    if (capacity == 0) capacity = DEFAULT_CAPACITY;
    if (lf > 1.0) lf = 1.0;
    
    loadFactor = lf;
    limit = (int) (capacity * lf);
    count = 0;
    
    keys = new int[capacity];
    values = new Object[capacity];
    next = new int[capacity];
    for (int i = 0; i < capacity; i++) {
      keys[i] = 0;
      next[i] = -1;
    }
  }
  
  /**
   * Maps the specified key to the specified value in this hashtable.
   * If an element with the specified key is already in the table only changes it's value.
   *
   * @param   key    the hashtable key
   * @param   value  the value
   */
  public void put(int key, Object value) {
    if (_put(key, value)) {
      count++;
    }
  }
  
  /**
   * Returns the value to which the specified key is mapped in this hashtable.
   *
   * @param   key  a key in the hashtable
   * @return  the value to which the key is mapped in this hashtable,
   *          null if the key is not mapped to any value in this hashtable
   */
  public Object get(int key) {
    int pos = find(key);
    return (pos == -1) ? null : values[pos];
  }
  
  /**
   * Removes the key (and its corresponding value) from this hashtable.
   * Does nothing if the key is not in the hashtable.
   *
   * @param   key  the key that needs to be removed
   */
  public Object remove(int key) {
    int pos = find(key);
    if (pos == -1) return null;
    next[pos] += next.length; // mark this field as empty
    count--;
    Object tmp = values[pos];
    values[pos] = null;
    return tmp;
  }
  
  /**
   * Removes all the elements from this hashtable.
   */
  public void removeAll() {
    for (int i = 0; i < values.length; i++) {
      values[i] = null;
      keys[i] = 0;
      next[i] = -1;
    }
    count = 0;
  }
  
  /**
   * Returns the count of elements currently in the table
   *
   * @return  the count of elements
   */
  public int size() {
    return count;
  }
  
  public int[] keys() {
    return keys;
  }
  
  public Object[] elements() {
    return values;
  }
  
  private int find(int key) {
    int pos = (int) (key & 0x7fffffff) % keys.length;
    int i = 0;
    
    while (next[pos] >= 0) {
      if (keys[pos] == key && next[pos] < next.length) return pos;
      if ((pos = next[pos]) >= next.length)            pos -= next.length;
      if (++i > next.length)                           return -1;
    }
    
    return -1;
  }
  
  private boolean _put(int key, Object value) {
    int index = find(key);
    if (index != -1) {
      values[index] = value;
      return false;
    }
    
    int pos = (int) (key & 0x7fffffff) % keys.length;
    
    while ((next[pos] >= 0) && (next[pos] < next.length)) pos = next[pos];
    
    keys[pos] = key;
    values[pos] = value;
    if (next[pos] < 0) next[pos] = (pos + step) % next.length;
    else               next[pos] -= next.length;
    return true;
  }
}